# Artifact for the Paper: Attack Analysis and Detection for the Combined Electric Vehicle Charging and Power Grid Domains
This repository provides the artifacts for the paper "Attack Analysis and Detection for the Combined Electric Vehicle Charging and Power Grid Domains" which was accepted at the 19th International Conference on Availability, Reliability and Security (ARES 2024).
In the paper, we propose an Intrusion Detection System (IDS) that combines regression-based charging load prediction with novelty detection-based anomaly identification. The proposed system considers features from both the EV charging and power grid domains, which is enabled by a novel co-simulation concept. 
Here, we provide the corresponding source code, which was used in the evaluation of the proposed solution.
Specifically, the code consists of an extension to 
[Wattson](https://github.com/fkie-cad/wattson/), which implements the co-simulation, and of [scikit](https://scikit-learn.org/stable/)-based code, which implements the IDS.
For EV charging data, we use the public
[ElaadNL](https://platform.elaad.io/download-data/)
data set.

The following sub-sections describe how to install the EV charging extension for Wattson (including the installation of Wattson based on their [README](https://github.com/fkie-cad/wattson/blob/master/README.md)), how to run the co-simulation, and how to run the IDS.

## Install Simulation Extension


```
#tested on ubuntu-22.04.3-desktop-amd64.iso
sudo apt update
sudo apt upgrade -y
sudo apt install -y python3-pip git gcc make perl docker.io python3-pyqt5

pip install pip==22.0.2

mkdir git
cd git

git clone https://github.com/fkie-cad/powerowl.git
cd powerowl
git reset --hard 1d51897c9313f0070d8f9b315aa941e4ab2f8ac2
sed -i -E "s/pandapower\.git/pandapower.git@v2.13.1/" setup.py
sed -i -E "s/plotly/plotly==5.18.0/" setup.py
sed -i -E "s/pandas/pandas==1.3.4/" setup.py

cd ..
sudo python3 -m pip install -e ./powerowl



git clone https://github.com/fkie-cad/wattson.git
cd wattson
git reset --hard 6329002f5dae3ac7990e17a0b479fed458dd6f9f
sed -i -E "s/pandapower\.git/pandapower.git@v2.13.1/" setup.py
sed -i -E "s/'powerowl/#'powerowl/" setup.py
sed -i -E "s/numba==0.57.1/numba==0.56.4/" setup.py
sed -i -E "s/packaging==20.3/packaging==21.3/" setup.py

cd ..


git clone https://code.fbi.h-da.de/seacop/ev-extension-for-wattson.git wattson-ev
wget "https://platform.elaad.io/download-data/filedownload.php?file=elaadnl_open_ev_datasets.xlsx&code=open" -O wattson-ev/scenarios/powerowl_example/extensions/cp_data/elaadnl_open_ev_datasets.xlsx

cp -r ./wattson-ev/wattson ./wattson/
cp -r ./wattson-ev/scenarios ./wattson/
cp ./wattson-ev/run_wattson.sh ./wattson/



sudo python3 wattson/setup.py wattson
sudo python3 -m pip install -e ./wattson

sudo pip install ujson pyoscp websockets ocpp flask CherryPy cheroot openpyxl
```


## Run Simulation

```
#Single simulation run with e.g.:
sudo python3 -m wattson wattson/scenarios/powerowl_example --no-cli --seed 5

#Full run (incl. attack and normal runs) with:
cd wattson
sudo bash run_wattson.sh
```


## IDS Dependencies and Execution:

```
#Install IDS dependencies with:
sudo pip install scikit-learn==1.3.2
cp -r ./wattson-ev/ids ./wattson/
```

Running the IDS requires the simulation results, which can be generated as described above.
Alternatively, simulation results can be downloaded from [elaadnl.zip](https://osf.io/fy8gp?view_only=5feb8899edd2419cb00fbf2668d3d9ae) (needs to be extracted by default to wattson-artifacts/).
The location of the simulation results can be configured in ids.conf

```
#Execute full IDS run (training and testing) with:
cd wattson/ids
bash run_ids.sh
```





