
from collections import defaultdict
from datetime import datetime
import gc
import gzip
import os
from pathlib import Path
import pickle
import re
import warnings
import numpy as np
import pandas as pd
from tqdm import tqdm

import json
from powerowl.simulators.pandapower import PandaPowerGridModel


import logging

logger = logging.getLogger("WATTSON_EV_IDS.FeaturesAUX")




def discretize_hour_balancing(ts):
    hour = ts.hour
    if hour in [7, 8, 9]:
        return "peak"
    elif hour in [16, 17, 18, 19, 20, 21]:
        return "peak"
    elif hour in [22, 23, 24, 0, 1, 2, 3, 4, 5, 6]:
        return "low"
    elif hour in [10, 11, 12, 13, 14, 15]:
        return "low"
def discretize_hour2(hour):
    if hour in [5, 6, 7, 8, 9, 10, 11]:
        return "Morning"
    elif hour in [12, 13, 14, 15, 16]:
        return "Afternoon"
    elif hour in [17, 18, 19, 20, 21]:
        return "Evening"
    elif hour in [21, 22, 23, 0, 1, 2, 3, 4]:
        return "Night"
def discretize_hour_only(hour):
    if hour in [9, 10, 11, 12, 13, 14, 15, 16]:
        return "Work"
    elif hour in [22, 23, 0, 1, 2, 3, 4, 5]:
        return "Sleep"
    elif hour in [17, 18, 19, 20, 21, 6, 7, 8]:
        return "Play"
def discretize_hour_ts(ts):
    hour = ts.hour
    return discretize_hour_only(hour)
def discretize_day_is_work(ts):
    day = ts.weekday()
    if day in [5, 6]:  # weekend
        return False
    else:
        return True
def discretize_hour_day(ts):
    hour = ts.hour
    day = ts.weekday()
    if day in [5, 6]:  # weekend
        if hour in [22, 23, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9]:
            return "High-Home"
        elif hour in [10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21]:
            return "High-Home"
            return "High-Leisure"
    else:
        if hour in [22, 23, 0, 1, 2, 3, 4, 5, 6, 7]:
            return "High-Home"
        elif hour in [8, 9, 10, 11, 12, 13, 14, 15, 16]:
            return "High-Work"
        elif hour in [17, 18, 19, 20, 21]:
            return "High-Home"
            return "High-Leisure"
        else:
            raise Exception("discretize_hour_only", hour)

def get_date_exog(sum_df, prefix):
    with warnings.catch_warnings():
        warnings.simplefilter('ignore', pd.errors.PerformanceWarning)
        # sum_df['day_name'] = sum_df.index.to_series().dt.day_name()
        sum_df[prefix+'dayofweek'] = sum_df.index.to_series().dt.dayofweek
        sum_df[prefix+'hour'] = sum_df.index.to_series().dt.hour
        sum_df[prefix+'discretize_hour_only'] = sum_df.index.to_series().apply(
            discretize_hour_ts)
        sum_df[prefix+'discretize_hour_day'] = sum_df.index.to_series().apply(
            discretize_hour_day)
        sum_df[prefix+'discretize_day_is_work'] = sum_df.index.to_series().apply(
            discretize_day_is_work)
        sum_df[prefix+'discretize_hour_balancing'] = sum_df.index.to_series().apply(
            discretize_hour_balancing)
    sum_df = sum_df.copy()


def get_date_exog_col(sum_df, col, prefix):
    with warnings.catch_warnings():
        warnings.simplefilter('ignore', pd.errors.PerformanceWarning)
        # sum_df['day_name'] = sum_df[[col]].to_series().dt.day_name()
        sum_df[prefix+col+'_dayofweek'] = sum_df[col].dt.dayofweek
        sum_df[prefix+col+'_hour'] = sum_df[col].dt.hour
        sum_df[prefix+col+'_discretize_hour_only'] = sum_df[col].apply(
            discretize_hour_ts)
        sum_df[prefix+col+'_discretize_hour_day'] = sum_df[col].apply(
            discretize_hour_day)
        sum_df[prefix+col+'_discretize_day_is_work'] = sum_df[col].apply(
            discretize_day_is_work)
        sum_df[prefix+col+'_discretize_hour_balancing'] = sum_df[col].apply(
            discretize_hour_balancing)
    sum_df = sum_df.copy()


def normalize_cols(df, suffix, ntype="std", skip_norm="_norm"):
    with warnings.catch_warnings():
        warnings.simplefilter('ignore', pd.errors.PerformanceWarning)
        if ntype == "minmax":
            for c in df.columns:
                if skip_norm is not None and skip_norm in c:
                    continue
                div = (df[c].max() - df[c].min())
                if div != 0:
                    df[c+suffix] = (df[c] - df[c].min()) / div
                else:
                    df[c+suffix] = 0
        elif ntype == "std":
            for c in df.columns:
                if skip_norm is not None and skip_norm in c:
                    continue
                div = df[c].std()
                if div != 0:
                    df[c+suffix] = (df[c] - df[c].mean()) / div
                else:
                    df[c+suffix] = 0
        else: raise ValueError(f"unk {ntype=}")
    df = df.copy()



def add_lags(df, num_lags=5, do_date=False, fillna="bfill", only_do_col=None):
    init_cols = df.columns
    do_cols = [c for c in init_cols if "is_attack" not in c]
    if not do_date:
        do_cols = [c for c in do_cols if "date_exog_" not in c]
    if only_do_col is not None:
        do_cols = [only_do_col]
    new_cols=[]
    with warnings.catch_warnings():
        warnings.simplefilter('ignore', pd.errors.PerformanceWarning)
        for l in range(1, num_lags+1):
            _new_cols = [(c+"_lag_"+str(l), c) for c in do_cols]
            df[[c[0] for c in _new_cols]] = df[[c for c in do_cols]].shift(l)
            new_cols.extend(_new_cols)
        #df = df.copy()
        for new_col, c in new_cols:
            if fillna == "mean":
                df[new_col].fillna(df[c].mean(), inplace=True)
            elif fillna == "bfill":
                df[new_col].fillna(method="bfill", inplace=True)
        df.rename(columns={c: c+"_lag_0" for c in init_cols}, inplace=True)
    return df

def add_lags2(df, num_lags=5, do_date=False, fillna="bfill"):
    init_cols = df.columns
    do_cols = [c for c in init_cols if "is_attack" not in c]
    if not do_date:
        do_cols = [c for c in do_cols if "date_exog_" not in c]
    new_cols=[]
    with warnings.catch_warnings():
        warnings.simplefilter('ignore', pd.errors.PerformanceWarning)
        for l in range(1, num_lags+1):
            for c in do_cols:
                new_col=c+"_lag_"+str(l)
                new_cols.append((new_col, c))
                df[new_col] = df[c].shift(l)
        #df = df.copy()
        for new_col, c in new_cols:
            if fillna == "mean":
                df[new_col].fillna(df[c].mean(), inplace=True)
            elif fillna == "bfill":
                df[new_col].fillna(method="bfill", inplace=True)
        df.rename(columns={c: c+"_lag_0" for c in init_cols}, inplace=True)
    return df

def add_all_lags(cpo_df_interp, num_lags=5, do_date=False, fillna="bfill", only_do_col=None):
    for group, cp_df_interp in tqdm(cpo_df_interp.items(), desc="add_all_lags"):
        cp_df_interp_l = add_lags(cp_df_interp, num_lags=num_lags, do_date=do_date, fillna=fillna, only_do_col=only_do_col)
        yield group, cp_df_interp_l



def get_grid_pp(DIR):
    grid_dir = os.path.join(DIR,"controller-export/power-grid/")
    grid_files = [f for f in os.listdir(grid_dir) if os.path.isfile(os.path.join(grid_dir, f))]
    
    for f in grid_files[:10]:
        regex_grid = re.compile(r"^WALL\-[\-\dT\+]+__SIM\-([\-\dT\+]+)\.powerowl\.p\.gz$")
        #WALL-2023-12-27T15-50-38-049544+00-00__SIM-2023-11-03T02-28-11-789274+00-00.powerowl.p.gz
        result = regex_grid.search(f)
        if result:
            f_time = "".join(result.group(1).rsplit("-", 1))
            f_time = pd.to_datetime(f_time, format="%Y-%m-%dT%H-%M-%S-%f%z") #2023-11-23T19-05-29-585438+00-00

            ff = os.path.join(grid_dir, f)
            p_grid = load_single_feat(ff)
            pd.reset_option('^display.', silent=True)

            pg = PandaPowerGridModel()
            pg.from_primitive_dict(p_grid)
            
            pp=pg.to_external()
            #print(pp)
            return pp
    logger.error(f"no pp grid found in {grid_dir}")
    return None



def interpolate_sim_time(ts, target):
    target = target.timestamp()
    if target in ts.index:
        return ts[target]
    ts1 = ts.sort_index()
    b = (ts1.index > target).argmax() # index of first entry after target
    s = ts1.iloc[b-1:b+1]
    # Insert empty value at target time.
    s = s.reindex(list(s.index.values) + [target])
    return s.interpolate(method='index', limit_direction="both").loc[target]

def get_grids_json(DIR, wall_sim_map, start_time=None): 
    grid_dir = os.path.join(DIR,"power_grid_exports/")
    grid_files = [f for f in os.listdir(grid_dir) if os.path.isfile(os.path.join(grid_dir, f))]
    
    pps=[]
    for f in tqdm(grid_files[:], desc="get_grids_json"):
        #power_grid_2024-02-03T19-04-14-008259+00-00.json
        regex_grid = re.compile(r"^power_grid_([\-\dT\+]+)\.json$")
        result = regex_grid.search(f)
        if result:
            f_time = "".join(result.group(1).rsplit("-", 1))
            f_time = pd.to_datetime(f_time, format="%Y-%m-%dT%H-%M-%S-%f%z") #2023-11-23T19-05-29-585438+00-00
            sim_time=interpolate_sim_time(wall_sim_map, pd.to_datetime(f_time))
            if start_time is not None:
                if np.isnan(sim_time):
                    continue
                s_t = datetime.utcfromtimestamp(sim_time)
                #print(f"{start_time} {s_t}")
                if s_t < start_time.tz_localize(None):
                    continue

            ff = os.path.join(grid_dir, f)
            
            #p_grid = load_single_feat(ff)
            with open(ff, 'r') as pg_f:
                p_grid = json.load(pg_f)
            
            p_grid["simtime"] = sim_time
            pd.reset_option('^display.', silent=True)
            pps.append(p_grid)
            #yield p_grid
    return pps


def get_elems_attatched_to_bus(pp, group_n):
    ret=[]
    lines=pp["line"]
    try:
        lines=lines[lines["from_bus"] < lines["to_bus"]]
        if (lines["to_bus"] == group_n).any():
            line = "line."+str(lines[lines["to_bus"] == group_n].sort_values(by="from_bus", ascending=True).index[0])
            ret.append(line)
    except Exception as e:
        logger.error(group_n)
        logger.error(pp["line"])
        logger.error(lines)
        raise e

    elems = ["load", "sgen", "storage"] #, "storage" only for non CP storages...; loads not included in default wattson
    for e in elems:
        try:
            el=pp[e]
            if e == "storage":
                el = el[el["type"] == "CP"]
                # print(el)
                # exit()
            if (el["bus"] == group_n).any():
                for el_idx in el[el["bus"] == group_n].index:
                    el = e+"."+str(el_idx)
                    ret.append(el)
        except Exception as e:
            logger.error(group_n)
            logger.error(el)
            raise e
    return ret


def get_ts_sim_map(DIR):
    grid_dir = os.path.join(DIR,"controller-export/power-grid/")
    grid_files = [f for f in os.listdir(grid_dir) if os.path.isfile(os.path.join(grid_dir, f))]
    
    lst=[]
    ind=[]
    for f in grid_files[:]:
        regex_grid = re.compile(r"^WALL\-([\-\dT\+]+)__SIM\-([\-\dT\+]+)\.powerowl\.p\.gz$")
        #WALL-2023-12-27T15-50-38-049544+00-00__SIM-2023-11-03T02-28-11-789274+00-00.powerowl.p.gz
        result = regex_grid.search(f)
        if result:
            wall_time = "".join(result.group(1).rsplit("-", 1))
            wall_time = pd.to_datetime(wall_time, format="%Y-%m-%dT%H-%M-%S-%f%z").timestamp() #2023-11-23T19-05-29-585438+00-00

            sim_time = "".join(result.group(2).rsplit("-", 1))
            sim_time = pd.to_datetime(sim_time, format="%Y-%m-%dT%H-%M-%S-%f%z").timestamp() #2023-11-23T19-05-29-585438+00-00

            lst.append(sim_time)
            ind.append(wall_time)

            #print(pp)
    x = pd.Series(lst, index = ind) 
    return x

def get_grid_measurements_from_export(DIR, pp, bus_ns=[], target_elem="load", target_var=".MEASUREMENT.active_power", start_time=None, time_delta_m=60):
    plot_l_dict=defaultdict(lambda:list())
    if not bus_ns:
        if "bus" in pp[target_elem]:
            bus_ns = sorted(set(i for i in pp[target_elem]["bus"].values))
        else:
            bus_ns = sorted(set(i for i in pp["load"]["bus"].values))
        logger.debug(f"no bus list provided, using all: {bus_ns}")

    wall_sim_map=get_ts_sim_map(DIR).sort_index()
    pps=get_grids_json(DIR, wall_sim_map, start_time=start_time)
    for bus_n in bus_ns:

        elems = get_elems_attatched_to_bus(pp, bus_n)
        if target_elem == "bus":
            elems = ["bus."+str(bus_n)]
        # print(elems)
        for elem in [e for e in elems if target_elem in e]:
            # print(elem)
            # exit()

            plot_t=[]
            #plot_b=[]
            plot_l=[]
            for grid in pps:
                if not np.isnan(grid["simtime"]):
                    plot_t.append(datetime.utcfromtimestamp(grid["simtime"]) + pd.Timedelta(minutes=time_delta_m))
                    #plot_b.append(grid["values"]["bus."+str(bus_n)".MEASUREMENT.active_power"])
                    plot_l.append(grid["values"][elem+target_var])
                else:
                    logger.debug(f"nan simtime for {grid['timestamp']}")
            #plot_b_s=pd.Series(plot_b, index = plot_t).sort_index()
            plot_l_s=pd.Series(plot_l, index = plot_t, name="grid_expo_"+elem+target_var).sort_index()
            plot_l_dict[bus_n].append(plot_l_s)
    return plot_l_dict

    
def save_feats(feat, OutDataDIR, t_name):
    Path(OutDataDIR+"/feats/"+t_name+"/").mkdir(parents=True, exist_ok=True)
    
    groups=[k for k in feat.keys()]
    for group in tqdm(groups, desc="save_feats", disable=(len(groups) == 1)):
        fdf_dict = feat[group]
        logger.debug(fdf_dict["cp_g_df"].columns)
        with gzip.open(OutDataDIR+"/feats/"+t_name+"/"+group+".gz", "wb") as f:
            pickle.dump(fdf_dict, f)
        del feat[group]
        gc.collect()

def load_feats_len(OutDataDIR, t_name):
    x_path = OutDataDIR+"/feats/"+t_name+"/"
    onlyfiles = [f for f in os.listdir(x_path) if os.path.isfile(os.path.join(x_path, f))]
    return len(onlyfiles)

# def load_feats2(OutDataDIR, t_name):
#     x_path = OutDataDIR+"/feats/"+t_name+"/"
#     onlyfiles = [f for f in os.listdir(x_path) if os.path.isfile(os.path.join(x_path, f))]
#     for fi in onlyfiles:
#         fdf_dict = load_single_feat(x_path+fi)
#         yield (fi.replace(".gz", ""), fdf_dict)




def load_feats(OutDataDIR, t_name):
    for fi, fi_p in iter_feats(OutDataDIR, t_name):
        fdf_dict = load_single_feat(fi_p)
        yield (fi, fdf_dict)

def iter_feats(OutDataDIR, t_name):
    x_path = OutDataDIR+"/feats/"+t_name+"/"
    onlyfiles = [f for f in os.listdir(x_path) if os.path.isfile(os.path.join(x_path, f))]
    for fi in onlyfiles:
        yield (fi.replace(".gz", ""), x_path+fi)

def load_single_feat(file_p):
    with gzip.open(file_p, "rb") as f:
        fdf_dict = pickle.load(f)
        return fdf_dict
    
def load_feats_CPg(OutDataDIR, t_name, cp_g):
    fi = OutDataDIR+"/feats/"+t_name+"/"+cp_g+".gz"
    if not os.path.isfile(os.path.join(fi)):
        logger.error(f"file {fi} not found")
        return None
                 
    with gzip.open(fi, "rb") as f:
        fdf_dict = pickle.load(f)
    return fdf_dict

def prune_feats(df_dict, feat_cases=None):
    if feat_cases is None or "all" in feat_cases:
        return
    
    feat_cols = df_dict["feat_cols"]


    add_expo = [feat_case for feat_case in feat_cases if "add_grid_load_expo_" in feat_case]
    if add_expo:
        if len(add_expo) > 1:
            logger.warning(f"multiple add_grid_load_expo_ sets provided: {add_expo}; using first")
        add_expo = add_expo[0]
        sub_case="static"
        if "expo_rnd" in add_expo:
            sub_case="rnd"
        add_expo_fac_min = int(add_expo.split("_")[-1])
        add_expo_fac_max = min(add_expo_fac_min+10, 100)
        #print(add_expo, add_expo_fac_min, add_expo_fac_max)
        expo_cols = [c for c in df_dict["cp_g_df"].columns if "grid_expo_" in c and sub_case+"_fac."+str(add_expo_fac_min) in c]
        #print(expo_cols)
        #print(feat_cols)
        feat_cols = feat_cols + expo_cols
        #print(feat_cols)
        #exit()

    if "add_bus_relations" in feat_cases:
        for add_c in [c for c in df_dict["cp_g_df"].columns if ".bus_and_sgen." in c or ".no_ev_load." in c]:
            feat_cols.append(add_c)

    for feat_case in feat_cases:
        if feat_case == "no_cps":
            feat_cols = [f for f in feat_cols if "CP_" not in f]
        elif feat_case == "no_cps_but_speed":
            feat_cols = [f for f in feat_cols if "CP_" not in f or "charge_speed" in f] 

        elif feat_case == "no_grid":
            #feat_cols = [f for f in feat_cols if "bus." not in f]
            feat_cols = [f for f in feat_cols if "grid_est_" not in f] 
            feat_cols = [f for f in feat_cols if "grid_meas_" not in f]
        elif feat_case == "no_grid_storage":
            #feat_cols = [f for f in feat_cols if "bus." not in f]
            feat_cols = [f for f in feat_cols if not ("grid_est_" in f and ".storage." in f)] 
            feat_cols = [f for f in feat_cols if not ("grid_meas_" in f and ".storage." in f)]
        elif feat_case == "no_grid_line":
            #feat_cols = [f for f in feat_cols if "bus." not in f]
            feat_cols = [f for f in feat_cols if not ("grid_est_" in f and ".line." in f)] 
            feat_cols = [f for f in feat_cols if not ("grid_meas_" in f and ".line." in f)]
        elif feat_case == "no_grid_sgen":
            #feat_cols = [f for f in feat_cols if "bus." not in f]
            feat_cols = [f for f in feat_cols if not ("grid_est_" in f and ".sgen." in f)] 
            feat_cols = [f for f in feat_cols if not ("grid_meas_" in f and ".sgen." in f)]
        elif feat_case == "no_grid_bus_rel_to_pred":
            pred_col = df_dict["pred_col"]
            feat_cols = [f for f in feat_cols if not ("grid_meas_" in f and ".bus." in f and "_relation_to_"+pred_col.replace("_lag_0", "") in f)]

        elif feat_case == "no_est":
            feat_cols = [f for f in feat_cols if "grid_est_" not in f] 
        elif feat_case == "no_meas":
            feat_cols = [f for f in feat_cols if "grid_meas_" not in f]

        elif feat_case == "no_time_diff":
            feat_cols = [f for f in feat_cols if "time_diff_" not in f]
        
        elif feat_case == "only_pred_lag":
            pred_col = df_dict["pred_col"]
            feat_cols = [f for f in feat_cols if "_lag_0" in f or f.startswith(pred_col.replace("_lag_0", "_lag_"))]
            
        elif feat_case == "only_norm":
            feat_cols = [f for f in feat_cols if "_norm" in f]
        elif feat_case == "only_norm_but_pred":
            pred_col = df_dict["pred_col"]
            feat_cols = [f for f in feat_cols if "_norm" in f or f.startswith(pred_col.replace("_lag_0", "_lag_"))]
        elif feat_case == "no_norm":
            feat_cols = [f for f in feat_cols if "_norm" not in f]

        elif feat_case == "no_norm1":
            feat_cols = [f for f in feat_cols if "_cp_g_norm" not in f and "_cp_norm" not in f]
        elif feat_case == "no_norm2":
            feat_cols = [f for f in feat_cols if "_group_norm" not in f]
            
        elif feat_case == "no_date1":
            feat_cols = [f for f in feat_cols if "date_exog_cp_g_" not in f and "date_exog_cp_" not in f]
        elif feat_case == "no_date2":
            feat_cols = [f for f in feat_cols if "date_exog_group_" not in f]

        elif feat_case == "no_hour_date":
            feat_cols = [f for f in feat_cols if not ("date_exog_" in f and "hour" in f and "discretize" not in f)]
        elif feat_case == "no_hour_only_date":
            feat_cols = [f for f in feat_cols if not ("date_exog_" in f and "discretize_hour_only" in f)]
        elif feat_case == "no_hour_day_date":
            feat_cols = [f for f in feat_cols if not ("date_exog_" in f and "discretize_hour_day" in f)]
        elif feat_case == "no_day_of_week_date":
            feat_cols = [f for f in feat_cols if not ("date_exog_" in f and "dayofweek" in f)]
        elif feat_case == "no_day_is_work_date":
            feat_cols = [f for f in feat_cols if not ("date_exog_" in f and "discretize_day_is_work" in f)]
        elif feat_case == "no_hour_balancing_date":
            feat_cols = [f for f in feat_cols if not ("date_exog_" in f and "discretize_hour_balancing" in f)]



    df_dict["feat_cols"] = feat_cols
