
import ast
from collections import defaultdict
from concurrent.futures import ProcessPoolExecutor, ThreadPoolExecutor
from copy import deepcopy
from datetime import datetime
import gc
import gzip
import hashlib
import json
import multiprocessing
import os
from pathlib import Path
import pickle
import sys
import threading
import time
import warnings
import pandas as pd
from sklearn.covariance import EllipticEnvelope
from sklearn.discriminant_analysis import StandardScaler
from sklearn.ensemble import IsolationForest
from sklearn.exceptions import ConvergenceWarning, UndefinedMetricWarning
from sklearn.metrics  import accuracy_score
from sklearn.calibration import cross_val_predict
from sklearn.metrics import confusion_matrix, mean_squared_error, precision_recall_fscore_support
from sklearn.model_selection import ParameterGrid
from sklearn.neighbors import LocalOutlierFactor
from sklearn.pipeline import make_pipeline
from sklearn.svm import OneClassSVM
from tqdm import tqdm


from joblib import parallel_backend
from threadpoolctl import threadpool_limits

import re

import logging

from features_aux import add_lags, get_date_exog, load_feats, load_feats_CPg, load_feats_len, normalize_cols, add_all_lags
from regression import get_cp_group_eval_dicts

logger = logging.getLogger("WATTSON_EV_IDS.FeaturesClf")

m_dict = multiprocessing.Manager()
CLF_CACHE=m_dict.dict()
#CLF_CACHE=dict()
with open("ids.conf", 'r') as f:
    conf = json.load(f)
    NUM_THREADS = conf["NUM_THREADS"]


feat_warn1=0
feat_warn2=0
def save_clf_data(feat, out_d, out_f):
    #cp_g_dict = {'reg': 'LinearSVR', 'shifts': '5', 'file': 'data/year/predictions/DSO.all/CP_1_LinearSVR_5.csv.gz', 'features': ['all']}
    Path(out_d).mkdir(parents=True, exist_ok=True)
    
    logger.debug(feat.columns)
    feat.to_csv(out_f)
    del feat
    gc.collect()

def add_grid_meas_relations(clf_df, cp_g_df, pred_col, speed_preds):
    get_rel_expo_cols = [c for c in cp_g_df.columns if ".load." in c and "_norm" not in c and "grid_expo_" in c and ".active_power" in c and "_relation_to_" not in c] # and "10_" in c
    get_rel_expo_cols_base = [c for c in cp_g_df.columns if ".load." in c and "_norm" not in c and "grid_expo_" in c and ".active_power_lag_0" in c and "_relation_to_" not in c]#.
    get_rel_expo_cols_map=defaultdict(lambda:list())
    
    if len(get_rel_expo_cols_base) > 1: #'grid_expo_0.0.load.14.active_power_lag_0', 'grid_expo_0.0.load.6.active_power_lag_0', 
        load_ns = [c.split(".")[3] for c in get_rel_expo_cols_base]
        for rel_expo in get_rel_expo_cols:
            for load_n in load_ns:
                if ".load."+load_n+"." in rel_expo:
                    rel_expo_base = rel_expo.replace(".load."+load_n+".",".load.total.")
                    get_rel_expo_cols_map[rel_expo_base].append(rel_expo)
        for rel_expo_base,rel_expos in get_rel_expo_cols_map.items():
            cp_g_df[rel_expo_base]=0
            get_rel_expo_cols.append(rel_expo_base)
            for rel_expo in rel_expos:
                cp_g_df[rel_expo_base]+=cp_g_df[rel_expo]
                

    # print([c for c in clf_df.columns])
    # print([c for c in cp_g_df.columns if "_norm" not in c and "grid_" in c and "sgen" in c])
    # print([c for c in cp_g_df.columns if "_norm" not in c and "grid_meas_" in c and "sgen" in c and ".active_power_lag_0" in c and "_relation_to_" not in c])
    # print([c for c in cp_g_df.columns if "_norm" not in c and "grid_meas_" in c and "bus" in c and ".active_power_lag_0" in c and "_relation_to_" not in c])
    # print(get_rel_expo_cols)
    #print(cp_g_df)

    new_cols=[]

    #max ev power = bus + sgen - load(s)
    bus_cols = [c for c in cp_g_df.columns if ".bus." in c and ".active_power" in c and "_norm" not in c and "_lag_0" in c and "_relation_to_" not in c]
    bus_col = bus_cols[0]
    sgen_cols = [c for c in cp_g_df.columns if ".sgen." in c and ".active_power" in c and "_norm" not in c and "_lag_0" in c and "_relation_to_" not in c]

    name_b_total = bus_col.replace("active_power","total_power")
    clf_df[name_b_total] = cp_g_df[bus_col]
    for sgen_col in sgen_cols:
        cp_g_df_sgen_col = cp_g_df[sgen_col]
        clf_df[name_b_total] += cp_g_df_sgen_col #grid_meas_585.10030.bus.11.active_power_lag_0
    new_cols.append(name_b_total)
    
    non_ev_power = bus_col.replace("active_power","non_ev_power")
    clf_df[non_ev_power] = clf_df[name_b_total] - cp_g_df[pred_col] # (bus + sgen) - ev load = loads
    new_cols.append(non_ev_power)
    
    non_ev_power_relation_to_bus = bus_col.replace("active_power","non_ev_power_relation_to_bus")
    clf_df[non_ev_power_relation_to_bus] = clf_df[non_ev_power] / clf_df[name_b_total]
    new_cols.append(non_ev_power_relation_to_bus)

    ev_power_relation_to_bus = bus_col.replace("active_power","only_ev_power_relation_to_bus")
    clf_df[ev_power_relation_to_bus] = cp_g_df[pred_col] / clf_df[name_b_total]
    new_cols.append(ev_power_relation_to_bus)
    
    for rec in get_rel_expo_cols:
        name = rec.replace("_lag_0","")+"_max_ev_power"+"_lag_0"
        clf_df[name] = clf_df[name_b_total] - cp_g_df[rec]
        new_cols.append(name)

        name2 = rec.replace("_lag_0","")+"_max_ev_power_relation_to_bus"+"_lag_0"
        clf_df[name2] = clf_df[name] / clf_df[name_b_total]
        new_cols.append(name2)

        name2 = rec.replace("_lag_0","")+"_non_ev_power_relation_to_expo"+"_lag_0"
        clf_df[name2] = cp_g_df[rec] / clf_df[non_ev_power]
        new_cols.append(name2)


        name = rec.replace("_lag_0","")+"_relation_to_"+pred_col.replace("_lag_0","")+"_lag_0" #relation: grid load measurement to received charge speed
        clf_df.loc[(cp_g_df[rec] != 0), name] = cp_g_df[pred_col] / cp_g_df[rec]
        clf_df.loc[(cp_g_df[rec] == 0), name] = 0
        new_cols.append(name)

        for speed_pred in speed_preds:
            name = rec.replace("_lag_0","")+"_relation_to_"+speed_pred.replace("_lag_0","")+"_lag_0" #relation: grid load measurement to predicted charge speeds
            clf_df.loc[(cp_g_df[rec] != 0), name] = clf_df[speed_pred] / cp_g_df[rec]
            clf_df.loc[(cp_g_df[rec] == 0), name] = 0
            new_cols.append(name)

        if len(bus_cols) >= 1:
            if len(bus_cols) > 1:
                logger.warning(f"len(bus_cols) > 1: {bus_cols}")
            bus_col = bus_cols[0]
            name = bus_col.replace("_lag_0","")+"_relation_to_"+rec.replace("_lag_0","")+"_lag_0" #relation: grid load measurement to bus measurement
            clf_df.loc[(cp_g_df[rec] != 0), name] = cp_g_df[bus_col] / cp_g_df[rec]
            clf_df.loc[(cp_g_df[rec] == 0), name] = 0
            new_cols.append(name)
        else:
            logger.warning(f"bus_cols empty: {[c for c in cp_g_df.columns if 'bus' in c]}")


        clf_df = clf_df.copy()

    # print(clf_df)
    # #print(new_cols)
    # exit()

    if False:
        new_cols_df = clf_df[new_cols]
        normalize_cols(new_cols_df, "_prune_norm")
        for c in new_cols_df.columns:
            if "_norm" in c and c not in cp_g_df.columns:
                new_c = c.replace("_lag_0_prune_norm", "_prune_norm_lag_0")
                clf_df[new_c] = new_cols_df[c]
                clf_df = clf_df.copy()
    return clf_df


def get_clf_feat_dfs(cp_g_pred, cp_g_df, pred_col):
    with warnings.catch_warnings():
        warnings.simplefilter('ignore', pd.errors.PerformanceWarning)
        pred_cols = [c for c in cp_g_pred.columns if "prediction_" in c]
        clf_df = cp_g_pred[pred_cols].copy()
        clf_df["prediction_mean"] = cp_g_pred[[c for c in cp_g_pred.columns if "prediction_" in c]].mean(axis=1)
        clf_df["prediction_min"] = cp_g_pred[[c for c in cp_g_pred.columns if "prediction_" in c and "_mean" not in c]].min(axis=1)
        clf_df["prediction_max"] = cp_g_pred[[c for c in cp_g_pred.columns if "prediction_" in c and "_mean" not in c]].max(axis=1)
        if len(pred_cols) >= 2:
            clf_df["prediction_mean_0_1"] = cp_g_pred[[c for c in cp_g_pred.columns if "prediction_0" in c or "prediction_1" in c]].mean(axis=1)
        if len(pred_cols) >= 3:
            clf_df["prediction_mean_0_1_2"] = cp_g_pred[[c for c in cp_g_pred.columns if "prediction_0" in c or "prediction_1" in c or "prediction_2" in c]].mean(axis=1)
            clf_df["prediction_mean_1_2"] = cp_g_pred[[c for c in cp_g_pred.columns if "prediction_1" in c or "prediction_2" in c]].mean(axis=1)
        if len(pred_cols) >= 4:
            clf_df["prediction_mean_2_3"] = cp_g_pred[[c for c in cp_g_pred.columns if "prediction_2" in c or "prediction_3" in c]].mean(axis=1)
        if len(pred_cols) >= 5:
            clf_df["prediction_mean_3_4"] = cp_g_pred[[c for c in cp_g_pred.columns if "prediction_3" in c or "prediction_4" in c]].mean(axis=1)
        # print(clf_df)
        # exit()
        pred_col_df = cp_g_df[pred_col].loc[clf_df.index]
        # if len(clf_df) != len(pred_col_df):
        #     print(clf_df)
        #     print(pred_col_df)
        for c in clf_df.columns:
            if "_mean" in c or "_min" in c or "_max" in c:
                clf_df[c+"_diff"] = clf_df[c] - pred_col_df
                clf_df[c+"_diff_abs"] = (clf_df[c] - pred_col_df).abs()
                #clf_df[c+"_diff_rmse"] = mean_squared_error(clf_df[c], pred_col_df, squared=False)
                clf_df = clf_df.copy()
            
        clf_df = add_grid_meas_relations(clf_df, cp_g_df, pred_col, ["prediction_mean", "prediction_min", "prediction_max"])    #pred_cols+ 
        # print(clf_df)
        # for c in clf_df.columns:
        #     print(c)
        # exit()

        normalize_cols(clf_df, "_clf_norm")
        #print(clf_df)
        return clf_df


def get_clf_feat_file_dicts(OutDataDIR):
    regex = re.compile(r"^(CP_\d+)_(\w+)_([\d\w]+).csv.gz$")

    ret_d=defaultdict(lambda:defaultdict(lambda:list()))
    DIR=OutDataDIR+"/clf_feats/"
    sub_folders = [name for name in os.listdir(DIR) if os.path.isdir(os.path.join(DIR, name))]
    for actor in sub_folders:
        if ".bak" in actor:
            logger.warning(f'skiping {OutDataDIR+"/clf_feats/"+actor}')
            continue
        if len(actor.split(".")) > 1:
            features = actor.split(".")[1:]
        else:
            features=["all"]
            logger.warning(f"assuming all features for {actor}")
        x_path=OutDataDIR+"/clf_feats/"+actor
        onlyfiles = [f for f in os.listdir(x_path) if os.path.isfile(os.path.join(x_path, f))]
        for f in onlyfiles:
            result = regex.search(f)
            if result:
                result.group(1)
                ret_d[actor][result.group(1)].append({"reg":result.group(2),"shifts":result.group(3),"file":os.path.join(x_path, f),"features":features})
            else:
                logger.error(f"unk file {f}")
                continue
    return ret_d


def get_clf_is_atk_dfs(OutDataDIR, group):
    is_atk_cols = ['meter_value_sampled_value_lag_0', 'custom_data_meter_no_atk_lag_0', 
                    'custom_data_energy_interval_lag_0', 'custom_data_original_energy_interval_lag_0',
                    'custom_data_average_power_lag_0', 'custom_data_original_average_power_lag_0', 
                    'meter_diff_lag_0', 'custom_data_meter_diff_lag_0',
                    'charge_speed_lag_0', 'custom_data_charge_speed_lag_0','is_attack_lag_0']

    is_atk_dfs = dict()

    cpos = [f for f in os.listdir(OutDataDIR+"/feats/") if os.path.isdir(os.path.join(OutDataDIR+"/feats/", f)) and f.startswith("CPO_")]
    for actor_prefix in sorted(cpos):
        act_feat = load_feats_CPg(OutDataDIR, actor_prefix, group) #generator
        df = act_feat["cp_g_df"]
        is_atk_dfs[actor_prefix] = df[is_atk_cols].copy()

    return is_atk_dfs


def _interp(s, targets, col="charge_speed_lag_0", method="time"):
    o_idx = list(s.index.values)
    n_idx = list(targets.values) # if v not in o_idx
    f_idx=set(o_idx + n_idx)
    s2 = s.reindex(pd.to_datetime(list(f_idx)))
    if col is None:
        return s2.interpolate(method=method).loc[targets]
    else:
        return s2[col].interpolate(method=method).loc[targets]

def _interp1(ts, target, col="charge_speed_lag_0"):
    if target in ts.index:
        return ts[col].loc[target]
    if target < ts.index.min() or target > ts.index.max():  #before or after table
        return 0.0
    ts1 = ts.sort_index()
    b = (ts1.index > target).argmax() # index of first entry after target
    s = ts1.iloc[b-1:b+1]
    if len(s)==0: #before or after table
        return 0.0
    # Insert empty value at target time.
    s = s.reindex(pd.to_datetime(list(s.index.values) + [target]))
    return s[col].interpolate(method='time').loc[target]
    
def get_clf_is_atk_grid(cp_group_grid_power, is_atk_dfs, target, col="charge_speed_lag_0", col_s=None):
    val = 0
    val_s = 0
    if col_s is None:
        col_s = "custom_data_" + col

    val += _interp(cp_group_grid_power, target, col=None)
    for actor, is_atk_df in is_atk_dfs.items():
        val_s += _interp(is_atk_df, target, col=col_s)

    return (val - val_s).rename(cp_group_grid_power.name+"_diff_to_"+col_s)

def get_clf_is_atk(is_atk_dfs, target, col="charge_speed_lag_0", col_s=None):
    val = 0
    val_s = 0
    if col_s is None:
        col_s = "custom_data_" + col
    for actor, is_atk_df in is_atk_dfs.items():
        val  +=  _interp(is_atk_df, target, col=col)
        if col != "is_attack_lag_0":
            val_s += _interp(is_atk_df, target, col=col_s)
    return (val - val_s).rename(col+"_diff")

def get_param_grid_len(clf_type):
    return len(ParameterGrid(get_param_grid_from_string(clf_type)))

def get_param_grid_from_string(clf_type, train_contamination=False): 
    if clf_type == "LocalOutlierFactor":
        if train_contamination == True:
            param_grid = {'algorithm': ['auto'], 'leaf_size': [30,20,40], 'metric': ['minkowski'], 'metric_params': [None], 'n_jobs': [NUM_THREADS], 'n_neighbors': [20,10,30], 'novelty': [True], 'p': [1,2], 'contamination': ['auto']}
        else:
            param_grid = {'algorithm': ['auto'], 'leaf_size': [30,20,40], 'metric': ['minkowski'], 'metric_params': [None], 'n_jobs': [NUM_THREADS], 'n_neighbors': [20,10,30], 'novelty': [True], 'p': [1,2], 'contamination': [1e-323]}
    elif clf_type == "OneClassSVM":
        param_grid = {'cache_size': [7000], 'kernel': ['rbf', 'sigmoid'], 'max_iter': [200], 'nu': [0.5,0.2,0.7], 'shrinking': [True], 'tol': [0.001], 'verbose': [False]}
        # param_grid1 = {'cache_size': [7000], 'coef0': [0.0, 0.5, 2.0], 'gamma': ['scale'], 'kernel': ['poly', 'sigmoid'], 'max_iter': [-1], 'nu': [0.5,0.2,0.7], 'shrinking': [True,False], 'tol': [0.001,0.01,0.0001], 'verbose': [False]}
        # param_grid2 = {'cache_size': [7000], 'degree': [3,6,9], 'gamma': ['scale'], 'kernel': ['poly'], 'max_iter': [-1], 'nu': [0.5,0.2,0.7], 'shrinking': [True,False], 'tol': [0.001,0.01,0.0001], 'verbose': [False]}
        # param_grid3 = {'cache_size': [7000], 'kernel': ['rbf', 'linear'], 'max_iter': [-1], 'nu': [0.5,0.2,0.7], 'shrinking': [True,False], 'tol': [0.001,0.01,0.0001], 'verbose': [False]}
        # param_grid = [param_grid1, param_grid2, param_grid3]
    elif clf_type == "EllipticEnvelope":
        # param_grid = {'assume_centered': [False,True], 'contamination': [1e-323], 'random_state': [12345], 'store_precision': [True], 'support_fraction': [None,0.1,0.5,0.8,0.9,0.99], "random_state": [12345]} #The covariance matrix of the support data is equal to 0, try to increase support_fractio
        if train_contamination == True:
            param_grid = {'assume_centered': [False,True], 'contamination': [0.1], 'random_state': [12345], 'store_precision': [True], 'support_fraction': [0.7,0.8,0.9,0.99]}
        else:
            param_grid = {'assume_centered': [False,True], 'contamination': [1e-323], 'random_state': [12345], 'store_precision': [True], 'support_fraction': [0.7,0.8,0.9,0.99]}
    elif clf_type == "IsolationForest":
        if train_contamination == True:
            param_grid = {'bootstrap': [False,True], 'contamination': ['auto'], 'max_features': [1.0,0.9], 'max_samples': ['auto',0.9,0.8], 'n_estimators': [50,100,200], 'n_jobs': [NUM_THREADS], 'random_state': [12345], 'verbose': [0], 'warm_start': [False]}
        else:
            param_grid = {'bootstrap': [False,True], 'contamination': [1e-323], 'max_features': [1.0,0.9], 'max_samples': ['auto',0.9,0.8], 'n_estimators': [50,100,200], 'n_jobs': [NUM_THREADS], 'random_state': [12345], 'verbose': [0], 'warm_start': [False]}
    else:
        logger.error(f"unk clf_type {clf_type}")
        param_grid = {}
    return param_grid
    
def get_clfs_from_string(clf_type, config=None, train_contamination=False): 
    param_grid = get_param_grid_from_string(clf_type, train_contamination=train_contamination)
    if config is None:
        pg = ParameterGrid(param_grid)
    else:
        pg = [config]
    for p in pg:
        if clf_type == "LocalOutlierFactor":
            yield LocalOutlierFactor(**p)
        elif clf_type == "OneClassSVM":
            yield OneClassSVM(**p)
        elif clf_type == "EllipticEnvelope":
            yield EllipticEnvelope(**p)
        elif clf_type == "IsolationForest":
            yield IsolationForest(**p)
        else:
            logger.error(f"unk clf_type {clf_type}")
            yield None



def get_clf_result_output_conc(best_d_ag, cp_g_dict, atk_cp_g_dict, train_df, atk_df, TRAIN_START_DATE, ATK_START_DATE, atk_labels_df, pbar=None, allow_diff_shifts=True):
    res_l=[]
    th_i=0
    # data_lock=threading.Lock()
    # data_lock2=threading.Lock()
    # pbar2=None
    # with ThreadPoolExecutor(NUM_THREADS) as pool:
    m = multiprocessing.Manager()
    data_lock=m.Lock()
    data_lock2=m.Lock()
    pbar2=pbar
    pbar=None
    with ProcessPoolExecutor(NUM_THREADS) as pool: #multiprocessing
        results=[]
        for clf, best_dict in best_d_ag.items():
            th_i+=1
            if th_i % 2 == 0:   
                with data_lock: 
                    results.append( pool.submit(get_clf_result_output, clf, best_dict, cp_g_dict, atk_cp_g_dict, train_df, atk_df, TRAIN_START_DATE, ATK_START_DATE, atk_labels_df, pbar, allow_diff_shifts, data_lock) )
            else:
                with data_lock2: 
                    results.append( pool.submit(get_clf_result_output, clf, best_dict, cp_g_dict, atk_cp_g_dict, train_df, atk_df, TRAIN_START_DATE, ATK_START_DATE, atk_labels_df, pbar, allow_diff_shifts, data_lock2) )
            time.sleep(0.05)

            if th_i % NUM_THREADS == 0 or th_i == 2 or th_i == 4:            
                for r in results:
                    timestamp1 = time.time()
                    clf_results = r.result()
                    timestamp2 = time.time()
                    logger.debug("r.result() took %.2f seconds" % (timestamp2 - timestamp1))
                    if pbar2 is not None:
                        pbar2.update(1)
                    if clf_results is not None:
                        res_l.append(clf_results)
                results=[]
                gc.collect()


        for r in results:
            clf_results = r.result()
            if pbar2 is not None:
                pbar2.update(1)
            if clf_results is not None:
                res_l.append(clf_results)

    return res_l

def do_get_clf_is_atk_conc(pool, atk_df, ATK_START_DATE, is_atk_dfs, cp_group_grid_power, pbar=None):
    return pool.submit(do_get_clf_is_atk_wrap, atk_df, ATK_START_DATE, is_atk_dfs, cp_group_grid_power, pbar=pbar)

def do_get_clf_is_atk_wrap(atk_df, ATK_START_DATE, is_atk_dfs, cp_group_grid_power, pbar=None):
    end_d = max([v.index.max() for v in is_atk_dfs.values()])
    atk_df = atk_df.loc[ATK_START_DATE:end_d] 
    # print(atk_df)
    # print([k for k in is_atk_dfs.keys()])
    # print(is_atk_dfs["CPO_0"])
    clf_is_atk = do_get_clf_is_atk(is_atk_dfs, atk_df, cp_group_grid_power)
    if pbar is not None:
        pbar.update(1)
    return clf_is_atk

def do_get_clf_is_atk(is_atk_dfs, atk_full_clf_feat, cp_group_grid_power):
    clf_is_atk1 = get_clf_is_atk(is_atk_dfs, atk_full_clf_feat.index)
    clf_is_atk2 = get_clf_is_atk(is_atk_dfs, atk_full_clf_feat.index, col="is_attack_lag_0")
    # print(cp_group_grid_power)
    # print(is_atk_dfs)
    clf_is_atk3 = get_clf_is_atk_grid(cp_group_grid_power, is_atk_dfs, atk_full_clf_feat.index)
    clf_is_atk4 = get_clf_is_atk_grid(cp_group_grid_power, is_atk_dfs, atk_full_clf_feat.index, col_s="charge_speed_lag_0")

    #clf_is_atk = clf_is_atk1.to_frame().join(clf_is_atk2)
    #print(clf_is_atk)
    clf_is_atk = pd.concat([clf_is_atk1,clf_is_atk2,clf_is_atk3,clf_is_atk4 ],axis=1)
    #print(clf_is_atk)
    #exit()
    clf_is_atk = clf_is_atk.loc[atk_full_clf_feat.index]
    clf_is_atk["is_attack1"] = 1
    clf_is_atk["is_attack2"] = 1
    clf_is_atk["is_attack3"] = 1
    clf_is_atk["is_attack4"] = 1
    clf_is_atk.loc[clf_is_atk[clf_is_atk1.name] != 0, "is_attack1"] = -1 #"charge_speed_lag_0_diff"
    clf_is_atk.loc[clf_is_atk[clf_is_atk2.name] != 0, "is_attack2"] = -1 #"is_attack_lag_0_diff"
    clf_is_atk.loc[clf_is_atk[clf_is_atk3.name] != 0, "is_attack3"] = -1 #grid vs should speed
    clf_is_atk.loc[clf_is_atk[clf_is_atk4.name] != 0, "is_attack4"] = -1 #grid vs reported speed
    # print(clf_is_atk)
    # print(clf_is_atk['is_attack1'].equals(clf_is_atk['is_attack2']))

    # clf_is_atk["is_attack_d"] = 1
    # clf_is_atk.loc[clf_is_atk["is_attack1"] != clf_is_atk["is_attack2"], "is_attack_d"] = -1

    ##print(clf_is_atk[clf_is_atk["is_attack_d"] == 1])
    # print(clf_is_atk[clf_is_atk["is_attack_d"] == -1]) #TODO: why the diff

    # print(atk_full_clf_feat)
    # print(clf_is_atk)
    # print(y_p)
    return clf_is_atk


def get_atk_full_clf_feat(atk_clf_feat_file, clf_f, atk_df, atk_labels_df, ATK_START_DATE, allow_diff_shifts=True):
    atk_cp_g_clf_feat = pd.read_csv(atk_clf_feat_file, index_col=0, parse_dates=[0])
    atk_df = atk_df.loc[atk_cp_g_clf_feat.index.min():atk_cp_g_clf_feat.index.max()] 

    if clf_f is None:
        clf_f=list(atk_cp_g_clf_feat.columns)+list(atk_df.columns)
    if allow_diff_shifts:
        clf_f_bak = clf_f
        clf_f_l = len(clf_f)
        atk_full_clf_feat_columns=list(atk_cp_g_clf_feat.columns)+list(atk_df.columns)
        clf_f = [c for c in clf_f if c in atk_full_clf_feat_columns]
        if len(clf_f) != clf_f_l:
            # logger.info(f"train and test used diff features! {clf_f_bak=}")
            # logger.info(f"train and test used diff features! {clf_f=}")
            # logger.info(f"train and test used diff features! {[f for f in clf_f_bak if f not in clf_f]}")
            # exit()
            
            global feat_warn1
            if feat_warn1 <=1:
                feat_warn1+=1
                logger.warning(f"train and test used diff features! {allow_diff_shifts=}")

    atk_cp_g_clf_feat = atk_cp_g_clf_feat[[c for c in clf_f if c in atk_cp_g_clf_feat.columns]]
    atk_df = atk_df[[c for c in clf_f if c in atk_df.columns]]

    #end_d = max([v.index.max() for v in is_atk_dfs.values()])
    end_d = atk_labels_df.index.max()
    atk_full_clf_feat = pd.concat([atk_cp_g_clf_feat, atk_df], axis=1).loc[ATK_START_DATE:end_d][clf_f]
    del atk_cp_g_clf_feat
    del atk_df
    gc.collect()
    return atk_full_clf_feat, clf_f

def get_clf_result_output(clf, best_dict, cp_g_dict, atk_cp_g_dict, train_df, atk_df, TRAIN_START_DATE, ATK_START_DATE, atk_labels_df, pbar=None, allow_diff_shifts=True, data_lock=None):
    #print(clf,best_dict)

    if data_lock is not None:
        data_lock.acquire()

    clf_feat_file = get_clf_feat_for_conf(cp_g_dict, best_dict)
    atk_clf_feat_file = get_clf_feat_for_conf(atk_cp_g_dict, best_dict, allow_diff_shifts=allow_diff_shifts)
    # print(cp_g_dict)
    # print(atk_cp_g_dict)
    # print(best_dict)
    if atk_clf_feat_file is None:
        logger.warning(f"no atk_clf_feat_file found for {clf=} in {atk_cp_g_dict=}")
        if pbar is not None:
            pbar.update(1)
        if data_lock is not None:
            data_lock.release()
        return None
    #print(clf_feat_file)
    #print(atk_clf_feat_file)
    clf_n = clf.split(".")[0]
    clf_c = best_dict["conf"]
    clf_c = ast.literal_eval(clf_c)
    _clf_f = best_dict["feat"]
    clf_o = best_dict["offset"]
    clf_r = best_dict["reg"]
    #print(f"TEST: {clf=} {clf_c=} {_clf_f=} {clf_o=} {clf_r=} {[k for k in best_dict.keys()]}")

    clf_only = clf.replace("."+clf_r, "")
    clf_features_string = clf_only.split(".")[1:]
    do_scale_pipeline = False
    if "only_norm" in clf_features_string:
        do_scale_pipeline = True

    #print(f"TEST: {clf_only=} {clf_r=} {clf_features_string=} {do_scale_pipeline=}")

    #print(atk_clf_feat_file)
    atk_full_clf_feat, clf_f = get_atk_full_clf_feat(atk_clf_feat_file, _clf_f, atk_df, atk_labels_df, ATK_START_DATE, allow_diff_shifts)


    cp_g_clf_feat = pd.read_csv(clf_feat_file, index_col=0, parse_dates=[0])
    train_df = train_df.loc[cp_g_clf_feat.index.min():cp_g_clf_feat.index.max()]

    cp_g_clf_feat = cp_g_clf_feat[[c for c in clf_f if c in cp_g_clf_feat.columns]]
    train_df = train_df[[c for c in clf_f if c in train_df.columns]]


    full_clf_feat = pd.concat([cp_g_clf_feat, train_df], axis=1).loc[TRAIN_START_DATE:ATK_START_DATE][clf_f]
    del cp_g_clf_feat
    del train_df
    gc.collect()


    # full_clf_feat = full_clf_feat[clf_f]
    # atk_full_clf_feat = atk_full_clf_feat[clf_f]
    # gc.collect()

    if data_lock is not None:
        data_lock.release()

    # print(clf_n, type(clf_n))
    # print(clf_c, type(clf_c))
    # print(clf_f, type(clf_f))
    # print(clf_o, type(clf_o))
    # exit()

    #is_atk_pred = do_clf(clf_n, clf_c, full_clf_feat, atk_full_clf_feat, offset_=0, pbar=pbar)
    try:
        y_p = do_clf(clf_n, clf_c, full_clf_feat, atk_full_clf_feat, pbar=pbar, overwrite_jobs=2, scale_pipeline=do_scale_pipeline)
    except Exception as e: #TODO: handle "ValueError: The covariance matrix of the support data is equal to 0, try to increase support_fraction"
        clf_n=clf_n+"_"+str(e)
        y_p = [1 for x in atk_full_clf_feat.values]

    #is_atk_pred_0 = [1 if p >= 0 else -1 for p in y_p]
    is_atk_pred = [1 if p >= clf_o else -1 for p in y_p]
    #print(is_atk_pred)

    # print(full_clf_feat)
    # print(atk_full_clf_feat)
    # print(is_atk_dfs["CPO_0"].loc[ATK_START_DATE:])
    timestamp1 = time.time()
    #clf_is_atk = do_get_clf_is_atk(is_atk_dfs, atk_full_clf_feat)
    clf_is_atk = atk_labels_df
    timestamp2 = time.time()
    logger.debug("do_get_clf_is_atk() took %.2f seconds" % (timestamp2 - timestamp1))

    clf_results = get_clf_eval(y=clf_is_atk['is_attack1'], y_pred=is_atk_pred)
    #clf_results = get_clf_eval(y=clf_is_atk['is_attack2'], y_pred=is_atk_pred)
    clf_results["clf"] =  clf
    clf_results["clf_n"] =  clf_n
    clf_results["clf_f"] =  clf_f
    clf_results["clf_o"] =  clf_o
    clf_results["clf_r"] =  clf_r
    #print(clf_results)
    clf_results["decision_function"] =  list(y_p)
    return clf_results



def get_clf_id(clf_n, estimator, df, feat_cols, conf):
    h = hashlib.new('sha256')
    h.update(repr(conf).encode("utf-8"))
    h.update(pd.util.hash_pandas_object(df, index=True).values)
    h.update(repr(sorted(feat_cols)).encode("utf-8"))
    h.update(repr(estimator).encode("utf-8"))
    ha = h.hexdigest()
    clf_id = (clf_n, ha)
    return clf_id

def do_clf(clf_type, config, full_clf_feat, atk_full_clf_feat, offset_=None, pbar=None, scale_pipeline=False, overwrite_jobs=None):
    # print([c for c in full_clf_feat.columns])
    # exit()
    X = full_clf_feat.values
    atk_X = atk_full_clf_feat.values

    if overwrite_jobs is not None:
        if "n_jobs" in config:
            config["n_jobs"] = overwrite_jobs

    estimators = [e for e in get_clfs_from_string(clf_type, config=config)] #
    if len(estimators) > 1:
        raise ValueError(f"len(estimators) > 1: {estimators=}")
    

    if not sys.warnoptions and (clf_type == "OneClassSVM" or clf_type == "EllipticEnvelope"):
        warnings.simplefilter("ignore")
        os.environ["PYTHONWARNINGS"] = "ignore" # Also affect subprocesses
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", category=ConvergenceWarning)
        for estimator in estimators: #should be only one since config...
            if scale_pipeline:
                estimator = make_pipeline(StandardScaler(),estimator)

            timestamp1 = time.time()
            clf_id = get_clf_id(clf_type, estimator, full_clf_feat, [c for c in full_clf_feat.columns], config)
            if clf_id in CLF_CACHE:
                logger.debug(f"hit {len(CLF_CACHE)}")
                estimator = deepcopy(CLF_CACHE[clf_id])
            else:
                logger.debug(f"miss {len(CLF_CACHE)}")
                estimator.fit(X=X)
                CLF_CACHE[clf_id] = estimator

            timestamp2 = time.time()
            logger.debug("fit() took %.2f seconds" % (timestamp2 - timestamp1))


            y_p = estimator.decision_function(X=atk_X)
            timestamp3 = time.time()
            logger.debug("decision_function() took %.2f seconds" % (timestamp3 - timestamp2))
            #y_pred = estimator.predict(X=atk_X)
            if pbar is not None:
                pbar.update(1)
            return y_p

def get_clf_eval(y, y_pred):
    with warnings.catch_warnings():
        warnings.simplefilter('ignore', UndefinedMetricWarning)
        #precision0, recall0, fscore0, support0 = precision_recall_fscore_support(y, y_pred)
        precision1, recall1, fscore1, support1 = precision_recall_fscore_support(y, y_pred, average="binary", labels=[1, -1])
        #precision2, recall2, fscore2, support2 = precision_recall_fscore_support(y, y_pred, average="weighted")
        tn, fp, fn, tp = confusion_matrix(y, y_pred, labels=[1, -1]).ravel()
        accuracy = accuracy_score(y, y_pred)

        if (tp+fn) != 0:
            tpr = tp/(tp+fn)
        else:
            tpr=0
        if (fp+tn) != 0:
            fpr = fp/(fp+tn)
        else:
            fpr=0
    ret = {
        "tpr": tpr,
        "fpr": fpr,
        "precision": precision1,
        "recall": recall1,
        "fscore": fscore1,
        "support": support1,
        "tn": tn,
        "fp": fp,
        "fn": fn,
        "tp": tp,
        "accuracy": accuracy,
    }
    return ret

def cross_val_clf(clf_type, full_clf_feat, feat_cols, clf_feat_cols, pred_col=None, only_clf_feat=False, pbar=None, scale_pipeline=False, train_contamination=False):
    full_feat_cols = sorted(feat_cols + clf_feat_cols)
    if only_clf_feat:
        full_feat_cols = sorted(clf_feat_cols)
    # print(full_clf_feat)
    # print(full_feat_cols)
    # # print([c for c in full_clf_feat.columns if "grid" in c])
    # exit()
    X = full_clf_feat[full_feat_cols].values

    if pred_col is None:
        y = [1] * len(X)
    else:
        y = full_clf_feat[pred_col].values

    ret_l = []
    estimators = get_clfs_from_string(clf_type, train_contamination=train_contamination) #get list w/ hyper params...
    # LOF: decision_function: The shift offset allows a zero threshold for being an outlier.
    offset_ = 0

    # IF: When the contamination parameter is set to "auto", the offset is equal to -0.5 as the scores of inliers are close to 0 and the scores of outliers are close to -1. 
    # LOF: inliers score around -1 (the higher, the less abnormal).
    # LOF: The offset is set to -1.5 (inliers score around -1), except when a
    #     contamination parameter different than "auto" is provided. 
    # if clf_type== "IsolationForest":
    #     offset_ = -0.5
    # elif clf_type== "LocalOutlierFactor":
    #     offset_ = -1.5

    
    for estimator in estimators:
        clf_params = estimator.get_params()
        if scale_pipeline:
            estimator = make_pipeline(StandardScaler(),estimator)
            if "make_pipeline(StandardScaler," not in clf_type:
                clf_type = f"make_pipeline(StandardScaler,{clf_type})"
        now = datetime.now()
        try:
            if not sys.warnoptions and ("OneClassSVM" in clf_type or "EllipticEnvelope" in clf_type):
                warnings.simplefilter("ignore")
                os.environ["PYTHONWARNINGS"] = "ignore" # Also affect subprocesses
            with warnings.catch_warnings():
                warnings.filterwarnings("ignore", category=ConvergenceWarning)

                nt=NUM_THREADS
                if "LocalOutlierFactor" in clf_type or "IsolationForest" in clf_type: #native multi-threading
                    nt=1 #doesnt seem to work...
                # if clf_type == "EllipticEnvelope" and not only_clf_feat:
                #     nt=1
                #if not only_clf_feat:
                #    nt=round(nt/2)
                pre_dispatch=nt
                #logger.info(f"{pre_dispatch=}, {nt=}")
                with parallel_backend("threading", n_jobs=nt):
                    with threadpool_limits(limits=nt, user_api='blas'):
                        os.environ["OMP_NUM_THREADS"] = str(nt)
                        os.environ["MKL_NUM_THREADS"] = str(nt)
                        os.environ["OPENBLAS_NUM_THREADS"] = str(nt)
                        os.environ["BLIS_NUM_THREADS"] = str(nt)
                        y_p = cross_val_predict(estimator, X, y=None, cv=5, n_jobs=nt, method="decision_function", pre_dispatch=pre_dispatch) #Returns -1 for anomalies/outliers and 1 for inliers.

                #print(y_p[:100])
                y_pred = [1 if p >= offset_ else -1 for p in y_p]
                # y_pred = cross_val_predict(estimator, X, y=None, cv=5, n_jobs=NUM_THREADS) #Returns -1 for anomalies/outliers and 1 for inliers.
        except Exception as e:
            logger.error(e)
            logger.error(str(estimator))
            # logger.error(full_clf_feat[full_feat_cols])
            # logger.error(full_clf_feat[[c for c in full_clf_feat.columns if "pred"in c]])
            ret = {
                "clf_type": clf_type,
                "clf_params": clf_params, 
                "full_feat_cols": full_feat_cols,
                "y_p": str(e),
            }
            if pbar is not None:
                pbar.update(1)
            ret_l.append(ret)
            #raise e
            continue

        with warnings.catch_warnings():
            warnings.simplefilter('ignore', UndefinedMetricWarning)
            #precision0, recall0, fscore0, support0 = precision_recall_fscore_support(y, y_pred)
            precision1, recall1, fscore1, support1 = precision_recall_fscore_support(y, y_pred, average="binary", labels=[1, -1])
            #precision2, recall2, fscore2, support2 = precision_recall_fscore_support(y, y_pred, average="weighted")
            tn, fp, fn, tp = confusion_matrix(y, y_pred, labels=[1, -1]).ravel()
            accuracy = accuracy_score(y, y_pred)

            # print(y[:100])
            # print(y_pred[:100])

            # print(precision1, recall1, fscore1, support1)
            # print(tn, fp, fn, tp)
            # print(accuracy)

        later = datetime.now()
        difference = (later - now).total_seconds()
        ret = {
            "clf_type": clf_type,
            #"clf": str(estimator),
            "clf_params": clf_params, 
            "full_feat_cols": full_feat_cols,
            #"y_true": y,
            "y_p": list(y_p),
            #"y_pred": list(y_pred),
            #"offset_":offset_,
            "precision": precision1,
            "recall": recall1,
            "fscore": fscore1,
            "tn": tn,
            "fp": fp,
            "fn": fn,
            "tp": tp,
            "accuracy": accuracy,
            "t_diff": difference,
        }
        if pbar is not None:
            pbar.update(1)
        ret_l.append(ret)
    return ret_l




def prune_clf_feats(df, feat_cases=None, df_dict=None):
    feat_cols = list(df.columns)
    if feat_cases is None or "all" in feat_cases:
        return feat_cols

    # print([c for c in df_dict["cp_g_df"].columns if  "_non_ev_power_" in c])
    # print([c for c in feat_cols if  "_non_ev_power_" in c])
    # exit()
    

    add_expo = [feat_case for feat_case in feat_cases if "add_grid_load_expo_" in feat_case]
    if add_expo:
        if df_dict is None:
            logger.error(f"add_grid_load_expo_ provided: {add_expo}; but no df_dict")
            exit()
        if len(add_expo) > 1:
            logger.warning(f"multiple add_grid_load_expo_ sets provided: {add_expo}; using first")
        add_expo = add_expo[0]
        sub_case="static"
        if "expo_rnd" in add_expo:
            sub_case="rnd"
        add_expo_fac_min = int(add_expo.split("_")[-1])
        #add_expo_fac_max = min(add_expo_fac_min+10, 100)
        #print(add_expo, add_expo_fac_min, add_expo_fac_max)
        expo_cols = [c for c in df_dict["cp_g_df"].columns if "grid_expo_" in c and sub_case+"_fac."+str(add_expo_fac_min) in c and "_relation_to_" not in c]
        feat_cols = [c for c in feat_cols if "grid_expo_" not in c or (sub_case+"_fac."+str(add_expo_fac_min) in c)]
        # print(expo_cols)
        # print(feat_cols)
        feat_cols = feat_cols + expo_cols
        #print(feat_cols)
    else:
        feat_cols = [f for f in feat_cols if "grid_expo_"not in f]

    # if "add_rel_to_cols" in feat_cases:
    #     add_rel_cols= [c for c in df_dict["cp_g_df"].columns if "_relation_to" in c]
    #     get_rel_expo_cols = [c for c in feat_cols if "_norm" not in c and "grid_expo_" in c]

    #     #print(df_dict["cp_g_df"])
    #     new_cols=[]
    #     for rec in get_rel_expo_cols:
    #         name = rec.replace("_lag_0","")+"_relation_to_"+df_dict["pred_col"].replace("_lag_0","")+"_lag_0"
    #         df_dict["cp_g_df"].loc[(df_dict["cp_g_df"][rec] != 0), name] = df_dict["cp_g_df"][df_dict["pred_col"]] / df_dict["cp_g_df"][rec]
    #         df_dict["cp_g_df"].loc[(df_dict["cp_g_df"][rec] == 0), name] = 0
    #         new_cols.append(name)

    #         bus_col = [c for c in df_dict["cp_g_df"].columns if "bus" in c and ".active_power" in c and "_norm" not in c and "_lag_0" in c and "_relation_to_" not in c][0]
    #         name = bus_col.replace("_lag_0","")+"_relation_to_"+rec.replace("_lag_0","")+"_lag_0"
    #         df_dict["cp_g_df"].loc[(df_dict["cp_g_df"][rec] != 0), name] = df_dict["cp_g_df"][bus_col] / df_dict["cp_g_df"][rec]
    #         df_dict["cp_g_df"].loc[(df_dict["cp_g_df"][rec] == 0), name] = 0
    #         new_cols.append(name)

    #     new_cols_df = df_dict["cp_g_df"][new_cols]
    #     normalize_cols(new_cols_df, "_prune_norm")
    #     for c in new_cols_df.columns:
    #         if "_norm" in c and c not in df_dict["cp_g_df"].columns:
    #             new_c = c.replace("_lag_0_prune_norm", "_prune_norm_lag_0")
    #             df_dict["cp_g_df"][new_c] = new_cols_df[c]
    #             new_cols.append(new_c)

    #     add_rel_cols = add_rel_cols + new_cols
    #     #print(df_dict["pred_col"].replace("_lag_0",""))
    #     feat_cols = feat_cols + add_rel_cols

    if feat_cases is not None and "set_1" in feat_cases:
        set_1 = ['prediction_0_diff', 'prediction_mean_diff', ]
        set_1 +=[f for f in feat_cols if "grid_meas_"in f and ".bus." in f and "non_ev_power_relation_to_bus_lag_0" in f and "_norm" not in f] #'grid_meas_0.0.bus.11.non_ev_power_relation_to_bus_lag_0', 
        set_1 +=[f for f in feat_cols if "grid_expo_"in f and ".load." in f and ".active_power_static_fac." in f and 
                 ("_max_ev_power_relation_to_bus_lag_0" in f or # 'grid_expo_0.0.load.7.active_power_static_fac.100_max_ev_power_relation_to_bus_lag_0', 
                  "_non_ev_power_relation_to_expo_lag_0" in f or # 'grid_expo_0.0.load.7.active_power_static_fac.100_non_ev_power_relation_to_expo_lag_0', 
                  "_relation_to_charge_speed_lag_0" in f or # 'grid_expo_0.0.load.7.active_power_static_fac.100_relation_to_charge_speed_lag_0',
                  "_relation_to_prediction_mean_lag_0" in f) and  # 'grid_expo_0.0.load.7.active_power_static_fac.100_relation_to_prediction_mean_lag_0',     
                  "_norm" not in f]
       
        feat_cols = [f for f in feat_cols if f in set_1]
        # print(set_1)
        # print(feat_cols)
        # exit()
        return feat_cols
    elif feat_cases is not None and "set_2" in feat_cases:
        set_2 = ['prediction_0_diff', 'prediction_mean_diff', ]       
        feat_cols = [f for f in feat_cols if f in set_2]
        return feat_cols
    elif feat_cases is not None and "set_3" in feat_cases:
        set_x = ['prediction_0_diff', 'prediction_mean_diff', ]       
        set_x +=[f for f in feat_cols if "grid_meas_"in f and ".bus." in f and ".non_ev_power_relation_to_bus_lag_0" in f and "_norm" not in f] #grid_meas_0.0.bus.11.non_ev_power_relation_to_bus_lag_0
        set_x +=[f for f in feat_cols if "grid_meas_"in f and ".bus." in f and ".only_ev_power_relation_to_bus_lag_0" in f and "_norm" not in f] #grid_meas_0.0.bus.11.only_ev_power_relation_to_bus_lag_0
        feat_cols = [f for f in feat_cols if f in set_x]
        # print(set_x)
        # print(feat_cols)
        # exit()
        return feat_cols
    elif feat_cases is not None and "set_4" in feat_cases:
        set_x = ['prediction_0_diff','prediction_1_diff','prediction_2_diff', 'prediction_mean_diff', ]       
        h_set = [c for c in df_dict["cp_g_df"].columns if "date_exog_group_hour_" in c and "_norm" not in c]
        set_x+=h_set
        feat_cols+=h_set
        feat_cols = [f for f in feat_cols if f in set_x]
        # print(set_x)
        # print(feat_cols)
        # exit()
        return feat_cols
    elif feat_cases is not None and "set_5" in feat_cases:
        set_x = [f for f in feat_cols if 'prediction_' in f and "_norm" not in f and "_rmse" not in f] 
        feat_cols = [f for f in feat_cols if f in set_x]
        return feat_cols
    elif feat_cases is not None and "set_6" in feat_cases:
        set_x = [f for f in feat_cols if 'prediction_' in f and "_norm" not in f] 
        h_set = [c for c in df_dict["cp_g_df"].columns if "date_exog_group_hour_" in c and "_norm" not in c]
        h_set +=[df_dict["pred_col"]]
        set_x+=h_set
        feat_cols+=h_set
        feat_cols = [f for f in feat_cols if f in set_x]
        return feat_cols
    elif feat_cases is not None and "set_7" in feat_cases:
        set_x = [f for f in feat_cols if 'prediction_' in f and "_norm" in f] 
        h_set = [c for c in df_dict["cp_g_df"].columns if "date_exog_group_hour_" in c and "_norm" in c]
        h_set +=["charge_speed_group_norm_lag_0"]
        set_x+=h_set
        feat_cols+=h_set
        feat_cols = [f for f in feat_cols if f in set_x]
        return feat_cols
    elif feat_cases is not None and "set_8" in feat_cases:
        set_x = [f for f in feat_cols if 'prediction_' in f and "_norm" not in f] 
        h_set = [c for c in df_dict["cp_g_df"].columns if "date_exog_group_hour_" in c and "_norm" not in c]
        set_x+=h_set
        feat_cols+=h_set
        feat_cols = [f for f in feat_cols if f in set_x]
        return feat_cols
    elif feat_cases is not None and "set_9" in feat_cases:
        set_x = [f for f in feat_cols if 'prediction_' in f and "_norm" in f] 
        h_set =["charge_speed_lag_0"]
        set_x+=h_set
        feat_cols+=h_set
        feat_cols = [f for f in feat_cols if f in set_x]
        return feat_cols
    elif feat_cases is not None and "set_10" in feat_cases:
        set_x = ['prediction_mean', 'prediction_min', 'prediction_max', 'prediction_mean_diff', 'prediction_min_diff', 'prediction_max_diff'] 
        h_set =["charge_speed_lag_0"]
        set_x+=h_set
        feat_cols+=h_set
        feat_cols = [f for f in feat_cols if f in set_x]
        # print(set_x)
        # print(feat_cols)
        # exit()
        return feat_cols
    elif feat_cases is not None and "set_11" in feat_cases:
        set_x = ['prediction_mean', 'prediction_min', 'prediction_max', 'prediction_mean_diff', 'prediction_min_diff', 'prediction_max_diff'] 
        feat_cols = [f for f in feat_cols if f in set_x]
        # print(set_x)
        # print(feat_cols)
        # exit()
        return feat_cols
    elif feat_cases is not None and "set_12" in feat_cases:
        set_x = ['prediction_mean_diff', 'prediction_min_diff', 'prediction_max_diff'] 
        feat_cols = [f for f in feat_cols if f in set_x]
        # print(set_x)
        # print(feat_cols)
        # exit()
        return feat_cols
    elif feat_cases is not None and "set_13" in feat_cases:
        set_x = [f for f in feat_cols if 'prediction_' in f and "_norm" not in f and "_rmse" not in f and "_diff" in f and "_abs" not in f and "_mean" in f] 
        feat_cols = [f for f in feat_cols if f in set_x]
        # print(set_x)
        # print(feat_cols)
        # exit()
        return feat_cols
    
    elif feat_cases is not None and "set_31" in feat_cases:
        set_x = ['prediction_mean', 'prediction_min', 'prediction_max', 'prediction_mean_diff', 'prediction_min_diff', 'prediction_max_diff']  
        set_x +=[f for f in feat_cols if "grid_meas_"in f and ".bus." in f and ".non_ev_power_relation_to_bus_lag_0" in f and "_norm" not in f] #grid_meas_0.0.bus.11.non_ev_power_relation_to_bus_lag_0
        set_x +=[f for f in feat_cols if "grid_meas_"in f and ".bus." in f and ".only_ev_power_relation_to_bus_lag_0" in f and "_norm" not in f] #grid_meas_0.0.bus.11.only_ev_power_relation_to_bus_lag_0
        feat_cols = [f for f in feat_cols if f in set_x]
        return feat_cols
    elif feat_cases is not None and "set_32" in feat_cases:
        set_x = ['prediction_mean', 'prediction_min', 'prediction_max', 'prediction_mean_diff', 'prediction_min_diff', 'prediction_max_diff']  
        set_x +=[f for f in feat_cols if "grid_meas_"in f and ".bus." in f and ".non_ev_power_relation_to_bus_lag_0" in f and "_norm" not in f] #grid_meas_0.0.bus.11.non_ev_power_relation_to_bus_lag_0
        set_x +=[f for f in feat_cols if "grid_meas_"in f and ".bus." in f and ".only_ev_power_relation_to_bus_lag_0" in f and "_norm" not in f] #grid_meas_0.0.bus.11.only_ev_power_relation_to_bus_lag_0
        h_set = [c for c in df_dict["cp_g_df"].columns if "date_exog_group_hour_" in c and "_norm" not in c]
        h_set +=["charge_speed_lag_0"]
        set_x += h_set
        feat_cols+=h_set
        feat_cols = [f for f in feat_cols if f in set_x]
        return feat_cols
    elif feat_cases is not None and "set_33" in feat_cases:
        set_x = ['prediction_mean_diff', 'prediction_min_diff', 'prediction_max_diff']  
        set_x +=[f for f in feat_cols if "grid_meas_"in f and ".bus." in f and ".non_ev_power_relation_to_bus_lag_0" in f and "_norm" not in f] #grid_meas_0.0.bus.11.non_ev_power_relation_to_bus_lag_0
        set_x +=[f for f in feat_cols if "grid_meas_"in f and ".bus." in f and ".only_ev_power_relation_to_bus_lag_0" in f and "_norm" not in f] #grid_meas_0.0.bus.11.only_ev_power_relation_to_bus_lag_0
        h_set = [c for c in df_dict["cp_g_df"].columns if "date_exog_group_hour_" in c and "_norm" not in c]
        set_x += h_set
        feat_cols+=h_set
        feat_cols = [f for f in feat_cols if f in set_x]
        return feat_cols
    elif feat_cases is not None and "set_34" in feat_cases:
        set_x = ['prediction_mean', 'prediction_min', 'prediction_max', 'prediction_mean_diff', 'prediction_min_diff', 'prediction_max_diff']  
        set_x +=[f for f in feat_cols if "grid_meas_"in f and ".bus." in f and ".non_ev_power_relation_to_bus_lag_0" in f and "_norm" not in f] #grid_meas_0.0.bus.11.non_ev_power_relation_to_bus_lag_0
        set_x +=[f for f in feat_cols if "grid_meas_"in f and ".bus." in f and ".only_ev_power_relation_to_bus_lag_0" in f and "_norm" not in f] #grid_meas_0.0.bus.11.only_ev_power_relation_to_bus_lag_0
        h_set =["charge_speed_lag_0"]
        set_x += h_set
        feat_cols+=h_set
        feat_cols = [f for f in feat_cols if f in set_x]
        # print(set_x)
        # print(feat_cols)
        # exit()
        return feat_cols
    
    elif feat_cases is not None and "set_35" in feat_cases:
        set_x = ['prediction_mean_diff', ]       
        set_x +=[f for f in feat_cols if "grid_meas_"in f and ".bus." in f and ".non_ev_power_relation_to_bus_lag_0" in f and "_norm" not in f] #grid_meas_0.0.bus.11.non_ev_power_relation_to_bus_lag_0
        set_x +=[f for f in feat_cols if "grid_meas_"in f and ".bus." in f and ".only_ev_power_relation_to_bus_lag_0" in f and "_norm" not in f] #grid_meas_0.0.bus.11.only_ev_power_relation_to_bus_lag_0
        set_x +=[f for f in feat_cols if "grid_expo_"in f and ".load." in f and ".active_power_static_fac." in f and 
                 ("_max_ev_power_relation_to_bus_lag_0" in f or # 'grid_expo_0.0.load.7.active_power_static_fac.100_max_ev_power_relation_to_bus_lag_0', 
                  "_non_ev_power_relation_to_expo_lag_0" in f or # 'grid_expo_0.0.load.7.active_power_static_fac.100_non_ev_power_relation_to_expo_lag_0', 
                  "_relation_to_charge_speed_lag_0" in f or # 'grid_expo_0.0.load.7.active_power_static_fac.100_relation_to_charge_speed_lag_0',
                  "_relation_to_prediction_mean_lag_0" in f) and  # 'grid_expo_0.0.load.7.active_power_static_fac.100_relation_to_prediction_mean_lag_0',     
                  "_norm" not in f]
        feat_cols = [f for f in feat_cols if f in set_x]
        # print(set_x)
        # print(feat_cols)
        # exit()
        return feat_cols
    elif feat_cases is not None and "set_352" in feat_cases:
        set_x = ['prediction_mean_diff', ]       
        set_x +=[f for f in feat_cols if "grid_meas_"in f and ".bus." in f and ".non_ev_power_relation_to_bus_lag_0" in f and "_norm" not in f] #grid_meas_0.0.bus.11.non_ev_power_relation_to_bus_lag_0
        set_x +=[f for f in feat_cols if "grid_meas_"in f and ".bus." in f and ".only_ev_power_relation_to_bus_lag_0" in f and "_norm" not in f] #grid_meas_0.0.bus.11.only_ev_power_relation_to_bus_lag_0
        set_x +=[f for f in feat_cols if "grid_expo_"in f and ".load." in f and ".active_power_" in f and "_fac." in f and #add_grid_load_expo_rnd_90
                 ("_max_ev_power_relation_to_bus_lag_0" in f or # 'grid_expo_0.0.load.7.active_power_static_fac.100_max_ev_power_relation_to_bus_lag_0', 
                  "_non_ev_power_relation_to_expo_lag_0" in f or # 'grid_expo_0.0.load.7.active_power_static_fac.100_non_ev_power_relation_to_expo_lag_0', 
                  "_relation_to_charge_speed_lag_0" in f or # 'grid_expo_0.0.load.7.active_power_static_fac.100_relation_to_charge_speed_lag_0',
                  "_relation_to_prediction_mean_lag_0" in f) and  # 'grid_expo_0.0.load.7.active_power_static_fac.100_relation_to_prediction_mean_lag_0',     
                  "_norm" not in f]
        feat_cols = [f for f in feat_cols if f in set_x]
        # print(set_x)
        # print(feat_cols)
        # exit()
        return feat_cols

    for feat_case in feat_cases:
        if feat_case == "only_norm":
            feat_cols = [f for f in feat_cols if "_norm" in f]
        elif feat_case == "no_norm":
            feat_cols = [f for f in feat_cols if "_norm" not in f]
        elif feat_case == "only_diff":
            feat_cols = [f for f in feat_cols if "_diff" in f or "_relation_to" in f]

    return feat_cols


def get_best_clf_conf_offset(y_p, max_fpr):
    if y_p["y_p"] is None:
        return None, None
    offset = 0
    y_p = sorted(y_p["y_p"])
    best_y_p = 0
    p_y_p = len(y_p) * max_fpr
    # print(y_p)
    for min_y in y_p:
        _best_y_p = len([y for y in y_p if y < min_y]) #len of false positives
        if _best_y_p > p_y_p: #more false postivies than allowed
            break
        offset = min_y
        best_y_p = _best_y_p
    # print(offset, best_y_p, p_y_p, len(y_p))
    # exit()
    return offset, best_y_p

def get_best_clf_conf2(dfr, max_fpr=0.005, ignore_fscore=False): # >= offset_ -> inlier=1
    #print(dfr)
    res = dfr.apply(lambda x: get_best_clf_conf_offset(x, max_fpr=max_fpr), axis=1)
    dfr["offset"], dfr["false_positives"]= zip(*res)
    if ignore_fscore:
        df_best = dfr
    else:
        df_best = dfr[dfr["fscore"] == dfr["fscore"].max()]
    df_best_o = df_best[df_best["offset"] == df_best["offset"].max()]
    df_best_f = df_best_o[df_best_o["t_diff"] == df_best_o["t_diff"].min()].iloc[0] 
    # print(df_best_f, df_best_f["offset"])
    # exit()
    return df_best_f, df_best_f["offset"]

def get_best_clf_conf(dfr, max_fpr=0.01):
    df_best = dfr[dfr["fscore"] == dfr["fscore"].max()]
    df_best_f = df_best[df_best["t_diff"] == df_best["t_diff"].min()].iloc[0]
    return df_best_f, 0

def to_float_list(val):
    val = val.replace("[", "").replace("]", "")
    try:
        ret = [float(v) for v in val.split(", ")]
    except Exception as e:
        logger.info(f"{e} with {val}")
        return None
    return ret

def _eval_tuning_clf(dfr):
    df_best_f, offset = get_best_clf_conf2(dfr.copy())
    # print(group,clf,reg,df_best_f["clf_params"],df_best_f["fscore"], offset)
    # exit()
    return  {"conf":df_best_f["clf_params"], "eval":df_best_f["fscore"], "feat":ast.literal_eval(df_best_f["full_feat_cols"]), "offset": offset}

def eval_tuning_clf(x_path, tqdm_n=0, full=False):
    onlyfiles = [f for f in os.listdir(x_path) if os.path.isfile(os.path.join(x_path, f))]

    #CP_4_LocalOutlierFactor.no_reg.only_norm.csv.gz
    regex = re.compile(r"^(CP_\d+)_([\w\.\_]+)\.csv\.gz$")
    eval_tuning_files=defaultdict(lambda: dict())
    for f in tqdm(onlyfiles, desc=f"eval_tuning {x_path} get eval_tuning_files", disable=tqdm_n>0):
        #cpo_ocpp_data_CP_1_0bc01518df38656accbde55bfb0e38591.csv.gz
        result = regex.search(f)
        if result: #
            f_df = pd.read_csv(os.path.join(x_path, f), index_col=0) # , converters={"clf_params":ast.literal_eval, "full_feat_cols":ast.literal_eval, "y_p":ast.literal_eval} , "features":ast.literal_eval, "clf_features":ast.literal_eval
            # f_df['clf_params'] = f_df['clf_params'].apply(lambda x: ast.literal_eval(x))
            # f_df['full_feat_cols'] = f_df['full_feat_cols'].apply(lambda x: ast.literal_eval(x))
            f_df['y_p'] = f_df['y_p'].apply(lambda x: to_float_list(x))
            # print(f_df)
            # print(f_df.columns)
            eval_tuning_files[result.group(1)][result.group(2)] = f_df
            #break
        else:
            logger.info(f"unk file {os.path.join(x_path, f)}")
            continue

    if full:
        return eval_tuning_files

    ret_dict=defaultdict(lambda: dict())
    results=[]
    #with ThreadPoolExecutor(NUM_THREADS) as pool:
    with ProcessPoolExecutor(NUM_THREADS) as pool:
        for group,group_d in tqdm(eval_tuning_files.items(), desc=f"eval_tuning {x_path} get ret_dict", disable=True):
            for clf, df in group_d.items():
                for index, rsf in df[["reg", 'shifts', 'features']].drop_duplicates().iterrows():
                    # print(rsf["reg"])
                    dfr = df[(df["reg"] == rsf["reg"]) & (df["shifts"] == rsf["shifts"]) & (df["features"] == rsf["features"])]
                    reg = dfr["reg"].iloc[0] +"."+ str(dfr["shifts"].iloc[0]) +"."+ ".".join(ast.literal_eval(dfr["features"].iloc[0]))

                    results.append( (group,clf,reg, pool.submit(_eval_tuning_clf, dfr)) )
                    #ret_dict[group][(clf,reg)] = _eval_tuning_clf(dfr)

        for group,clf,reg,r in tqdm(results, desc=f"eval_tuning {x_path} get ret_dict", disable=tqdm_n>0):
            ret_dict[group][(clf,reg)] = r.result()
    return ret_dict


def get_eval_dicts_clf(OutDataDIR):
    ret_d=dict()
    DIR=OutDataDIR+"/clf_eval/"

    sub_folders = [name for name in os.listdir(DIR) if os.path.isdir(os.path.join(DIR, name)) if name.startswith("CPO_")]
    for cpo in sub_folders:
        ret_dict=eval_tuning_clf(OutDataDIR+"/clf_eval/"+cpo)
        ret_d[cpo] = ret_dict

    sub_folders = [name for name in os.listdir(DIR) if os.path.isdir(os.path.join(DIR, name)) if name.startswith("DSO")]
    for dso in sub_folders:
        ret_dict=eval_tuning_clf(OutDataDIR+"/clf_eval/"+dso)
        ret_d[dso] = ret_dict

    return ret_d


def save_eval_dicts_clf(best_d, OutDataDIR):
    for actor in sorted(best_d.keys()):
        ret_dict = best_d[actor]
        Path(OutDataDIR+"/clf_eval_dicts/"+actor+"/").mkdir(parents=True, exist_ok=True)
        for group,ret_d2 in tqdm(ret_dict.items(), desc=f"saving {actor}"):
            with gzip.open(OutDataDIR+"/clf_eval_dicts/"+actor+"/"+group+".gz", "wb") as f:
                pickle.dump(ret_d2, f)


def load_eval_dicts_clf(OutDataDIR):
    DIR = OutDataDIR+"/clf_eval_dicts/"
    sub_folders = [name for name in os.listdir(DIR) if os.path.isdir(os.path.join(DIR, name))]

    best_d=defaultdict(lambda: defaultdict(lambda: dict()))
    for _x_path in sub_folders:
        x_path = os.path.join(DIR,_x_path)
        onlyfiles = [f for f in os.listdir(x_path) if os.path.isfile(os.path.join(x_path, f))]
        for fi in onlyfiles:
            with gzip.open(os.path.join(x_path,fi), "rb") as f:
                fdf_dict = pickle.load(f)
                best_d[_x_path][fi.replace(".gz", "")] = fdf_dict
    return best_d

def get_clf_feat_for_conf(cp_g_dict, best_dict, allow_diff_shifts=False):
    #print(cp_g_dict) #[{'reg': 'GradientBoostingRegressor', 'shifts': '5', 'file': 'data/elaadnl/clf_feats/DSO.all/CP_11_GradientBoostingRegressor_5.csv.gz', 'features': ['all']}, ...
    #print(best_dict["reg"].split("."))
    best_dict_reg = best_dict["reg"].split(".")
    for cp_g_d in cp_g_dict:
        if cp_g_d["reg"] == best_dict_reg[0] and cp_g_d["shifts"] == best_dict_reg[1] and  cp_g_d["features"] == best_dict_reg[2:]:
            return cp_g_d["file"]
    for cp_g_d in cp_g_dict:
        if cp_g_d["reg"] == best_dict_reg[0] and allow_diff_shifts and  cp_g_d["features"] == best_dict_reg[2:]:
            
            global feat_warn2
            if feat_warn2 <=1:
                feat_warn2+=1
                logger.warning(f"train and test used diff shifts for features! {allow_diff_shifts=}")
            return cp_g_d["file"]
    return None

def get_atks_old(l_buffer=2): #based on EVAtksScript; TODO: auto gen...
    atk_l=[]
    step=0
    times = [ #Timedelta from sim start time [select attack day]; (hour, minute) to start attack [select attack time]
        (pd.Timedelta(days=1),  0, 20, None), #atk 0 getSeed
        (pd.Timedelta(days=1),  0, 40, None), #atk 1 getCPs
    ]
    atk_cycles=27
    atk_cycles_actual=0
    for x in range(0, atk_cycles, 5): #[0, 5, 10, 15, 20, 25]
        times.append((pd.Timedelta(days=0+x),  8, 00, "atk 2 MAD inc")) #atk 2 MAD inc
        times.append((pd.Timedelta(days=0+x), 13, 00, "atk 3 FDI red")) #atk 3 FDI red
        times.append((pd.Timedelta(days=1+x),  8, 00, "atk 4 FDI inc (no limit)")) #atk 4 FDI inc (no limit)
        times.append((pd.Timedelta(days=1+x), 18, 00, "atk 5 MAD inc and FDI same")) #atk 5 MAD inc and FDI same
        times.append((pd.Timedelta(days=2+x),  8, 00, "atk 6 MAD red")) #atk 6 MAD red
        times.append((pd.Timedelta(days=2+x), 18, 00, "atk 7 FDI inc")) #atk 7 FDI inc
        times.append((pd.Timedelta(days=3+x),  8, 00, "atk 8 MAD inc x2")) #atk 8 MAD inc x2
        times.append((pd.Timedelta(days=3+x), 18, 00, "atk 9 MAD red and FDI inc")) #atk 9 MAD red and FDI inc
        times.append((pd.Timedelta(days=4+x),  8, 00, "atk 10 MAD inc and FDI red")) #atk 10 MAD inc and FDI red
        times.append((pd.Timedelta(days=4+x), 18, 00, "atk 11 MAD inc x1.5")) #atk 11 MAD inc x1.5
        atk_cycles_actual+=1
    steps_per_cycle=10

    times.append((pd.Timedelta(days=31), 23, 59, None)) #Shutdown

    offset = pd.Timedelta(days=2)
    
    setup_times = [0,1]
    times = [(x[0]+offset, x[1], x[2], x[3]) if i not in setup_times else (x[0], x[1], x[2], x[3]) for i,x in enumerate(times)]

    start_sim_clock_time = pd.to_datetime("2023-11-29 00:00:00")

    for x in setup_times:
        atk_time = times[step]
        atk_time_sim = start_sim_clock_time + atk_time[0]
        atk_time_h = atk_time[1]
        atk_time_m = atk_time[2]
        atk_time_sim = atk_time_sim.replace(hour=atk_time_h, minute=atk_time_m)

        atk_l.append({"type":atk_time[3], "start":atk_time_sim, "end":atk_time_sim})
        step += 1

    for x in range(atk_cycles_actual):
        for i in range(steps_per_cycle):
            atk_dur_h = x+1 #*6 *365
            atk_time = times[step]
            if atk_time[3] is None:
                break
            
            atk_time_sim = start_sim_clock_time + atk_time[0]
            
            atk_time_h = atk_time[1] -1 #rnd times
            atk_time_m = atk_time[2]
            # atk_time_h = atk_time[1] + self.rand.randint(low=-1,high=2)
            # atk_time_m = atk_time[2] + self.rand.randint(low= 0,high=60-atk_time[2])
            atk_time_sim = atk_time_sim.replace(hour=atk_time_h, minute=atk_time_m)

            length=15 * 4 * atk_dur_h
            #l_buffer=2#10
            length+=60*(1+3 +l_buffer) #1+3 rnd times
            end_diff = pd.Timedelta(minutes=length)
            atk_time_sim_end = atk_time_sim+end_diff
            atk_l.append({"type":atk_time[3], "start":atk_time_sim, "end":atk_time_sim_end})
            step += 1

    
    for x in ["shutdown"]:
        atk_time = times[step]
        atk_time_sim = start_sim_clock_time + atk_time[0]
        atk_time_h = atk_time[1]
        atk_time_m = atk_time[2]
        atk_time_sim = atk_time_sim.replace(hour=atk_time_h, minute=atk_time_m)

        atk_l.append({"type":atk_time[3], "start":atk_time_sim, "end":atk_time_sim})
        step += 1

    return atk_l


def get_atks(l_buffer=2, start_sim_offset_h=0): #based on EVAtksScript; TODO: auto gen...
    atk_l=[]
    step=0
    times = [ #Timedelta from sim start time [select attack day]; (hour, minute) to start attack [select attack time]
        (pd.Timedelta(days=1),  0, 20, None), #atk 0 getSeed
        (pd.Timedelta(days=1),  0, 40, None), #atk 1 getCPs
    ]
    atk_cycles=27
    atk_cycles_actual=0
    for x in range(0, atk_cycles, 10): #[0, 10, 20,]
        times.append((pd.Timedelta(days=0+x),  7, 00, "atk 2 MAD inc")) #atk 2 MAD inc
        times.append((pd.Timedelta(days=1+x), 14, 00, "atk 3 FDI red")) #atk 3 FDI red
        times.append((pd.Timedelta(days=2+x),  8, 00, "atk 4 FDI inc (no limit)")) #atk 4 FDI inc (no limit)
        times.append((pd.Timedelta(days=3+x), 18, 00, "atk 5 MAD inc and FDI same")) #atk 5 MAD inc and FDI same
        times.append((pd.Timedelta(days=4+x),  8, 00, "atk 6 MAD red")) #atk 6 MAD red
        times.append((pd.Timedelta(days=5+x), 18, 00, "atk 7 FDI inc")) #atk 7 FDI inc
        times.append((pd.Timedelta(days=6+x),  8, 00, "atk 8 MAD inc x2")) #atk 8 MAD inc x2
        times.append((pd.Timedelta(days=7+x), 18, 00, "atk 9 MAD red and FDI inc")) #atk 9 MAD red and FDI inc
        times.append((pd.Timedelta(days=8+x),  8, 00, "atk 10 MAD inc and FDI red")) #atk 10 MAD inc and FDI red
        times.append((pd.Timedelta(days=9+x), 18, 00, "atk 11 MAD inc x2 (no limit) and FDI same")) #atk 11 MAD inc x2 (no limit) and FDI same
        atk_cycles_actual+=1    
    steps_per_cycle=10

    times.append((pd.Timedelta(days=31), 23, 59, None)) #Shutdown

    offset = pd.Timedelta(days=2)
    
    setup_times = [0,1]
    times = [(x[0]+offset, x[1], x[2], x[3]) if i not in setup_times else (x[0], x[1], x[2], x[3]) for i,x in enumerate(times)]

    start_sim_clock_time = pd.to_datetime("2023-11-29 00:00:00")

    for x in setup_times:
        atk_time = times[step]
        atk_time_sim = start_sim_clock_time + atk_time[0]
        atk_time_h = atk_time[1]
        atk_time_m = atk_time[2]
        atk_time_sim = atk_time_sim.replace(hour=atk_time_h, minute=atk_time_m)

        atk_l.append({"type":atk_time[3], "start":atk_time_sim, "end":atk_time_sim})
        step += 1

    for x in range(atk_cycles_actual):
        for i in range(steps_per_cycle):
            atk_dur_h = x+1 #*6 *365
            atk_dur_h = min(atk_dur_h, 4) #max 4 hours...
            atk_time = times[step]
            if atk_time[3] is None:
                break
            
            atk_time_sim = start_sim_clock_time + atk_time[0]
            
            atk_time_h = atk_time[1] - 1 #rnd times
            atk_time_h += start_sim_offset_h 
            atk_time_m = atk_time[2]
            # atk_time_h = atk_time[1] + self.rand.randint(low=-1,high=2)
            # atk_time_m = atk_time[2] + self.rand.randint(low= 0,high=60-atk_time[2])
            atk_time_sim = atk_time_sim.replace(hour=atk_time_h, minute=atk_time_m)

            length=15 * 4 * atk_dur_h
            #l_buffer=2#10
            length+=60*(1+3 +l_buffer) #1+3 rnd times
            end_diff = pd.Timedelta(minutes=length)
            atk_time_sim_end = atk_time_sim+end_diff
            atk_l.append({"type":atk_time[3], "start":atk_time_sim, "end":atk_time_sim_end})
            step += 1

    
    for x in ["shutdown"]:
        atk_time = times[step]
        atk_time_sim = start_sim_clock_time + atk_time[0]
        atk_time_h = atk_time[1]
        atk_time_m = atk_time[2]
        atk_time_sim = atk_time_sim.replace(hour=atk_time_h, minute=atk_time_m)

        atk_l.append({"type":atk_time[3], "start":atk_time_sim, "end":atk_time_sim})
        step += 1

    return atk_l