
import base64
from datetime import datetime
import ipaddress
import logging
import pickle
import sys
import threading
import time
import zlib

import pandas as pd
from scenarios.powerowl_example.extensions.cp_data_factory import CPDataFactory
from wattson.cosimulation.control.messages.wattson_event import WattsonEvent
from wattson.cosimulation.control.scenario_extension import ScenarioExtension

from typing import Dict, List
from wattson.hosts.cp.cp_profile_provider_interface import CPPowerProfileProviderInterface
from wattson.powergrid.simulator.default_configurations.mtu_default_configuration import MtuDefaultConfiguration
from wattson.powergrid.simulator.power_grid_simulator import PowerGridSimulator

from wattson.services.configuration.service_configuration import ServiceConfiguration

from wattson.services.configuration.configuration_expander import ConfigurationExpander
from wattson.services.configuration.configuration_store import ConfigurationStore

from wattson.cosimulation.simulators.network.components.wattson_network_host import WattsonNetworkHost
from wattson.cosimulation.simulators.network.components.wattson_network_switch import WattsonNetworkSwitch
from wattson.cosimulation.simulators.network.emulators.wattson_network_emulator import WattsonNetworkEmulator
import typing
from copy import deepcopy
from powerowl.layers.powergrid.values.grid_value import GridValue
from powerowl.layers.powergrid.values.grid_value_context import GridValueContext
from powerowl.layers.powergrid.elements.attribute_specification import AttributeSpecification

from wattson.cosimulation.simulators.network.components.wattson_network_interface import WattsonNetworkInterface
from wattson.services.wattson_python_service import WattsonPythonService
from wattson.util.misc import dynamic_load_class

from wattson.util.progress_printer import ProgressPrinter

import numpy as np
import pandapower as pp
from wattson.util.random import Random
from numpy import nan
from itertools import islice 

from wattson.cosimulation.control.argument_parser import get_argument_parser
#sudo python3 -m wattson wattson-ev/scenarios/powerowl_example


# MAX_MSG_SIZE=1024 * 16 * 2 * 2
MAX_MSG_SIZE=65527
OSCP_INTERVAL_M=15

class EVChargingExtension(ScenarioExtension):

    def provides_pre_physical(self) -> bool:
        return False
    
    def extend_pre_physical(self):
        config = self.config["config"]

        parser = get_argument_parser()
        args = parser.parse_args()
        r_seed = args.random if args.random else 1
        #r_seed = config["r_seed"]
        Random.set_base_seed(r_seed)

        logger = self.co_simulation_controller.logger.getChild("EVChargingExtension.PrePhysical")
        l_format = '%(asctime)s - %(name)s - %(levelname)s - %(thread)d.%(process)d -  %(message).40s'
        for h in logger.handlers:
            formatter = logging.Formatter(l_format)
            h.setFormatter(formatter)

        network_emulator = self.co_simulation_controller.network_emulator
        physical_simulator = self.co_simulation_controller._physical_simulator

        logger.info(f"EVChargingExtension loaded {config, network_emulator}")

        wattson_client_query_socket = config.get("wattson_client_query_socket")
        wattson_client_publish_socket = config.get("wattson_client_publish_socket")
        logger.info(f"EVChargingExtension loaded {wattson_client_query_socket}")
        logger.info(f"EVChargingExtension loaded {wattson_client_publish_socket}")

        service_configuration = ServiceConfiguration()
        # TODO: Better loading with sanitizing?
        for key, value in config.items():
            service_configuration[key] = value
        logger.info(f"EVChargingExtension loaded {service_configuration}")


        # configuration_store = network_emulator.get_configuration_store()
        # if configuration_store is None:
        #     configuration_store = ConfigurationStore()
        # expander = ConfigurationExpander(configuration_store=configuration_store)
        nn = network_emulator.get_nodes()[0]
        nn = network_emulator.get_node("n361")
        logger.info(f"EVChargingExtension loaded {network_emulator.get_management_ips()}")
        # expanded_configuration = expander.expand_node_configuration(nn, config)
        # logger.info(f"EVChargingExtension loaded {expanded_configuration}")

        for n in network_emulator.get_nodes():
            logger.info(f"EVChargingExtension loaded {n.display_name, n.entity_id}")

        node = WattsonNetworkHost(id="TestCP1")
        network_emulator.add_host(node)

        for n in network_emulator.get_nodes():
            logger.info(f"EVChargingExtension loaded {n.display_name, n.entity_id}")


    def provides_post_physical(self) -> bool:
        return True


    def extend_post_physical(self):

        config = self.config["config"]
        dataset = config.get("dataset")


        parser = get_argument_parser()
        args = parser.parse_args()
        r_seed = args.random if args.random else 1
        #r_seed = config["r_seed"]
        Random.set_base_seed(r_seed)
        rand = Random.get_instance("EVChargingExtension")
        self.rand = rand

        #logger = self.co_simulation_controller.logger.getChild("EVChargingExtension")
        logger = logging.getLogger("Wattson.EVChargingExtension")
        logger.propagate = False
        while logger.hasHandlers():
            if len(logger.handlers) > 0:
                logger.removeHandler(logger.handlers[0])
            else:
                break
        logger.handlers.clear()
        l_format = '%(asctime)s - %(name)s - %(levelname)s - %(thread)d.%(process)d -  %(message).220s'
        l_format = '%(asctime)s - %(name)s - %(levelname)s - %(thread)d.%(process)d -  %(message)s'
        formatter = logging.Formatter(l_format)
        h2 = logging.StreamHandler(sys.stdout)
        h2.setFormatter(formatter)
        logger.addHandler(h2)
        self.logger = logger

        logger.info(f"EVChargingExtension extend_post_physical {r_seed}")
        network_emulator, physical_simulator = self.co_simulation_controller.get_simulators()
        network_emulator = typing.cast(WattsonNetworkEmulator, network_emulator)

        cs = physical_simulator.get_configuration_store()
        for k,v in cs.items():
            logger.debug(f"EVChargingExtension loaded {k}")

        for k,v in cs["!rtu_map"](None, None).items():
            logger.debug(f"EVChargingExtension rtu_map {k, v}")

        # logger.info(f"EVChargingExtension loaded mtu_ids {cs['!mtu_ids'](None, None)}") #['375']
        logger.info(f"EVChargingExtension loaded mtus {cs['!mtus'](None, None)}") #['n375']

        bus_to_rtu=dict()
        for k,v in cs["!datapoints"].items():
            # logger.info(f"EVChargingExtension datapoints {k}")
            bus_list=[]
            for v2 in v:
                if "sources" not in v2["providers"]:
                    continue
                for v3 in v2["providers"]["sources"]:
                    v4 = v3["provider_data"]["grid_element"]
                    if v4.startswith("bus") and v4 not in bus_list:
                        bus_list.append(v4)
                        logger.debug(f"EVChargingExtension datapoints {'mtu' if k in cs['!mtu_ids'](None, None) else 'rtu', k,v4}")
                        if k in cs['!mtu_ids'](None, None):
                            pass
                        else: 
                            bus_to_rtu[v4] = k

        
        from powerowl.simulators.pandapower.pandapower_grid_model import PandaPowerGridModel
        # import typing
        # physical_simulator = typing.cast(PowerGridSimulator, physical_simulator)
        # pandapower_model = physical_simulator._grid_model.to_external()
        # pandapower_model2 = typing.cast(PandaPowerGridModel, physical_simulator._grid_model).get_panda_power_net()
        
        # logger.info(f"EVChargingExtension loaded pandapower_model {pandapower_model['bus']}")
        # logger.info(f"EVChargingExtension loaded pandapower_model2 {pandapower_model2['bus']}")

        # from powerowl.layers.powergrid.elements.bus import Bus
        # pandapower_model2_bus = typing.cast(PandaPowerGridModel, physical_simulator._grid_model).get_pandapower_table(Bus())
        # pandapower_model2_bus2 = typing.cast(PandaPowerGridModel, physical_simulator._grid_model).get_element_type_by_table("bus")
        # logger.info(f"EVChargingExtension loaded pandapower_model2_bus {pandapower_model2_bus}")
        # logger.info(f"EVChargingExtension loaded pandapower_model2_bus2 {pandapower_model2_bus2}")
        
        # pandapower_model2_bus3 = typing.cast(PandaPowerGridModel, physical_simulator._grid_model).get_elements_by_type("bus")
        # for b in pandapower_model2_bus3:
        #     logger.info(f"EVChargingExtension loaded pandapower_model2_bus2 {b.to_dict()}")


        # This pandapower network includes the following parameter tables:
        # - bus (15 element)
        # - load (18 element)
        # - sgen (13 element)
        # - storage (2 element)
        # - switch (8 element)
        # - ext_grid (1 elements)
        # - line (15 element)
        # - trafo (2 element)
        # - bus_geodata (15 element)

        pandapower_model = physical_simulator._grid_model.to_external()
        physical_simulator = typing.cast(PowerGridSimulator, physical_simulator)
        physical_simulator._wattson_time
        start_datetime = physical_simulator.get_configuration_store().get("configuration", {}).get("power-grid", {}).get("profile-loader", {}).get("wattson_time", {}).get("start_datetime", {})
        wattson_time_ref_sim = datetime.strptime(start_datetime, "%Y-%m-%d %H:%M:%S")
        logger.info(f"EVChargingExtension loaded wattson_time_ref_sim {wattson_time_ref_sim}")
        # logger.info(f"EVChargingExtension loaded pandapower_model {pandapower_model}")
        # logger.info(f"EVChargingExtension loaded pandapower_model {pandapower_model['trafo']}")
        # logger.info(f"EVChargingExtension loaded pandapower_model {pandapower_model['bus']}")
        logger.info(f"EVChargingExtension loaded pandapower_model storage len {len(pandapower_model['storage'])}")

        #cp num from grid size
        cp_num_mean=85 #based on https://dl.acm.org/doi/pdf/10.1145/3564625.3568136
        cp_num_stdev=10
        # cp_num_mean=2 
        # cp_num_stdev=0

        #cp_num_mean=100 #use
        #cp_num_stdev=0 #use

        pd.reset_option('^display.', silent=True)

        ACN_min_sec_between_meter=60*1
        ELAAD_min_sec_between_meter=-1
        #load cp data
        if type(dataset) == list:
            cp_data_l = []
            cps_l = []
            for dset in dataset:
                cpdf = CPDataFactory(dset, wattson_time_ref_sim)
                cpdi = cpdf.get_interface()
                logger.info(f"EVChargingExtension load cp data get_data {dset=}")
                cp_data = cpdi.get_data(cpdi.get_data_cp_num())
                cp_data_l.append(cp_data)

                logger.info(f"EVChargingExtension load cp data make_load_cps")
                #pass data to make loads for inital pwr and scaling
                if "acn" in dset:
                    _cp_idxs = self.make_load_cps_same_bus(pandapower_model, cp_data, init_load_scaling=config.get("init_load_scaling"))
                    cps = pandapower_model["storage"].loc[_cp_idxs]
                    cps.loc[:, "min_sec_between_meter"] = ACN_min_sec_between_meter
                else:
                    _cp_idxs = self.make_load_cps(pandapower_model, cp_data, cps_per_bus=cp_num_mean, cp_stdev=cp_num_stdev, init_load_scaling=config.get("init_load_scaling"))
                    cps = pandapower_model["storage"].loc[_cp_idxs]
                    cps.loc[:, "min_sec_between_meter"] = ELAAD_min_sec_between_meter
                cps_l.append(cps)
            
            cp_data=dict()
            for d in cp_data_l:
                cp_data.update(d)
            cps=pd.concat(cps_l)

        else:           
            cpdf = CPDataFactory(dataset, wattson_time_ref_sim)
            cpdi = cpdf.get_interface()
            logger.info(f"EVChargingExtension load cp data get_data {dataset=}")
            cp_data = cpdi.get_data(cpdi.get_data_cp_num())

            logger.info(f"EVChargingExtension load cp data make_load_cps")
            #pass data to make loads for inital pwr and scaling
            if "acn" in dataset:
                _cp_idxs = self.make_load_cps_same_bus(pandapower_model, cp_data, init_load_scaling=config.get("init_load_scaling"))
                cps = pandapower_model["storage"].loc[_cp_idxs]
                cps.loc[:, "min_sec_between_meter"] = ACN_min_sec_between_meter
            else:
                _cp_idxs = self.make_load_cps(pandapower_model, cp_data, cps_per_bus=cp_num_mean, cp_stdev=cp_num_stdev, init_load_scaling=config.get("init_load_scaling"))
                cps = pandapower_model["storage"].loc[_cp_idxs]
                cps.loc[:, "min_sec_between_meter"] = ELAAD_min_sec_between_meter

        # print(cp_data)

        # raise Exception("done")

        physical_simulator._grid_model = PandaPowerGridModel()
        physical_simulator._grid_model.from_external(pandapower_model)
        pandapower_model = physical_simulator._grid_model.to_external()
        # logger.info(f"EVChargingExtension loaded pandapower_model {pandapower_model}")
        # logger.info(f"EVChargingExtension loaded pandapower_model {pandapower_model['bus']}")
        # logger.info(f"EVChargingExtension loaded pandapower_model {pandapower_model['load']}")
        logger.info(f"EVChargingExtension loaded pandapower_model storage len {len(pandapower_model['storage'])}")

        physical_simulator = typing.cast(PowerGridSimulator, physical_simulator)
        # cp_profile = CPPowerProfileProviderInterface(physical_simulator._profile_thread._profile_provider)
        # physical_simulator._profile_thread._profile_provider_factory.interface = cp_profile
        # physical_simulator._profile_thread._profile_provider = cp_profile
        # raise Exception("es")
        

        all_cp_ids = [_cp_id for _, row in cps.iterrows() for _cp_id in row["cp_list"].split("_")]

        mtu_n = cs['!mtus'](None, None)[0]
        mtu_host = network_emulator.get_host(mtu_n)
        dso_ip = [iface for iface in mtu_host.get_interfaces() if not iface.is_management][0].get_ip_address()

        CPO_service = {"services": [
            {"service-type":"python",
             "module":"wattson.hosts.cp.cpo_deployment",
             "class":"CPODeployment",
             "config":{"random_seed":r_seed, #set later
                        "ip": "!management_ips.ctrl",
                        "cpo_node_id": "",
                        "cp_ids": all_cp_ids,
                        "mtu_ip": "!management_ips.!mtus.0",
                        "dso_ip": str(dso_ip),
                         "oscp_interval_m": OSCP_INTERVAL_M, #oscp measurement interval in minutes
                        # "oscp_interval_m": 60, #oscp measurement interval in minutes
                        #"oscp_interval_m": 60 * 8,
                        "wattson_client_query_socket": "!sim-control-query-socket",
                        "wattson_client_publish_socket": "!sim-control-publish-socket"}
            },]}
        
        cpo_ip=None
        cpo_i=0
        for iface in mtu_host.get_interfaces():
            if not iface.is_management:
                if iface.get_link().interface_b.node != mtu_host:
                    conn = iface.get_link().interface_b #get switch for connection of CPO
                else:
                    conn = iface.get_link().interface_a

                CPO_service_config=deepcopy(CPO_service)
                CPO_service_config["services"][0]["config"]["random_seed"] = self.rand.randint(low=1, high=100000)
                cpo_id = "CPO_"+str(cpo_i)
                CPO_service_config["services"][0]["config"]["cpo_node_id"] = cpo_id
                node = WattsonNetworkHost(id=cpo_id, config=CPO_service_config)
                cpo_i+=1
                network_emulator.add_host(node)
                cpo_ip = network_emulator.get_unused_ip(subnet=conn.get_subnet())
                # cpo_ip = ipaddress.IPv4Network(f"{cpo_ip}/{iface.get_subnet_prefix_length()}", strict=False)
                logger.debug(f"EVChargingExtension network_emulator rtu_host new_ip {cpo_ip,iface.get_subnet_prefix_length()}")

                interface_a = WattsonNetworkInterface(node=node, ip_address=cpo_ip, subnet_prefix_length=iface.get_subnet_prefix_length())
                network_emulator.add_interface(node, interface_a)
                interface_b = WattsonNetworkInterface(node=conn.node)
                network_emulator.add_interface(conn.node, interface_b)
                network_emulator.connect_interfaces(interface_a, interface_b)
                node.update_default_route()
                # break

        

                # Create MTU DSO OSCP configuration
                configuration = ServiceConfiguration()
                dso_service_class = dynamic_load_class("wattson.hosts.cp.dso_deployment" + "." + "DSODeployment")
                configuration.update({"random_seed": self.rand.randint(low=1, high=100000),
                                "ip": "!management_ips.ctrl",
                                "cpo_node_id": cpo_id,
                                "cp_ids": all_cp_ids,
                                "cpo_ip":str(cpo_ip),
                                "mtu_ip": "!management_ips.!mtus.0",
                                "dso_ip": str(dso_ip),
                                "wattson_client_query_socket": "!sim-control-query-socket",
                                "wattson_client_publish_socket": "!sim-control-publish-socket"})
                mtu_host.add_service(WattsonPythonService(dso_service_class, configuration, mtu_host))

                #mtu_host.add_service()
                logger.debug(f"mtu_host conf: {mtu_host.get_config().get('services', [])}")
                logger.debug(f"mtu_host conf: {mtu_host.get_service(service_id=1).id, mtu_host.get_service(service_id=1).name, mtu_host.get_service(service_id=1).get_start_command()}")
                logger.debug(f"mtu_host conf: {mtu_host.get_config(), mtu_host.get_services(), mtu_host.get_service(service_id=1)._service_configuration}")
                # raise Exception("done")
                break


        if not cpo_ip:
            raise Exception("No CPO:",cpo_ip)
        
        CP_service = {"services": [
            {"service-type":"python",
             "module":"wattson.hosts.cp.cp_deployment",
             "class":"CPDeployment",
             "config":{"random_seed":r_seed, #set later
                       "cp_id":"",
                       "cp_node_id": "",
                       "scaling":1.0,
                       "cp_data":"",
                       "cpo_ip":str(cpo_ip),
                       "min_sec_between_meter": -1, #skip sending messages if too fast (in seconds; eg 60=1min; -1=deactivate); except session start/end
                    #    "min_sec_between_meter": 60*15, #skip sending messages if too fast (in seconds; eg 60=1min; -1=deactivate); except session start/end
                    #    "min_sec_between_meter": 60*60*4, #4h
                        "ip": "!management_ips.ctrl",
                        "mtu_ip": "!management_ips.!mtus.0",
                        "wattson_client_query_socket": "!sim-control-query-socket",
                        "wattson_client_publish_socket": "!sim-control-publish-socket"}
            },]}
        logger.info(f"EVChargingExtension adding CPs to Network {len(cps.index)}")
        progress_printer = ProgressPrinter(max_progress=len(cps.index), on_stop_margin=True)
        progress_printer.start()

        physical_simulator.grid_model._kwargs["cp_data_dicts"] = dict()
        self.physical_simulator = physical_simulator
        for index, row in cps.iterrows():
            rtu = bus_to_rtu["bus."+str(row["bus"])]
            rtu_host = network_emulator.get_host(rtu)
            # logger.info(f"index {index}")
            # logger.info(f"row {row}")
            logger.debug(f"EVChargingExtension network_emulator rtu_host {rtu_host.display_name, rtu_host.class_id}")
            logger.debug(f"EVChargingExtension network_emulator rtu_host {rtu_host.get_config()}")
            # logger.info(f"EVChargingExtension network_emulator rtu_host {rtu_host.get_subnets()[0].hosts()}")
            for iface in rtu_host.get_interfaces():
                if not iface.is_management:
                    logger.debug(f"EVChargingExtension network_emulator rtu_host {iface.get_link().interface_a.display_name,iface.get_link().interface_b.display_name}")
                    logger.debug(f"EVChargingExtension network_emulator rtu_host {iface.get_link().interface_a.node.display_name, iface.get_link().interface_b.node.display_name}")
                    if iface.get_link().interface_b.node != rtu_host:
                        conn = iface.get_link().interface_b
                    else:
                        conn = iface.get_link().interface_a
                    if type(conn.node) != WattsonNetworkSwitch:
                        raise Exception("type(conn) != WattsonNetworkSwitch")
                    
                    config=deepcopy(CP_service)
                    config["services"][0]["config"]["random_seed"] = self.rand.randint(low=1, high=100000)
                    cp_data_dict = dict((_cp_id, cp_data[_cp_id]) for _cp_id in row["cp_list"].split("_"))
                    cp_data_dict_c = zlib.compress(pickle.dumps(cp_data_dict))
                    cp_data_dict_base = base64.b64encode(cp_data_dict_c).decode("utf-8")
                    # config["services"][0]["config"]["cp_id"] = "CP"+str(index)             
                    config["services"][0]["config"]["cp_id"] = row["name"]
                    config["services"][0]["config"]["scaling"] = row["scaling"]
                    config["services"][0]["config"]["cp_data"] = cp_data_dict_base
                    cp_node_id = "Node"+row["name"]
                    config["services"][0]["config"]["cp_node_id"] = cp_node_id
                    config["services"][0]["config"]["min_sec_between_meter"] = row["min_sec_between_meter"]
                    
                    node = WattsonNetworkHost(id=cp_node_id, config=config)
                    network_emulator.add_host(node)

                    physical_simulator.grid_model._kwargs["cp_data_dicts"][row["name"]] = cp_data_dict

                    # cp_elem = physical_simulator.grid_model.get_element("storage",index) 
                    # cp_data_as = AttributeSpecification(name = "cp_data",
                    #                                     context = GridValueContext.GENERIC,
                    #                                     value_type = dict,
                    #                                     default_value = dict(),
                    #                                     required = False)
                    # cp_data_gv = GridValue.from_specification(cp_data_as, cp_elem)
                    # cp_data_gv.set_value(cp_data_dict, wattson_time_ref_sim.timestamp())
                    # cp_elem.set("cp_data", GridValueContext.GENERIC, cp_data_gv)

                    # conn_sub = ipaddress.IPv4Network(f"{iface.get_ip_address()}/16", strict=False)
                    # new_ip = network_emulator.get_unused_ip(subnet=conn_sub)
                    new_ip = network_emulator.get_unused_ip(subnet=conn.get_subnet())
                    # new_ip = ipaddress.IPv4Network(f"{new_ip}/{iface.get_subnet_prefix_length()}", strict=False)
                    logger.debug(f"EVChargingExtension network_emulator rtu_host new_ip {new_ip,iface.get_subnet_prefix_length()} more like /16")

                    # interface_a = WattsonNetworkInterface(node=node, ip_address=new_ip, subnet_prefix_length=16)
                    interface_a = WattsonNetworkInterface(node=node, ip_address=new_ip, subnet_prefix_length=iface.get_subnet_prefix_length())
                    network_emulator.add_interface(node, interface_a)
                    interface_b = WattsonNetworkInterface(node=conn.node)
                    network_emulator.add_interface(conn.node, interface_b)
                    link = network_emulator.connect_interfaces(interface_a, interface_b)

                    udr = node.update_default_route()
                    for iface2 in network_emulator.get_host(node.entity_id).get_interfaces():
                        logger.debug(f"EVChargingExtension network_emulator rtu_host connect_nodes {iface2.display_name}")
                        logger.debug(f"EVChargingExtension network_emulator rtu_host connect_nodes {iface2.get_link().interface_a.display_name,iface2.get_link().interface_b.display_name}")
                    break

            # break
            progress_printer.inc()
        progress_printer.stop()
        
        # physical_simulator._grid_model.from_primitive_dict(physical_simulator._grid_model.to_primitive_dict())

        # pandapower_model = physical_simulator._grid_model.to_external()
        # physical_simulator._grid_model = PandaPowerGridModel()
        # physical_simulator._grid_model.from_external(pandapower_model)

        # physical_simulator._grid_model.from_primitive_dict(physical_simulator._grid_model.to_primitive_dict())
        
        # logger.info(f"EVChargingExtension network_emulator rtu_host connect_nodes {[k for k in physical_simulator.grid_model._kwargs['cp_data_dicts'].keys()]}")

        # raise Exception("done")


        self.thread_atk = threading.Thread(target = self.atk_sim_server, daemon=True)#, args = (10, )
        self.thread_atk.start()
        #thread.join()
        
        pass


    def scale_post_cps(self, net, init_load_scaling=1.0):
        #logger = self.co_simulation_controller.logger.getChild("EVChargingExtension")
        logger = self.logger
        # - load (18 element)
        # - sgen (13 element)
        # - storage (2 element)
        total_load_p_mw = net["load"]["p_mw"] * net["load"]["scaling"]
        total_storage1_p_mw = net["storage"][net["storage"]["is_cp"] != True]["p_mw"] * net["storage"][net["storage"]["is_cp"] != True]["scaling"]
        total_storage2_p_mw = net["storage"]["p_mw"] * net["storage"]["scaling"]
        total_sgen_p_mw = net["sgen"]["p_mw"] * net["sgen"]["scaling"]

        total_p_mw_pre = total_load_p_mw.sum() + total_storage1_p_mw.sum() - total_sgen_p_mw.sum()

        net["load"]["scaling"] *= init_load_scaling
        total_load_p_mw = net["load"]["p_mw"] * net["load"]["scaling"]

        total_p_mw_post = total_load_p_mw.sum() + total_storage2_p_mw.sum() - total_sgen_p_mw.sum()
        fac = total_p_mw_pre/total_p_mw_post
        logger.info(f"scale_post_cps {total_p_mw_pre} -> {total_p_mw_post}, fac {fac} w/ {total_storage2_p_mw.sum()=}")
        net["load"]["scaling"] *= fac
        net["storage"]["scaling"] *= fac
        net["sgen"]["scaling"] *= fac

        total_load_p_mw = net["load"]["p_mw"] * net["load"]["scaling"]
        total_storage2_p_mw = net["storage"]["p_mw"] * net["storage"]["scaling"]
        total_sgen_p_mw = net["sgen"]["p_mw"] * net["sgen"]["scaling"]
        total_p_mw_post2 = total_load_p_mw.sum() + total_storage2_p_mw.sum() - total_sgen_p_mw.sum()
        logger.info(f"scale_post_cps post {total_p_mw_post2} w/ {total_storage2_p_mw.sum()=}")
        if abs(total_p_mw_pre - total_p_mw_post2) > 0.01:
            raise Exception("scaling error abs(total_p_mw_pre - total_p_mw_post2) > 0.01",total_p_mw_pre,total_p_mw_post2)


    def make_load_cps_same_bus(self, net, cps, init_load_scaling=1.0):
        applicable_loads = net["load"][net["load"]["in_service"] == True]
        busses = sorted(applicable_loads["bus"].drop_duplicates().tolist())

        used_busses = sorted(net["storage"][net["storage"]["type"] == "CP"]["bus"].drop_duplicates().tolist())
        new_busses = [b for b in busses if b not in used_busses]
        self.logger.info(f"EVChargingExtension create_storage {busses=} {new_busses=}")
        if len(new_busses) > 0:
            busses=new_busses
        else:
            raise Exception("no new busses")

        cp_keys = sorted(cps.keys())
        self.rand.shuffle(cp_keys)
        self.rand.shuffle(busses)
        do_bus = busses[0]

        self.logger.info(f"EVChargingExtension create_storage {len(cp_keys)} on bus {do_bus}")
        progress_printer = ProgressPrinter(max_progress=len(cp_keys), on_stop_margin=True)
        progress_printer.start()

        ret_idx=[]
        idx = self.make_create_cp(net, do_bus, cp_keys, cps, progress_printer)
        ret_idx.append(idx)

        progress_printer.stop()

        self.scale_post_cps(net, init_load_scaling=init_load_scaling)

        self.logger.info("len storages %s", len(net.storage))
        return ret_idx
    

    def make_load_cps(self, net, cps, cps_per_bus=85, cp_stdev=10, init_load_scaling=1.0):
        applicable_loads = net["load"][net["load"]["in_service"] == True]
        num_busses = len(applicable_loads["bus"].drop_duplicates())
        #r = np.random.RandomState(int(round(Random.float(lowest=0, highest=10000))))
        cp_num_list = self.rand.normal(cps_per_bus, cp_stdev, num_busses)
        cp_num_list = [int(round(abs(elem))) for elem in cp_num_list]
        total_cp_num = sum(cp_num_list)

        logger = self.logger
        logger.info("make_load_cps %s", total_cp_num)
        while total_cp_num > len(cps.keys()):
            logger.warning(f"{cps_per_bus=} -- total_cp_num > len(cps.keys()) {total_cp_num, len(cps.keys())}")
            logger.warning(f"reducing {cps_per_bus=}")

            cps_per_bus-=1
            cp_num_list = self.rand.normal(cps_per_bus, cp_stdev, num_busses)
            cp_num_list = [int(round(abs(elem))) for elem in cp_num_list]
            total_cp_num = sum(cp_num_list)
            if cps_per_bus == 0:
                raise Exception(f"{cps_per_bus=} -- total_cp_num > len(cps.keys()) {total_cp_num, len(cps.keys())}")
        
        cp_keys = sorted(cps.keys())
        self.rand.shuffle(cp_keys)

        #logger = self.co_simulation_controller.logger.getChild("EVChargingExtension")
        # cps.reset_index(drop=True, inplace=True)

        logger.info(f"EVChargingExtension create_storage {total_cp_num}")
        progress_printer = ProgressPrinter(max_progress=total_cp_num, on_stop_margin=True)
        progress_printer.start()

        CP_Input = iter(cp_keys) 
        CP_Output = [list(islice(CP_Input, elem)) for elem in cp_num_list] 
        ret_idx=[]
        for b,cp_l in zip(sorted(applicable_loads["bus"].drop_duplicates().tolist()), CP_Output):
            idx = self.make_create_cp(net, b, cp_l, cps, progress_printer)
            ret_idx.append(idx)

        progress_printer.stop()

        self.scale_post_cps(net, init_load_scaling=init_load_scaling)

        logger.info("len storages %s", len(net.storage))
        return ret_idx

    def make_create_cp(self, net, b, cp_l, cps, progress_printer):
        # logger.info(f"EVChargingExtension loaded b.cp {b,cp_l}")
        cp_dict = dict()
        cp_dict["bus"] = b
        cp_dict["p_mw"] = 0
        cp_dict["max_p_mw"] = 0
        cp_dict["name"] = "CP_"+str(b)
        cp_dict["max_e_mwh"] = 0
        cp_dict["soc_percent"] = self.rand.choice([0.5,.6,.7], size=1, replace=True)[0]
        cp_dict["q_mvar"]=0; cp_dict["sn_mva"]=nan; cp_dict["scaling"]=1.0; cp_dict["min_e_mwh"]=0.0002
        cp_dict["in_service"]=True; cp_dict["type"]="CP"; cp_dict["min_p_mw"]=0.0; cp_dict["is_cp"]=True
        cp_dict["cp_list"]=[]
        for cp_id in cp_l:
            cp_df = cps[cp_id]
            cp_dict["cp_list"].append(cp_id)
            cp_dict["p_mw"] += cp_df["AveragePower"].iloc[0] / 1000.0 #kW -> mw
            cp_dict["max_p_mw"] += cp_df["AveragePower"].max() / 1000.0 #kW -> mw
            cp_dict["min_p_mw"] = -1 * cp_dict["max_p_mw"]
            cp_dict["max_e_mwh"] += self.rand.choice([79.5/1000, 45.0/1000, 64.0/1000, 45.6/1000], size=1, replace=True) #not (really) applicable for CP use case; numbers for Model 3 (Tesla) ID.3 (VW) Kona (Hyundai) Zoe (Renault)
            cp_dict["soc_percent"] = np.mean([cp_dict["soc_percent"], self.rand.choice([.3,.4,.5,.6,.7,.8,.9], size=1, replace=True)[0]])
            progress_printer.inc()
        cp_dict["cp_list"] = "_".join(cp_dict["cp_list"])

        ret = pp.create_storage(net = net, **cp_dict)
        self.logger.debug(f"EVChargingExtension create_storage {ret}")
        return ret


    def get_cp_df(self, find_cp):
        for cp_id in self.physical_simulator.grid_model._kwargs['cp_data_dicts'].keys():
            for cp_n in self.physical_simulator.grid_model._kwargs['cp_data_dicts'][cp_id].keys():
                if cp_n == find_cp:
                    return self.physical_simulator.grid_model._kwargs['cp_data_dicts'][cp_id][cp_n]
        return None
    
    def atk_sim_server(self):
        import socket
        self._run=True

        if self.co_simulation_controller._simulation_control_server and self.co_simulation_controller._simulation_control_server.event_is_set(WattsonEvent.START):
            time.sleep(15)
        while self._run:
            time.sleep(2)
            self.logger.debug(f"EVChargingExtension waiting...")
            if self.co_simulation_controller._simulation_control_server and self.co_simulation_controller._simulation_control_server.event_is_set(WattsonEvent.START):
                break


        #UDP_IP = str(wattson_client_query_socket).replace("tcp://", "").split(":")[0]
        #self.co_simulation_controller.sim_control_query_socket_string

        # self.logger.info(f"sock bind? {socket.gethostbyname(socket.gethostname())}")
        # s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        # s.connect(("8.8.8.8", 80))
        # self.logger.info(f"sock bind? {s.getsockname()[0]}")
        # s.close()

        self.co_simulation_controller._sim_control_host.get_namespace().thread_attach()
                         

        UDP_IP=self.co_simulation_controller._sim_control_host.get_management_interface().ip_address
        #UDP_IP="127.0.0.1"
        UDP_PORT = 5005
        self.logger.info(f"sock bind: {str(UDP_IP), UDP_PORT}")

        sock = socket.socket(socket.AF_INET, # Internet
                            socket.SOCK_DGRAM) # UDP
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        sock.bind((str(UDP_IP), UDP_PORT))
        #sock.setblocking(False)

        while self._run:
            # try:
            #     data = await asyncio.wait_for(loop.sock_recv(sock, 1024), timeout=60)
            # except asyncio.TimeoutError:
            #     pass
            data, atk_ip = sock.recvfrom(MAX_MSG_SIZE)

            rec_data = pickle.loads(zlib.decompress(data))
            self.logger.info(f"received message: {rec_data, atk_ip}")
            if rec_data["msg"] == 'test':
                # self.logger.info(f"cp_data_dicts: {list(self.physical_simulator.grid_model._kwargs['cp_data_dicts']['CP_1'].keys())}")
                #cp_df = self.physical_simulator.grid_model._kwargs['cp_data_dicts']['CP_1']['16c21e52cca558e0bdd8ebb8095fc7cf1']
                # cp_df.iloc[-1, cp_df.columns.get_loc('MaxPower')] = 100
                # self.logger.info(f"cp_data_dicts: {str(self.physical_simulator.grid_model._kwargs['cp_data_dicts']['CP_1']['16c21e52cca558e0bdd8ebb8095fc7cf1']['MaxPower'])}")
                #self.physical_simulator.grid_model._kwargs["cp_data_dicts"] = None
                EVChargingExtension.send_to_and_check(sock, b"ok", atk_ip, self.logger)
            elif rec_data["msg"] == 'pwr_atk':
                cp_keys = set(cp_key for _cp_id in self.physical_simulator.grid_model._kwargs['cp_data_dicts'] for cp_key in self.physical_simulator.grid_model._kwargs['cp_data_dicts'][_cp_id])
                self.logger.info(f"pwr_atk: {list(_atk_cp for _atk_cp in rec_data['atk_cps'] if _atk_cp in cp_keys)}")
                EVChargingExtension.send_to_and_check(sock, b"ok", atk_ip, self.logger) #early ok to avoid delays/timeouts #do_pwr_atk can take 5+ seconds...

                s = rec_data["start"]
                e = rec_data["end"]
                for _atk_cp in rec_data['atk_cps']:
                    cp_df = self.get_cp_df(_atk_cp)
                    if cp_df is None:
                        self.logger.error(f"pwr_atk cp not found: {_atk_cp}")
                        raise Exception(f"pwr_atk cp not found: {_atk_cp}")

                    EVChargingExtension.do_pwr_atk(cp_df, s, e, set_to=rec_data["set_to"], add=rec_data["add"], fac=rec_data["fac"], ignore_limits=rec_data["ignore_limits"], atk_type=rec_data["atk_type"])
                    # if _atk_cp == "ba3d45d4bf5714a2d6bc4360466731a31":
                    #     #self.logger.info(cp_df["AveragePower"])
                    #     self.logger.info(cp_df.loc[s:e, "AveragePower"])

                # if (self.physical_simulator.grid_model._kwargs['cp_data_dicts']['CP_1']['ba3d45d4bf5714a2d6bc4360466731a31']["AveragePower"] >= 100).any():
                #     self.logger.error(f"AveragePower >= 100")
                # EVChargingExtension.send_to_and_check(sock, b"ok", atk_ip, self.logger)

            # data, addr = sock.recvfrom(1024) # buffer size is 1024 bytes
        self.logger.info("atk_sim_server done.")

    @staticmethod
    def send_to_and_check(sock, out_msg, to_ip, logger):
        if len(out_msg) > MAX_MSG_SIZE:
            logger.error(f"send_and_recv snd msg to large: {len(out_msg)} > {MAX_MSG_SIZE}")
        sock.sendto(out_msg, to_ip)

    @staticmethod
    def do_pwr_atk(cp_df, s, e, set_to, add, fac, ignore_limits, atk_type="UNK"):
        if set_to == None:
            cp_df.loc[s:e, "AveragePower"] = cp_df.loc[s:e, "OriginalAveragePower"]
        elif isinstance(set_to, (int, float, complex)) and not isinstance(set_to, bool):
            cp_df.loc[s:e, "AveragePower"] = set_to
        else:
            raise Exception(f"Unknown type for set_to {set_to, type(set_to)}")
        
        cp_df.loc[s:e, "AveragePower"] += add
        cp_df.loc[s:e, "AveragePower"] *= fac
        cp_df.loc[s:e, "atk_type"] = atk_type
        
        if not ignore_limits:
            CPcMaxPower = cp_df["MaxPower"].max() #* 1000
            cp_df.loc[(cp_df.index >= s) & (cp_df.index <= e) & (cp_df["AveragePower"] > CPcMaxPower), "AveragePower"] = CPcMaxPower
            CPcMinPower = -1* CPcMaxPower
            cp_df.loc[(cp_df.index >= s) & (cp_df.index <= e) & (cp_df["AveragePower"] < CPcMinPower), "AveragePower"] = CPcMinPower                  
        cp_df.loc[(cp_df.index >= s) & (cp_df.index <= e) & (cp_df["IntervalFactor"] == 0), "AveragePower"] = 0 #should not happen since nan (from div by zero) are set to avg instead of 0 in get_data...

        cp_df.loc[s:e, "EnergyInterval"] = cp_df.loc[s:e, "AveragePower"] * cp_df.loc[s:e, "IntervalFactor"]