from collections import defaultdict
from concurrent.futures import ThreadPoolExecutor
from copy import deepcopy
import datetime
from threading import Thread
import time
import traceback
import typing
import zlib
from scenarios.powerowl_example.extensions.ev_charging_extension import MAX_MSG_SIZE, EVChargingExtension
from subprocess import Popen, PIPE

from wattson.apps.script_controller.interface import SoloScript
from time import sleep
from wattson.cosimulation.control.argument_parser import get_argument_parser
from wattson.powergrid.wrapper.power_owl_measurement import PowerOwlMeasurement
from wattson.time.wattson_time_type import WattsonTimeType

from typing import TYPE_CHECKING, Optional, Dict, Any

from wattson.util import get_logger
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from wattson.apps.script_controller import ScriptControllerApp

import pandas as pd
import pickle

from powerowl.layers.powergrid.elements.storage import Storage
from wattson.util.random import Random
from random import random as PyRand

from wattson.powergrid.remote.remote_grid_value import RemoteGridValue

import socket


class EVAtksScript(SoloScript):
    def __init__(self, controller: 'ScriptControllerApp', config: Optional[Dict] = None):
        super().__init__(controller)
        self.times = [ #Timedelta from sim start time [select attack day]; (hour, minute) to start attack [select attack time]
            (pd.Timedelta(days=1),  0, 20,), #atk 0 getSeed
            (pd.Timedelta(days=1),  0, 40,), #atk 1 getCPs
            # (pd.Timedelta(days=2), 12, 00,), #atk 2
            # (pd.Timedelta(days=2), 18, 00,), #atk 3
            # (pd.Timedelta(days=3), 12, 00,), #atk 4
            # (pd.Timedelta(days=3), 18, 00,), #atk 5
            # (pd.Timedelta(days=4), 12, 00,), #atk 6
            # (pd.Timedelta(days=4), 18, 00,), #atk 7
            # (pd.Timedelta(days=5), 18, 00,), #atk 8
            # (pd.Timedelta(days=31), 23, 59,), #Shutdown
        ]
        self.atk_cycles=27
        self.num_of_atks=10
        self.atk_cycles_actual=0
        # for x in range(0, self.atk_cycles, round(self.num_of_atks/2)): #[0, 5, 10, 15, 20, 25]
        #     self.times.append((pd.Timedelta(days=0+x),  7, 00,)) #atk 2 MAD inc
        #     self.times.append((pd.Timedelta(days=0+x), 14, 00,)) #atk 3 FDI red
        #     self.times.append((pd.Timedelta(days=1+x),  8, 00,)) #atk 4 FDI inc (no limit)
        #     self.times.append((pd.Timedelta(days=1+x), 18, 00,)) #atk 5 MAD inc and FDI same
        #     self.times.append((pd.Timedelta(days=2+x),  8, 00,)) #atk 6 MAD red
        #     self.times.append((pd.Timedelta(days=2+x), 18, 00,)) #atk 7 FDI inc
        #     self.times.append((pd.Timedelta(days=3+x),  8, 00,)) #atk 8 MAD inc x2
        #     self.times.append((pd.Timedelta(days=3+x), 18, 00,)) #atk 9 MAD red and FDI inc
        #     self.times.append((pd.Timedelta(days=4+x),  8, 00,)) #atk 10 MAD inc and FDI red
        #     self.times.append((pd.Timedelta(days=4+x), 18, 00,)) #atk 11 MAD inc x1.5
        #     self.atk_cycles_actual+=1
        for x in range(0, self.atk_cycles, self.num_of_atks): #[0, 10, 20,]
            self.times.append((pd.Timedelta(days=0+x),  7, 00,)) #atk 2 MAD inc
            self.times.append((pd.Timedelta(days=1+x), 14, 00,)) #atk 3 FDI red
            self.times.append((pd.Timedelta(days=2+x),  8, 00,)) #atk 4 FDI inc (no limit)
            self.times.append((pd.Timedelta(days=3+x), 18, 00,)) #atk 5 MAD inc and FDI same
            self.times.append((pd.Timedelta(days=4+x),  8, 00,)) #atk 6 MAD red
            self.times.append((pd.Timedelta(days=5+x), 18, 00,)) #atk 7 FDI inc
            self.times.append((pd.Timedelta(days=6+x),  8, 00,)) #atk 8 MAD inc x2
            self.times.append((pd.Timedelta(days=7+x), 18, 00,)) #atk 9 MAD red and FDI inc
            self.times.append((pd.Timedelta(days=8+x),  8, 00,)) #atk 10 MAD inc and FDI red
            self.times.append((pd.Timedelta(days=9+x), 18, 00,)) #atk 11 MAD inc x2 (no limit) and FDI same
            self.atk_cycles_actual+=1

        self.times.append((pd.Timedelta(days=31), 23, 59,)) #Shutdown
        
        # parser = get_argument_parser()
        # args = parser.parse_args()
        # r_seed = args.random if args.random else 1
        # Random.set_base_seed(r_seed)

        self.pre_send_atk_msg_h=3 #send attack message x hours before attack execution to avoid delays #do_pwr_atk can take 5+ seconds for all CPs...

        self.case = "test2"
        if self.case == "base":
            # Shift execution for Timedelta except setup_times
            self.offset = pd.Timedelta(days=334)
            self.include_attacks = False
        elif self.case == "test":
            self.offset = pd.Timedelta(days=1) #30
            self.include_attacks = True
            self.run_tracert_done=False
        elif self.case == "test2":
            self.offset = pd.Timedelta(days=1) #30
            self.include_attacks = True
            self.run_tracert_done=False
        elif self.case == "atk":
            self.offset = pd.Timedelta(days=2)
            self.include_attacks = True
        else:
            raise Exception(f"unk case: {self.case}")
        
        self.setup_times = [0,1]
        self.times = [(x[0]+self.offset, x[1], x[2]) if i not in self.setup_times else (x[0], x[1], x[2]) for i,x in enumerate(self.times)]

        # Whether to also simulate the attackers
        self.attack_p = config["attack_p"]
        self.attack_seed = config["attack_seed"]
        self.rand=None

        self.step = 0 #index for times list
        self.start_timestamp = 0
        self.start_datetime: datetime.datetime = datetime.datetime.now()
        logger = self.controller.logger.getChild("EVAtksScript")
        self.logger = logger
        self.logger.info(f"EVAtks Script config {self.case, config}")


        self.RUN=True
        # self.sock = socket.socket(socket.AF_INET, # Internet
        #                     socket.SOCK_DGRAM) # UDP
        self.sock = dict()


    def rec_oscp_to_state_est(self, UDP_PORT=50099):
        self.logger.info(f"send_token_oob sock bind: {str(self.controller.host_ip), UDP_PORT}")

        sock2 = socket.socket(socket.AF_INET, # Internet
                            socket.SOCK_DGRAM) # UDP
        sock2.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        sock2.bind((str(self.controller.host_ip), UDP_PORT))
        #sock.setblocking(False)
        while self.RUN:
            data, cpo_ip = sock2.recvfrom(MAX_MSG_SIZE)
            #data = data.decode("utf-8")
            rec_data = pickle.loads(zlib.decompress(data))
            self.logger.debug(f"received message: {rec_data, cpo_ip}")
            self.oscp_to_state_est(rec_data[0], rec_data[1])

    def oscp_to_state_est(self, cp_group_id, value):
        for element in self.controller.grid_wrapper.get_grid_elements(element_type="storage"):
            if element.get_property("type").get_value() == "CP":
                if element.get_generic_value('name') == cp_group_id: #'CP_1':
                    grid_value = element.get_measurement("active_power")
                    self.logger.debug(f"{cp_group_id, value, grid_value.to_dict()}")
                    measurement = PowerOwlMeasurement(grid_value, value, 12, self.controller.grid_wrapper.virtual_time.time())
                    for estimator_info in self.controller.grid_wrapper._state_estimators.values():
                        estimator_info["estimator"].measure(measurement)
                    break

    def run(self):
        self.logger.info(f"Running EVAtks Script {self.times}")

        self.rec_oscp_to_state_est_thread = Thread(target = self.rec_oscp_to_state_est, daemon=True)
        self.rec_oscp_to_state_est_thread.start()

        while not self.controller.wattson_client.is_started:
            sleep(1)

        self.start_datetime = self.controller.wattson_time.start_datetime_local(WattsonTimeType.SIM)
        self.start_timestamp = self.start_datetime.timestamp()

        sim_clock_time = self.controller.wattson_time.start_datetime_local(WattsonTimeType.SIM)
        sim_clock_time = pd.to_datetime(sim_clock_time).tz_localize(None).timestamp()
        sim_time_iso = datetime.datetime.fromtimestamp(sim_clock_time).isoformat()

        self.logger.info(f"Simulation started at {self.start_datetime.isoformat(), sim_time_iso}")
        self.start_datetime = datetime.datetime.fromtimestamp(sim_clock_time)
        self.start_timestamp = sim_clock_time


        #{'type': 'storage', 'index': 2, 'attributes': 
        # {<GridValueContext.PROPERTY: 'PROPERTY'>: {'minimum_reactive_power': None Var, 'maximum_reactive_power': None Var, 'minimum_active_power': -869030.7799999997 W, 
        #   'maximum_active_power': 869030.7799999997 W, 'minimum_charge': 200.0 Wh, 'maximum_charge': 5004399.999999999 Wh, 'nominal_power': nan VA, 'controllable': False, 
        #   'type': 'CP', 'bus': <powerowl.layers.powergrid.elements.bus.Bus object at 0x7f7f1ffc1ab0>}, 
        # <GridValueContext.CONFIGURATION: 'CONFIGURATION'>: {'state_of_charge': 0 %, 'target_reactive_power': 0.0 Var, 'target_active_power': 0.0 W, 'cos_phi': None, 
        #   'in_service': True, 'scaling': 0.9467879327564719}, 
        # <GridValueContext.MEASUREMENT: 'MEASUREMENT'>: {'state_of_charge': 0.0 %, 'reactive_power': 0.0 Var, 'active_power': 0.0 W}, 
        # <GridValueContext.ESTIMATION: 'ESTIMATION'>: {'state_of_charge': 0 %, 'reactive_power': nan Var, 'active_power': nan W}, 
        # <GridValueContext.GENERIC: 'GENERIC'>: {'observable': True, 'name': 'CP_1'}}}
        # for element in self.controller.remote_grid_model.get_elements_by_type("storage"):
        #     if type(element) == Storage and element.get_property("type").get_value() == "CP":
        #         self.logger.info(f"EVAtks {element.to_dict()}")
                
        #         rem_val = self.controller.remote_grid_model.get_grid_value_by_identifier(element.get_identifier()+".CONFIGURATION.scaling")
        #         self.logger.info(f"EVAtks {rem_val.to_dict()}")
        #         if not rem_val.set_value(0.5):
        #             self.logger.info(f"EVAtks error setting...")
        #         break
        
        # #self.controller.sync()
        # for element in self.controller.remote_grid_model.get_elements_by_type("storage"):
        #     if type(element) == Storage and element.get_property("type").get_value() == "CP":
        #         self.logger.info(f"EVAtks {element.to_dict()}")
                
        #         rem_val = self.controller.remote_grid_model.get_grid_value_by_identifier(element.get_identifier()+".CONFIGURATION.scaling")
        #         self.logger.info(f"EVAtks {rem_val.to_dict()}")
        #         break


        ext_ip = str(self.controller.wattson_client_query_socket).replace("tcp://", "").split(":")[0]
        #ext_ip="127.0.0.1"
        MESSAGE = "test"
        data = self.send_and_recv(ext_ip, MESSAGE, retry=15)
        self.ext_ip = ext_ip


        self._sleep_next(rnd_time=False, do_pre_send=False) #0
        self.logger.info("Getting Seed #0")
        # Set Active Power to 30%
        self.logger.info("getSeed")
        cpo_i = [i for i in self.controller.wattson_client.get_remote_network_node("CPO_0").get_interfaces() if i.is_management][0]
        cpo_ip = str(cpo_i.get_ip_address())
        MESSAGE = "getSeed"
        data = self.send_and_recv(cpo_ip, MESSAGE, retry=15)
        cpo_seed = pickle.loads(zlib.decompress(data))
        comb_seed = float(str(cpo_seed) + str(self.attack_seed))
        self.logger.info(f"getSeed {cpo_seed, cpo_ip} with {self.attack_seed=} -> {comb_seed=}")
        Random.set_base_seed(comb_seed)
        self.cpo_ip = cpo_ip

        self._sleep_next(rnd_time=False, do_pre_send=False) #1
        self.logger.info("Getting CPs #1")
        cpip_ids=defaultdict(lambda: list()) #cp ip to list of cp_ids
        cp_map=dict() #cp_id to node and ip
        for cp_n in [n for n in self.controller.wattson_client.get_remote_network_nodes() if n.entity_id.startswith("NodeCP_")]:
            cp_i = [i for i in cp_n.get_interfaces() if i.is_management][0]
            cp_ip = str(cp_i.get_ip_address())            

            self.logger.info(f"sending to {cp_n.entity_id}")
            MESSAGE = "getCPs"
            data = self.send_and_recv(cp_ip, MESSAGE, retry=15)
            cp_list = pickle.loads(zlib.decompress(data))
            self.logger.info(f"received message: {cp_ip, len(cp_list)}")
            for cp in cp_list:
                cpip_ids[cp_ip].append(cp)
                if cp in cp_map:
                    self.logger.error(f"dup cp {cp_n.entity_id, cp, cp_map[cp]}")
                else:
                    cp_map[cp] = {"cp_n":cp_n, "cp_ip":cp_ip}

        self.cpip_ids = cpip_ids
        self.cp_map = cp_map
        self.pool = ThreadPoolExecutor(len(self.cpip_ids.keys()))

        rand = Random.get_instance("ev_atks")
        self.rand = rand

        for x in range(self.atk_cycles_actual):
            atk_cps = self.rand.choice(sorted(list(cp_map.keys())), size=round(self.attack_p * len(cp_map.keys())), replace=False)
            self.logger.info(f"atk_cps entity_id: {sorted(set(cp_map[cp]['cp_n'].entity_id for cp in atk_cps))}")
            self.logger.info(f"atk_cps: {len(atk_cps), atk_cps}")
            self.atk_cps = atk_cps

            atk_dur_h = x+1 #*6 *365
            atk_dur_h = min(atk_dur_h, 4) #max 4 hours...

            atk_nums=[n for n in range(self.num_of_atks)]
            self.rand.shuffle(atk_nums)

            for atk_num in atk_nums:
                self.logger.info(f"atk_num: {atk_num}")
                self._sleep_next() #2
                if self.include_attacks:
                    self.select_pwr_atk(atk_num, atk_dur_h)


        self._sleep_next(rnd_time=False, do_pre_send=False) #12
        self.logger.info("Sending shutdown request")
        self.controller.wattson_client.request_shutdown()

        self.pool.shutdown()
        self.logger.info("Script completed")
        return
    

    def pickle_send_and_recv(self, to_ip, pwr_atk):
        pwr_atk_msg=zlib.compress(pickle.dumps(pwr_atk))
        data = self.send_and_recv(to_ip, pwr_atk_msg)
        return data

    def delayed_send_and_recv(self, to_ip, msg, retry=3, recv=True, timeout=5):
        s_time = self.rand.rand()
        #self.logger.info(f"delayed_send_and_recv: {s_time=}")
        sleep(s_time)
        self.send_and_recv(to_ip, msg, retry=retry, recv=recv, timeout=timeout)

    def run_tracert(self, use_cp_id="NodeCP_12"):
        all_cps=[]
        cp_ips=[]
        for cp, cp_d in self.cp_map.items():
            all_cps.append(cp_d["cp_ip"])
            if cp_d["cp_n"].entity_id == use_cp_id:
                cp_ips.append(cp_d["cp_ip"])
        msg_i=100
        while msg_i > 0:
            for cp_ip in cp_ips:
                msg_i-=1
                MESSAGE="tracert"
                data = self.send_and_recv(cp_ip, MESSAGE, retry=15, do_log=False)
                output_l = data.decode("utf-8").splitlines()
                self.logger.info(f"tracert result: {cp_ip, len(cp_ips), len(all_cps), output_l}")
                if msg_i == 0:
                    break
        self.controller.wattson_client.request_shutdown()
        self.pool.shutdown()
        # p = Popen(['tracepath', self.cpo_ip], stdin=PIPE, stdout=PIPE, stderr=PIPE)
        # output, err = p.communicate()
        # output_l = output.decode("utf-8").splitlines()
        # self.logger.info(f"{output_l=}")

    def select_pwr_atk(self, atk_num, atk_dur_h): 
        if self.case == "test" and not self.run_tracert_done:
            self.run_tracert_done=True
            self.run_tracert_thread = Thread(target = self.run_tracert, daemon=True)
            self.run_tracert_thread.start()

        if atk_num == 0:
            self.logger.info("Sending pwr_atk request #2")
            self.pwr_atk(atk_type = "MAD", set_to=1, add=1, fac=99, start_time=None, length=15 * 4 * atk_dur_h)

        elif atk_num == 1:
            self.logger.info("Sending pwr_atk request #3")
            self.pwr_atk(atk_type = "FDI", set_to=0, length=15 * 4 * atk_dur_h)

        elif atk_num == 2:
            self.logger.info("Sending pwr_atk request #4")
            self.pwr_atk(atk_type = "FDI", add=1, fac=5, ignore_limits=True, length=15 * 4 * atk_dur_h)

        elif atk_num == 3:
            self.logger.info("Sending pwr_atk request #5")
            # cur_sim_clock_time = self.controller.wattson_time.to_local_datetime(WattsonTimeType.SIM)
            # cur_sim_clock_time = pd.to_datetime(cur_sim_clock_time).tz_localize(None)
            # cur_sim_clock_time = cur_sim_clock_time + pd.Timedelta(hours=self.pre_send_atk_msg_h)
            start_time_p = self.pwr_atk(atk_type = "MAD_pre_FDI", set_to=1, add=1, fac=99,    start_time=None, length=15 * 4 * atk_dur_h) #mad attack
            self.pwr_atk(atk_type = "FDI_post_MAD", fac=1, ignore_limits=False, start_time=start_time_p, length=15 * 4 * atk_dur_h) #and report normal

        elif atk_num == 4:
            self.logger.info("Sending pwr_atk request #6")
            self.pwr_atk(atk_type = "MAD", set_to=0, length=15 * 4 * atk_dur_h)

        elif atk_num == 5:
            self.logger.info("Sending pwr_atk request #7")
            self.pwr_atk(atk_type = "FDI", set_to=1, add=1, fac=99, ignore_limits=False, length=15 * 4 * atk_dur_h)

        elif atk_num == 6:
            self.logger.info("Sending pwr_atk request #8")
            self.pwr_atk(atk_type = "MAD", fac=2, start_time=None, length=15 * 4 * atk_dur_h)

        elif atk_num == 7:
            self.logger.info("Sending pwr_atk request #9")
            start_time_p = self.pwr_atk(atk_type = "MAD_pre_FDI", set_to=0,     start_time=None, length=15 * 4 * atk_dur_h) #mad reduction attack
            self.pwr_atk(atk_type = "FDI_post_MAD", fac=2, ignore_limits=False, start_time=start_time_p, length=15 * 4 * atk_dur_h) #and report increase

        elif atk_num == 8:
            self.logger.info("Sending pwr_atk request #10")
            start_time_p = self.pwr_atk(atk_type = "MAD_pre_FDI", set_to=1, add=1, fac=99,    start_time=None, length=15 * 4 * atk_dur_h) #mad attack
            self.pwr_atk(atk_type = "FDI_post_MAD", fac=0, ignore_limits=False, start_time=start_time_p, length=15 * 4 * atk_dur_h) #and report reduction

        elif atk_num == 9:
            self.logger.info("Sending pwr_atk request #11")
            #self.pwr_atk(atk_type = "MAD", fac=1.5, start_time=None, length=15 * 4 * atk_dur_h)
            start_time_p = self.pwr_atk(atk_type = "MAD_pre_FDI", fac=2, ignore_limits=True,    start_time=None, length=15 * 4 * atk_dur_h) #mad attack
            self.pwr_atk(atk_type = "FDI_post_MAD", fac=1, ignore_limits=False, start_time=start_time_p, length=15 * 4 * atk_dur_h) #and report normal
        
        else:
            self.logger.error(f"unknown attack {atk_num=}")

    #start_time as datetime; length in min; set_to/add in kw, fac as *
    #attacks always start from base value, i.e., subsequent atks override previous
    def pwr_atk(self, atk_type, set_to=None, add=0, fac=1, ignore_limits=False, start_time=None, length=15): 
        try:
            msg="pwr_atk"
            if ignore_limits and atk_type == "MAD":
                self.logger.warning(f"pwr_atk with ignore_limits and MAD atk_type.") #we assume that MAD attacks should be limited to the physical limits of CPs
            pwr_atk={"msg":msg,"atk_type":atk_type,"set_to":set_to,"add":add,"fac":fac,"ignore_limits":ignore_limits,"start":None,"end":None,"atk_cps":self.atk_cps}

            if start_time is None:
                cur_sim_clock_time = self.controller.wattson_time.to_local_datetime(WattsonTimeType.SIM)
                cur_sim_clock_time = pd.to_datetime(cur_sim_clock_time).tz_localize(None)
                start_time = cur_sim_clock_time
                start_time = start_time + pd.Timedelta(hours=self.pre_send_atk_msg_h)
            pwr_atk["start"] = start_time
            pwr_atk["end"] = pwr_atk["start"] + pd.Timedelta(minutes=length)
            self.logger.info(f"pwr_atk {pwr_atk, self.ext_ip}")
            pool_res = []
            if atk_type in ["MAD", "MAD_pre_FDI"]:
                # pwr_atk_msg=zlib.compress(pickle.dumps(pwr_atk))
                # data = self.send_and_recv(self.ext_ip, pwr_atk_msg) #changes simulated pwr consumption (via CPPowerProfileProviderInterface)
                # data = self.pickle_send_and_recv(self.ext_ip, pwr_atk)
                pool_res.append(self.pool.submit(self.pickle_send_and_recv, self.ext_ip, pwr_atk))

            if atk_type in ["MAD", "FDI", "FDI_post_MAD"]:
                for cp_ip in sorted(self.cpip_ids.keys()):
                    cpids = self.cpip_ids[cp_ip]
                    pwr_atk_cp = deepcopy(pwr_atk)
                    pwr_atk_cp["atk_cps"] = []
                    for _atk_cp in self.atk_cps:
                        if _atk_cp in cpids:
                            pwr_atk_cp["atk_cps"].append(_atk_cp)
                    if len(pwr_atk_cp["atk_cps"]) > 0:
                        # pwr_atk_cp_msg=zlib.compress(pickle.dumps(pwr_atk_cp))
                        # data = self.send_and_recv(cp_ip, pwr_atk_cp_msg) #changes reported (via OCPP) pwr consumption
                        pool_res.append(self.pool.submit(self.pickle_send_and_recv, cp_ip, pwr_atk_cp))
            for r in pool_res:
                data = r.result()


                # cp_ips = set(self.cp_map[_atk_cp]["cp_ip"] for _atk_cp in self.atk_cps)
                # for cp_ip in sorted(cp_ips): #cp_c
                #     data = self.send_and_recv(cp_ip, pwr_atk_msg) #changes reported (via OCPP) pwr consumption
        except Exception as e:
            self.logger.error(f"pwr_atk error {e=}")
            self.logger.error(traceback.format_tb(e.__traceback__))
            pass
        return start_time


    def send_and_recv(self, to_ip, msg, retry=3, recv=True, timeout=5, do_log=True):
        if type(msg) == str:
            msg = zlib.compress(pickle.dumps({"msg":msg}))
        if to_ip not in self.sock:
            self.sock[to_ip] = socket.socket(socket.AF_INET, # Internet
                                 socket.SOCK_DGRAM) # UDP
        sock = self.sock[to_ip]
        sock.settimeout(timeout)
        try:
            UDP_PORT = 5005
            if do_log:
                self.logger.info(f"send_and_recv snd {len(msg) if len(msg) > 10 else str(msg), str(to_ip), UDP_PORT}")
            EVChargingExtension.send_to_and_check(sock, msg, (to_ip, UDP_PORT), self.logger)

            if not recv:
                return None
            data, from_ip = sock.recvfrom(MAX_MSG_SIZE)
            if to_ip != from_ip[0]:
                raise Exception(f"send_and_recv to_ip != from_ip {to_ip, from_ip}")
            if do_log:
                self.logger.info(f"send_and_recv rec {len(data) if len(data) > 10 else str(data), from_ip}")
        except socket.timeout as et:
            self.logger.info(f"send_and_recv timeout {retry, et}")
            if retry > 0:
                data = self.send_and_recv(to_ip, msg, retry=retry-1)
            else:
                data=None
        except BlockingIOError as et:
            self.logger.info(f"send_and_recv BlockingIOError {retry, et}")
            if retry > 0:
                data = self.send_and_recv(to_ip, msg, retry=retry-1)
            else:
                data=None
        except Exception as e: #BlockingIOError
            self.logger.error(f"send_and_recv Exception {e}")
            raise e

        sock.settimeout(None)
        return data

    def _sleep_next(self, rnd_time=True, do_pre_send=True):
        if self.step >= len(self.times):
            self.logger.error(f"no next atk {self.step, len(self.times)}.")
            return False
        atk_time = self.times[self.step]

        start_sim_clock_time = self.start_datetime
        atk_time_sim = start_sim_clock_time + atk_time[0]
        
        if rnd_time:
            atk_time_h = atk_time[1] + self.rand.randint(low=-1,high=2)
            atk_time_m = atk_time[2] + self.rand.randint(low= 0,high=60-atk_time[2])
        else:
            atk_time_h = atk_time[1]
            atk_time_m = atk_time[2]
        atk_time_sim = atk_time_sim.replace(hour=atk_time_h, minute=atk_time_m)
        if do_pre_send:
            atk_time_sim = atk_time_sim - pd.Timedelta(hours=self.pre_send_atk_msg_h)

        cur_sim_clock_time = self.controller.wattson_time.to_local_datetime(WattsonTimeType.SIM)
        cur_sim_clock_time = pd.to_datetime(cur_sim_clock_time).tz_localize(None)

        sim_clock_time_ts = cur_sim_clock_time.timestamp()
        atk_time_sim_ts = atk_time_sim.timestamp()
        offset = atk_time_sim_ts - sim_clock_time_ts
        offset_in_sim = offset / self.controller.wattson_time.speed
        if offset_in_sim < 0:
            offset_in_sim = 0
        self.logger.info(f"next atk at {datetime.datetime.fromtimestamp(atk_time_sim_ts).isoformat()} (current = {datetime.datetime.fromtimestamp(sim_clock_time_ts).isoformat()}, waiting offset_in_sim = {offset_in_sim}).")

        sleep(offset_in_sim)
        self.step += 1