import base64
import pickle
import zlib
from wattson.services.deployment import PythonDeployment
import threading
from wattson.util import get_logger
from wattson.hosts.cp.cp_service import CPService

#sudo python3 -m wattson wattson-ev/scenarios/powerowl_example

class CPDeployment(PythonDeployment):
    def __init__(self, configuration):
        super().__init__(configuration)
        self.restart_event = threading.Event()
        self.app=None

        self.config = configuration
        self.mtu_ip = self.config["mtu_ip"]
        self.management_ip = self.config["ip"]
        self.cp_id = self.config["cp_id"]
        self.cpo_ip = self.config["cpo_ip"]
        self.cp_data = self.config["cp_data"]
        self.cp_data = base64.b64decode(self.cp_data.encode("utf-8"))
        self.cp_data = pickle.loads(zlib.decompress(self.cp_data))
        self.scaling = self.config["scaling"]
        self.cp_node_id = self.config["cp_node_id"]
        self.min_sec_between_meter = float(self.config["min_sec_between_meter"])

        self.wattson_client_query_socket = self.config.get("wattson_client_query_socket")
        self.wattson_client_publish_socket = self.config.get("wattson_client_publish_socket")

        # self.scripts = self.config.get("scripts", [])
        # self.scenario_path = self.config.get("scenario_path")
        # self.export_config = self.config.get("export_config", {})
        # if "folder" not in self.export_config:
        #     self.export_config["folder"] = self.artifacts_dir
        # self.statistics = self.config.get("statistics", {})

        self.logger = get_logger("CPService", "CPService")
        self.logger.debug(f"__init__ CPService {self.config}")

    def restart(self):
        self.restart_event.set()

    def start(self):
        self.logger.info("Starting CPService")
        self.app = CPService(host_ip=self.management_ip,
                             cp_id=self.cp_id,
                             cpo_ip=self.cpo_ip,
                             cp_data=self.cp_data,
                             mtu_ip=self.mtu_ip,
                             scaling=self.scaling,
                             cp_node_id=self.cp_node_id,
                             min_sec_between_meter=self.min_sec_between_meter,
                             wattson_client_query_socket=self.wattson_client_query_socket,
                             wattson_client_publish_socket=self.wattson_client_publish_socket)
        self.app.start() 

    def stop(self):
        if self.app:
            self.app.stop()