from wattson.services.deployment import PythonDeployment
import threading
from wattson.util import get_logger
from wattson.hosts.cp.cpo_service import CPOService

#sudo python3 -m wattson wattson-ev/scenarios/powerowl_example

class CPODeployment(PythonDeployment):
    def __init__(self, configuration):
        super().__init__(configuration)
        self.restart_event = threading.Event()
        self.app=None

        self.config = configuration
        self.mtu_ip = self.config["mtu_ip"] #mtu mgmt iface ip
        self.dso_ip = self.config["dso_ip"] #mtu actual iface ip
        self.management_ip = self.config["ip"] #"!management_ips.ctrl"
        self.all_cp_ids = self.config["cp_ids"]
        self.cpo_node_id = self.config["cpo_node_id"]
        self.oscp_interval_m = self.config["oscp_interval_m"]

        self.wattson_client_query_socket = self.config.get("wattson_client_query_socket")
        self.wattson_client_publish_socket = self.config.get("wattson_client_publish_socket")

        # self.scripts = self.config.get("scripts", [])
        # self.scenario_path = self.config.get("scenario_path")
        # self.export_config = self.config.get("export_config", {})
        # if "folder" not in self.export_config:
        #     self.export_config["folder"] = self.artifacts_dir
        # self.statistics = self.config.get("statistics", {})

        self.logger = get_logger("CPOService", "CPOService")
        self.logger.info(f"__init__ CPService {self.config}")

    def restart(self):
        self.restart_event.set()

    def start(self):
        self.logger.info("Starting CPOService")
        self.app = CPOService(mgmt_ip=self.management_ip,
                                  mtu_ip=self.mtu_ip,
                                  dso_ip=self.dso_ip,
                                  all_cp_ids=self.all_cp_ids,
                                  cpo_node_id=self.cpo_node_id,
                                  oscp_interval_m=self.oscp_interval_m,
                                  wattson_client_query_socket=self.wattson_client_query_socket,
                                  wattson_client_publish_socket=self.wattson_client_publish_socket)
        self.app.start()

    def stop(self):
        if self.app:
            self.app.stop()