
#based on https://github.com/NOWUM/pyoscp/blob/master/start.py

from copy import deepcopy
import datetime
import logging
import os
import json
import pickle
import secrets
import socket
from threading import Thread
import time
import zlib
import pandas as pd

import requests
from oscp import createBlueprint
from flask import Flask, redirect
from oscp.RegistrationManager import RegistrationDictMan, createOscpHeader
from oscp.json_models import UpdateGroupMeasurements
from scenarios.powerowl_example.extensions.ev_charging_extension import MAX_MSG_SIZE, EVChargingExtension
from wattson.time.wattson_time_type import WattsonTimeType
from wattson.util import get_logger
import concurrent.futures
import traceback

try:
    from cheroot.wsgi import Server as WSGIServer, PathInfoDispatcher
except ImportError:
    from cherrypy.wsgiserver import CherryPyWSGIServer as WSGIServer, WSGIPathInfoDispatcher as PathInfoDispatcher


class CPO_OSCP:
    def __init__(self, cpo_ocpp_service, DST_PORT=5001, DST_IP="127.0.0.1", HOST_PORT=5002, HOST_IP="127.0.0.1", HOST_URL=None, BASE_TOKEN_FOR_DSO=secrets.token_urlsafe(32)):
        
        self.logger = get_logger("CPO_OSCP")

        self.cpo_ocpp_service = cpo_ocpp_service
        self.DST_PORT=DST_PORT
        self.DST_IP=DST_IP
        self.HOST_PORT=HOST_PORT
        self.HOST_IP=HOST_IP
        self.HOST_URL = 'http://'+self.HOST_IP+':'+str(self.HOST_PORT) if HOST_URL is None else HOST_URL
        self.BASE_TOKEN_FOR_DSO = BASE_TOKEN_FOR_DSO

        self.RUN=True
        #init():
        self.regman = None
        self.server = None
        self.fpm = None
        self.BASE_TOKEN_FOR_CPO1=None
        #start_server():
        self.thread = None
        self.thread2 = None
        #start_server_threaded(self):
        self.server_thread = None
        #init_threaded(self):
        self.init_thread = None


    class FlexibilityProviderManager():

        def __init__(self, cpo_server, registrationmanager: RegistrationDictMan):
            self.stuff = []
            self.version_urls = [{'version': '2.0', 'base_url': cpo_server.HOST_URL + '/oscp/fp/2.0'}]
            self.registrationmanager = registrationmanager
            self.cpo_server = cpo_server
            self.th_exec = None
            self.sends = []

        def start(self):
            s = {"token":self.cpo_server.BASE_TOKEN_FOR_DSO, "version_url":self.version_urls}
            res = requests.post("http://"+self.cpo_server.DST_IP+":"+str(self.cpo_server.DST_PORT)+"/oscp/cp/2.0/register", json=s, headers=createOscpHeader(self.cpo_server.BASE_TOKEN_FOR_CPO1)) #exchanged OOB
            self.cpo_server.logger.info(res)
            self.cpo_server.logger.info(res.text)

        def get_sim_time(self):
            sim_clock_time = self.cpo_server.cpo_ocpp_service.wattson_time.to_local_datetime(WattsonTimeType.SIM) #same as ProfileLoader
            sim_clock_time = pd.to_datetime(sim_clock_time).tz_localize(None).timestamp()
            sim_clock_time = datetime.datetime.fromtimestamp(sim_clock_time)
            return sim_clock_time
        
        
        def send_mes_th(self, base_url, token, value, group, sim_clock_time, init_sim_clock_time, sleep_time=0.5):
            m = {"value": value, "phase": "ALL", "unit": "WH", "energy_type": "TOTAL", "direction": "NET", "measure_time": str(sim_clock_time), "initial_measure_time": str(init_sim_clock_time)}
            s = {"group_id":group, "measurements": [m]}

            try:
                res = requests.post(base_url+"/update_group_measurements", json=s, headers=createOscpHeader(token))
                if res.status_code != 204:
                    self.cpo_server.logger.error(f"{base_url, token, value, group, sim_clock_time, res}")
            except Exception as e:
                self.cpo_server.logger.error(f"Exception {e=} w/ {base_url, token, value, group, sim_clock_time}")
                res=None

            #time.sleep(sleep_time)
            return (group, res)

        def send_mes(self):
            base_url, token = self.registrationmanager._url_by_token(self.cpo_server.BASE_TOKEN_FOR_DSO)
            self.cpo_server.logger.info(f"/update_group_measurements for base_url, token: {base_url, token}")

            if token is not None:
                values, group_times, sim_time = self.cpo_server.cpo_ocpp_service.get_group_measurements()
                
                if self.th_exec is None:
                    max_workers = max(1, len(values.keys()))
                    self.th_exec = concurrent.futures.ThreadPoolExecutor(max_workers=max_workers)
                else:
                    self.cpo_server.logger.info(f"sends: {[fut.result() for fut in self.sends]}")
                    self.sends=[]

                if self.th_exec._max_workers < len(values.keys()):
                    self.th_exec.shutdown()
                    self.th_exec = concurrent.futures.ThreadPoolExecutor(max_workers=len(values.keys()))
                
                sim_clock_time=None
                #sim_clock_time = self.get_sim_time()
                for group, value in values.items():
                    try:
                        #sim_clock_time = max(group_times[group])
                        sim_clock_time = sim_time
                        init_sim_clock_time = min(group_times[group])
                        if True: #value != 0:     
                            #self.send_mes_th(base_url, token, values)
                            send_fut = self.th_exec.submit(self.send_mes_th, base_url, token, value, group, sim_clock_time, init_sim_clock_time)
                            self.sends.append(send_fut)
                    except Exception as e:
                        self.cpo_server.logger.error(f"send_mes: {e=}, {values=}, {group_times=}")


                #self.cpo_server.logger.info(res.text)


    class CapacityOptimizerManager():
        def __init__(self):
            pass
    class EnergyProviderManager():
        def __init__(self):
            pass
    class FlexibilityManager(object):
        def __init__(self):
            pass
    class CapacityProviderManager():
        def __init__(self):
            pass



    def get_token_oob(self, UDP_IP="127.0.0.1", UDP_PORT = 5005, retry=15):
        UDP_IP = UDP_IP
        UDP_PORT = UDP_PORT
        self.logger.info(f"get_token_oob sock send to DSO: {str(UDP_IP), UDP_PORT}")

        sock = socket.socket(socket.AF_INET, # Internet
                            socket.SOCK_DGRAM) # UDP
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        sock.settimeout(10)
        #sock.setblocking(False)

        m="getToken"
        msg = zlib.compress(pickle.dumps({"msg": m}))
        EVChargingExtension.send_to_and_check(sock, msg, (UDP_IP, UDP_PORT), self.logger)

        try:
            data, from_ip = sock.recvfrom(MAX_MSG_SIZE)
            if UDP_IP != from_ip[0]:
                raise Exception(f"get_token_oob UDP_IP != from_ip {UDP_IP, from_ip}")
            self.logger.info(f"get_token_oob rec {len(data) if len(data) > 10 else str(data), from_ip}")
        except socket.timeout as et:
            self.logger.info(f"get_token_oob timeout {retry, et}")
            if retry > 0:
                data = zlib.compress(pickle.dumps(self.get_token_oob(UDP_IP, UDP_PORT, retry=retry-1)))
            else:
                data = zlib.compress(pickle.dumps(None))

        token = pickle.loads(zlib.decompress(data))
        self.logger.info(f"get_token_oob token: {UDP_IP, token}")
        return token

    
    def init(self, dso_oob_ip="127.0.0.1", dso_oob_port = 5005):
        app = Flask(__name__)

        @app.route('/', methods=['POST', 'GET'])
        def home():
            return redirect('/oscp/ui')

        version_urls = [{'version': '2.0', 'base_url': self.HOST_URL + '/oscp/fp/2.0'}]
        regman = RegistrationDictMan(version_urls, filename="./endpoints_fp.json")


        self.BASE_TOKEN_FOR_CPO1=self.get_token_oob(dso_oob_ip, dso_oob_port)

        # inject dependencies here
        cpm = self.CapacityProviderManager()
        com = self.CapacityOptimizerManager()
        fpm = self.FlexibilityProviderManager(self, regman) #needs self.cpo_server.BASE_TOKEN_FOR_CPO1
        epm = self.EnergyProviderManager()

        injected_objects = {'energyprovider': epm,
                            'capacityprovider': cpm,
                            'capacityoptimizer': com,
                            'flexibilityprovider': fpm,
                            'registrationmanager': regman}

        config = 'oscp.json'
        if os.path.exists(config):
            self.logger.info(f'reading config file {config}')
            with open(config, 'r') as f:
                conf = json.load(f)
                regman._updateService(conf['token'], None, None)
        else:
            self.logger.info(f'config file {config} does not exist')
        regman._updateService(self.BASE_TOKEN_FOR_DSO, None, None)


        # the injected_objects are used to route requests from the namespace to the
        # logic containing classes like ForecastManager
        blueprint = createBlueprint(injected_objects, actor='fp') #CPO

        # using a blueprint allows us to initialize everything without an app context
        # this is cleaner, as the oscp module never needs access to the actual Flask server
        app.register_blueprint(blueprint)


        d = PathInfoDispatcher({'/': app})
        server = WSGIServer((self.HOST_IP, self.HOST_PORT), d)

        self.regman = regman
        self.server = server
        self.fpm = fpm
        return


    def start_fpm(self, wait_s):
        _,_ = self.regman._url_by_token(self.BASE_TOKEN_FOR_DSO)
        time.sleep(wait_s)
        self.fpm.start()

    def send_mes(self, wait_s, interval):
        time.sleep(wait_s)
        while self.RUN:
            try:
                self.fpm.send_mes()
                time.sleep(interval)
            except Exception as e:
                self.logger.error(f"error in send_mes {e=}")
                self.logger.error(traceback.format_tb(e.__traceback__))
                time.sleep(interval)
                

    def start_server(self, init_delay, init_meas_delay, meas_interval):
        self.regman.start()

        self.thread = Thread(target = self.start_fpm, args = (init_delay, ))
        self.thread.start()
        self.thread2 = Thread(target = self.send_mes, args = (init_delay+init_meas_delay, meas_interval, ))
        self.thread2.start()

        try:
            self.server.start()
        except KeyboardInterrupt:
            self.server.stop()

        self.RUN=False

        self.thread.join()
        self.thread2.join()

        if self.fpm.th_exec is not None:
            self.fpm.th_exec.shutdown()

        self.regman.stop()

    def start_server_threaded(self, init_delay=20, init_meas_delay=5, meas_interval=15):
        self.logger.info(f"start_server_threaded: {init_delay, init_meas_delay, meas_interval}")
        self.server_thread = Thread(target = self.start_server, args = (init_delay, init_meas_delay, meas_interval, ))
        self.server_thread.start()

    def init_threaded(self, dso_oob_ip="127.0.0.1", dso_oob_port = 5005):
        self.init_thread = Thread(target = self.init, args = (dso_oob_ip, dso_oob_port, ))
        self.init_thread.start()

if __name__ == '__main__':
    cpo_server = CPO_OSCP()
    cpo_server.init_threaded() #get token
    cpo_server.init_thread.join()
    cpo_server.start_server_threaded()

# if __name__ == '__main__':
#     cpo_server = CPO_OSCP()
#     regman,server,fpm = cpo_server.init()
#     regman.start()
#     print(regman.getEndpoints())


#     thread = Thread(target = cpo_server.start_fpm, args = (1, ))
#     thread.start()
#     thread2 = Thread(target = cpo_server.send_mes, args = (2, 1, ))
#     thread2.start()
    
#     # app.run(port=self.HOST_PORT)
#     try:
#         server.start()
#     except KeyboardInterrupt:
#         server.stop()

#     cpo_server.RUN=False

#     thread.join()
#     thread2.join()

#     regman.stop()