
#based on https://github.com/NOWUM/pyoscp/blob/master/start.py

from copy import deepcopy
import datetime
import logging
import os
import json
import pickle
import secrets
import socket
from threading import Thread
import time
import zlib
import pandas as pd

import requests
from oscp import createBlueprint
from flask import Flask, redirect
from oscp.RegistrationManager import RegistrationDictMan, createOscpHeader
from oscp.json_models import UpdateGroupMeasurements
from scenarios.powerowl_example.extensions.ev_charging_extension import MAX_MSG_SIZE, OSCP_INTERVAL_M, EVChargingExtension
from wattson.time.wattson_time_type import WattsonTimeType
from wattson.util import get_logger

try:
    from cheroot.wsgi import Server as WSGIServer, PathInfoDispatcher
except ImportError:
    from cherrypy.wsgiserver import CherryPyWSGIServer as WSGIServer, WSGIPathInfoDispatcher as PathInfoDispatcher


class DSO_OSCP:
    def __init__(self, HOST_PORT=5001, HOST_IP="127.0.0.1", HOST_URL=None, BASE_TOKEN_FOR_CPO1=secrets.token_urlsafe(32), mgmt_ip=None, wattson_time=None):
        
        self.logger = get_logger("DSO_OSCP")

        self.HOST_PORT=HOST_PORT
        self.HOST_IP=HOST_IP
        self.HOST_URL = 'http://'+self.HOST_IP+':'+str(self.HOST_PORT) if HOST_URL is None else HOST_URL
        self.BASE_TOKEN_FOR_CPO1 = BASE_TOKEN_FOR_CPO1
        self.mgmt_ip = mgmt_ip
        self.wattson_time=wattson_time

        self.RUN=True
        #init():
        self.regman = None
        self.server = None
        self.fpm = None
        #start_server_threaded(self):
        self.server_thread = None
        #oob_threaded(self):
        self.oob_thread = None

        self.sock = socket.socket(socket.AF_INET, # Internet
                            socket.SOCK_DGRAM) # UDP


    class FlexibilityProviderManager():
        def __init__(self):
            pass
    class CapacityOptimizerManager():
        def __init__(self):
            pass
    class EnergyProviderManager():
        def __init__(self):
            pass
    class FlexibilityManager(object):
        def __init__(self):
            pass

    class CapacityProviderManager():
        def __init__(self, dso_server):
            dso_server.logger.info(f'CapacityProviderManager started')
            self.dso_server = dso_server
            self.last_cp_g_meas = dict()

        def handleUpdateGroupMeasurements(self, payload, token):
            sim_time=None
            sim_speed_fac=OSCP_INTERVAL_M / 60 # 0.25 for 15 min
            if self.dso_server.wattson_time is not None:
                sim_clock_time = self.dso_server.wattson_time.to_local_datetime(WattsonTimeType.SIM) #same as ProfileLoader
                sim_time = pd.to_datetime(sim_clock_time).tz_localize(None).isoformat()

            self.dso_server.logger.info(f'handleUpdateGroupMeasurements {sim_time, sim_speed_fac, payload, token}')
            #handleUpdateGroupMeasurements ({'group_id': 'CP_3', 'measurements': [{'value': 1412, 'phase': 'ALL', 'unit': 'WH', 'energy_type': 'FLEXIBLE', 'direction': 'NET', 'measure_time': 'CP_9', 'initial_measure_time': 'CP_9'}]}, 'UApoXNb43t8835FZQ3nXCONl8wsp0XeoeyeZq1KFpyM')
            cp_group_id = payload['group_id']
            value = payload['measurements'][0]["value"]
            measure_time = payload['measurements'][0]["measure_time"]
            measure_time = pd.to_datetime(measure_time)
            if cp_group_id in self.last_cp_g_meas:
                last_value = self.last_cp_g_meas[cp_group_id][0]
                last_measure_time = self.last_cp_g_meas[cp_group_id][1]
                measure_time_diff = (measure_time - last_measure_time).total_seconds()
                time_fac = measure_time_diff / (60*60) #s to h
                value_diff = value - last_value #in wh
                if time_fac != 0:
                    send_val = value_diff / time_fac #wh to w
                else:
                    send_val=0
                self.dso_server.send_oscp_to_state_est(cp_group_id, send_val)
            self.last_cp_g_meas[cp_group_id] = (value, measure_time)



    def send_token_oob(self, UDP_IP="127.0.0.1", UDP_PORT = 5005):
        self.logger.info(f"send_token_oob sock bind: {str(UDP_IP), UDP_PORT}")

        sock = socket.socket(socket.AF_INET, # Internet
                            socket.SOCK_DGRAM) # UDP
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        sock.bind((str(UDP_IP), UDP_PORT))
        #sock.setblocking(False)

        while self.RUN:
            # try:
            #     data = await asyncio.wait_for(loop.sock_recv(sock, 1024), timeout=60)
            # except asyncio.TimeoutError:
            #     pass
            data, cpo_ip = sock.recvfrom(MAX_MSG_SIZE)
            #data = data.decode("utf-8")
            rec_data = pickle.loads(zlib.decompress(data))
            self.logger.info(f"received message: {rec_data, cpo_ip}")
            if rec_data["msg"] == 'getToken':
                self.logger.info(f"sending message: {self.BASE_TOKEN_FOR_CPO1, cpo_ip}")
                out_msg = zlib.compress(pickle.dumps(self.BASE_TOKEN_FOR_CPO1))
                EVChargingExtension.send_to_and_check(sock, out_msg, cpo_ip, self.logger)

            # data, addr = sock.recvfrom(1024) # buffer size is 1024 bytes
        self.logger.info("send_token_oob done.")

    def send_oscp_to_state_est(self, cp_group_id, value, UDP_PORT=50099,):
        self.logger.info(f"send_oscp_to_state_est: {str(self.mgmt_ip), UDP_PORT, cp_group_id, value}")
        msg = zlib.compress(pickle.dumps((cp_group_id, value)))
        EVChargingExtension.send_to_and_check(self.sock, msg, (self.mgmt_ip, UDP_PORT), self.logger)
    
    def init(self):
        app = Flask(__name__)

        @app.route('/', methods=['POST', 'GET'])
        def home():
            return redirect('/oscp/ui')

        version_urls = [{'version': '2.0', 'base_url': self.HOST_URL + '/oscp/cp/2.0'}]
        regman = RegistrationDictMan(version_urls, filename="./endpoints_cp.json")


        # inject dependencies here
        cpm = self.CapacityProviderManager(self)
        com = self.CapacityOptimizerManager()
        fpm = self.FlexibilityProviderManager()
        epm = self.EnergyProviderManager()

        injected_objects = {'energyprovider': epm,
                            'capacityprovider': cpm,
                            'capacityoptimizer': com,
                            'flexibilityprovider': fpm,
                            'registrationmanager': regman}

        config = 'oscp.json'
        if os.path.exists(config):
            self.logger.info(f'reading config file {config}')
            with open(config, 'r') as f:
                conf = json.load(f)
                regman._updateService(conf['token'], None, None)
        else:
            self.logger.info(f'config file {config} does not exist')
        regman._updateService(self.BASE_TOKEN_FOR_CPO1, None, None)


        # the injected_objects are used to route requests from the namespace to the
        # logic containing classes like ForecastManager
        blueprint = createBlueprint(injected_objects, actor='cp') #CPO

        # using a blueprint allows us to initialize everything without an app context
        # this is cleaner, as the oscp module never needs access to the actual Flask server
        app.register_blueprint(blueprint)


        d = PathInfoDispatcher({'/': app})
        server = WSGIServer((self.HOST_IP, self.HOST_PORT), d)

        self.regman = regman
        self.server = server
        self.fpm = fpm
        return


    def start_server(self):
        self.regman.start()

        try:
            self.server.start()
        except KeyboardInterrupt:
            self.server.stop()

        self.RUN=False

        self.regman.stop()

    def start_server_threaded(self):
        self.server_thread = Thread(target = self.start_server)
        self.server_thread.start()

    def oob_threaded(self, dso_oob_ip="127.0.0.1", dso_oob_port = 5005):
        self.oob_thread = Thread(target = self.send_token_oob, args = (dso_oob_ip, dso_oob_port, ), daemon=True)
        self.oob_thread.start()

if __name__ == '__main__':
    cpo_server = DSO_OSCP()
    cpo_server.init() #get token
    cpo_server.oob_threaded()
    # cpo_server.start_server_threaded()
    # cpo_server.server_thread.join()
    cpo_server.start_server()

# if __name__ == '__main__':
#     cpo_server = CPO_OSCP()
#     regman,server,fpm = cpo_server.init()
#     regman.start()
#     print(regman.getEndpoints())


#     thread = Thread(target = cpo_server.start_fpm, args = (1, ))
#     thread.start()
#     thread2 = Thread(target = cpo_server.send_mes, args = (2, 1, ))
#     thread2.start()
    
#     # app.run(port=self.HOST_PORT)
#     try:
#         server.start()
#     except KeyboardInterrupt:
#         server.stop()

#     cpo_server.RUN=False

#     thread.join()
#     thread2.join()

#     regman.stop()