
from concurrent.futures import ProcessPoolExecutor, ThreadPoolExecutor
from copy import deepcopy
from datetime import datetime
import json
import os
import re
from subprocess import PIPE, Popen
import time
from cycler import cycler
import matplotlib
from matplotlib.dates import DateFormatter
from numpy import mean
import numpy as np

from sklearn.metrics import RocCurveDisplay, auc, confusion_matrix, roc_auc_score, roc_curve
import yaml


with open("ids.conf", 'r') as f:
    conf = json.load(f)
    configs = conf["configs"]
    # print(conf)
    # exit()
    NUM_THREADS = conf["NUM_THREADS"]
    if "affinity_mask" in conf:
        affinity_mask = conf["affinity_mask"]
        pid = os.getpid()
        os.sched_setaffinity(pid, affinity_mask) 


os.environ["OMP_NUM_THREADS"] = str(NUM_THREADS)
os.environ["MKL_NUM_THREADS"] = str(NUM_THREADS)
os.environ["OPENBLAS_NUM_THREADS"] = str(NUM_THREADS)
os.environ["BLIS_NUM_THREADS"] = str(NUM_THREADS)

import argparse
import ast
from collections import defaultdict
import gzip
import pickle


from pathlib import Path
import pandas as pd
import matplotlib.pyplot as plt

import logging

from tqdm import tqdm
from load_data import clean_dataset_l, get_dataset_l, get_measurements_expo, print_data_df, get_cpo_ocpp_data, get_dso_oscp_data, get_file_dfs, get_measurements, get_estimations, plot_cpo_ocpp_data, plot_dso_oscp_data
from features_cpo import get_cpo_files, read_cpo_files, get_cpo_group_feat_dfs
from features_dso import get_dso_files, get_elems_attatched_to_bus, get_grid_pp, oscp_speed_workaround, read_dso_files, get_dso_group_feat_dfs, read_dso_measurements, read_dso_estimations
from features_aux import iter_feats, load_feats_CPg, load_feats_len, load_single_feat, prune_feats, save_feats, load_feats, get_grids_json, get_ts_sim_map, interpolate_sim_time, get_grid_measurements_from_export
from regression import eval_tuning, get_cp_group_eval_dicts, get_eval_dicts, get_prediction_dicts, get_regression_pred, get_regression_pred_conc, get_regression_pred_conc_kfold, optimize_regression
from features_clf import _interp, cross_val_clf, do_clf, do_get_clf_is_atk_conc, get_atks, get_clf_eval, get_clf_feat_dfs, get_clf_feat_file_dicts, get_clf_feat_for_conf, get_clf_is_atk, get_clf_is_atk_dfs, get_clf_result_output, get_clf_result_output_conc, get_eval_dicts_clf, get_param_grid_len, load_eval_dicts_clf, prune_clf_feats, save_clf_data, save_eval_dicts_clf, to_float_list

logger = logging.getLogger("WATTSON_EV_IDS")

# os.environ["QT_QPA_PLATFORM"] = "wayland"
#os.environ["QT_QPA_PLATFORM"] = "xcb"

# py ids.py -c=load_data -v -d test
# py ids.py -c=get_features_dso -l 4 -v -d test
# py ids.py -c=get_features_cpo -l 4 -v -d test
# py ids.py -c=train_reg_dso -r test -v -d test
# py ids.py -c=train_reg_cpo -r test -f no_cps no_grid -v -d test
# py ids.py -c=eval_tuning -v -d elaadnl
# py ids.py -c=do_pred_dso -s 5 -v -d elaadnl
# py ids.py -c=do_pred_cpo -s 5 -v -d elaadnl

# py ids.py -c=get_features_clf -v -d elaadnl

# py ids.py -c=optimize_clf -C LocalOutlierFactor -F no_reg only_norm -v -d elaadnl
# py ids.py -c=optimize_clf -C LocalOutlierFactor -F no_reg no_norm -v -d elaadnl
# py ids.py -c=optimize_clf -C LocalOutlierFactor -F only_norm -v -d elaadnl
# py ids.py -c=optimize_clf -C LocalOutlierFactor -F no_norm -v -d elaadnl


# py ids.py -c=do_clf_dso -v -d elaadnl_atk


for k in configs.keys():
    configs[k]["DataPointMapsDIR"] = "../scenarios/powerowl_example/data-points"
    configs[k]["OutDataDIR"] = "data/"+k
    if "atk" in k:
        sub_folders = [os.path.join(configs[k]["DIR"], name) for name in os.listdir(configs[k]["DIR"]) if os.path.isdir(os.path.join(configs[k]["DIR"], name)) if name.startswith("atk_")]
        configs[k]["DIR"] = sub_folders

for k in configs.keys():
    if "TRAIN_START_DATE" in configs[k]:
        configs[k]["TRAIN_START_DATE"] += " 00:00:00.0"
        configs[k]["VALIDATION_START_DATE"] += " 00:00:00.0"
        configs[k]["ATK_START_DATE"] += " 00:00:00.0"

dataset_choices = get_dataset_l(configs)

#missing measurements in /home/dk/git/wattson-artifacts/elaadnl/atk_4_1.0_powerowl_example; eg for data points 818 ie bus.7

pd.reset_option('^display.', silent=True)
DO_OSCP_WORKAROUND=True
USE_TEX=True

if __name__ == "__main__":
    parser = argparse.ArgumentParser(prog='ids.py', usage='%(prog)s [options]')
    parser.add_argument('-d', '--dataset', nargs='+',
                        choices=dataset_choices,
                        help='dataset selection')           # 
    parser.add_argument('-c', '--case',
                        choices=["test", 'load_data', "get_features_dso", "get_features_cpo",
                                 'plot_atks', 'plot_atks2', 'get_load_yml', "plot_base",
                                 "train_reg_dso",'train_reg_cpo', 
                                 "eval_tuning", "do_pred_dso", "do_pred_cpo", 
                                 "get_features_clf", "get_features_clf_cpo", "get_features_clf_dso",
                                 "optimize_clf", "optimize_clf_cpo", "optimize_clf_dso",
                                 "eval_tuning_clf", "print_eval_tuning_clf",
                                 "get_is_atk_dfs", "get_is_atk_dfs_cpo", "get_is_atk_dfs_dso",
                                 "do_clf", "do_clf_cpo", "do_clf_dso",
                                 "eval_clf_results1", "eval_clf_results2", "eval_clf_results3", "eval_clf_results4"],
                        help='evaluation case selection')      # 
    parser.add_argument('-f', '--features',
                        help="feature set selection",
                        default=["all"], nargs='+',
                        choices=["all", "no_cps", "no_cps_but_speed", 
                                 "no_grid", "no_grid_storage", "no_grid_line", "no_grid_sgen", "no_grid_bus_rel_to_pred", "add_bus_relations", ]+
                                 ["add_grid_load_expo_static_"+str(x) for x in range(10,110,10)]+
                                 ["add_grid_load_expo_rnd_"+str(x) for x in range(2,21,2)]+
                                 ["no_est", "no_meas", "only_pred_lag", "no_time_diff",                                  
                                 "only_norm", "only_norm_but_pred", "no_norm", "no_norm1", "no_norm2", "no_date1", "no_date2",
                                 "no_hour_date", "no_hour_only_date", "no_hour_day_date", "no_day_of_week_date", "no_day_is_work_date", "no_hour_balancing_date"],)
    parser.add_argument('-r', '--reg_type',
                        choices=['RandomForestRegressor', "DecisionTreeRegressor",'GradientBoostingRegressor', "HistGradientBoostingRegressor", "LinearSVR", "MLPRegressor", "test"],
                        help='regression algorithm selection (only for train_reg_dso/cpo)', default=None)      # 
    parser.add_argument('-s', '--shifts',
                        help="amount of (shifted) preditctions to generate (>=1) (only for do_pred_dso/cpo)",
                        type=int, default=1)
    parser.add_argument('-l', '--lags',
                        help="Amount of lags to add during feature extraction (only for get_features_dso/cpo)",
                        type=int, default=4)
    parser.add_argument('-C', '--clf_type',
                        choices=['LocalOutlierFactor', "OneClassSVM", "EllipticEnvelope", "IsolationForest", "test"],
                        help='classification/novelty detection algorithm selection', default=None)      # 
    parser.add_argument('-F', '--clf_features',
                        help="novelty detection feature set selection",
                        default=["all"], nargs='+',
                        choices=["all", "only_norm", "no_norm", "copy_reg", "no_reg", "reg_no_cps", "train_conta", "only_diff"]+
                                 ["set_1", "set_2", "set_3", "set_4", "set_5", "set_6", "set_7", "set_8", "set_9", "set_10", "set_11", "set_12", "set_13"]+
                                 ["set_31", "set_32", "set_33", "set_34", "set_35", "set_352"]+
                                 ["add_grid_load_expo_static_"+str(x) for x in range(10,110,10)]+
                                 ["add_grid_load_expo_rnd_"+str(x) for x in range(2,21,2)],)
    
    parser.add_argument('-A', '--atk_subset',
                        help="set considered attack subset",
                        choices=["mad", "fdi", "combo"],)
    
    parser.add_argument('-a', '--affinity_mask',
                        help="set CPU affinity_mask",
                        default=[], nargs='+',)
    
    parser.add_argument('-e', '--eval_fac',
                        help="Factor for evaluation attack size",
                        type=str)
    
    parser.add_argument('-n', '--noise',
                        help="Add noise to grid measurements for testing",
                        type=int)
    
    parser.add_argument(
        '-U', '--use_cpo_pred',
        help="Use CPO predictions at DSO.",
        action="store_const", dest="use_cpo_pred", const=True,
        default=False,
    )
    parser.add_argument(
        '-R', '--recursive_regression',
        help="Use recursive_regression instead of training multiple regression models per prediction distance.",
        action="store_const", dest="recursive_regression", const=True,
        default=False,
    )
    parser.add_argument(
        '-O', '--overwrite',
        help="Overwrite existing files.",
        action="store_const", dest="overwrite", const=True,
        default=False,
    )
    parser.add_argument(
        '-D', '--debug',
        help="Print lots of debugging statements",
        action="store_const", dest="loglevel", const=logging.DEBUG,
        default=logging.WARNING,
    )
    parser.add_argument(
        '-v', '--verbose',
        help="Be verbose",
        action="store_const", dest="loglevel", const=logging.INFO,
    )
    # parser.print_help()

    args = parser.parse_args()
    NUM_LAGS=args.lags
    logging.basicConfig(level=args.loglevel)
    args.dataset = clean_dataset_l(args.dataset, configs) #
    logger.info(f"{args.dataset=}, {args.case=}, {args.loglevel=}, {args.features=}, {args.reg_type=}, {args.shifts=}, {args.lags=}, {args.clf_type=}, {args.clf_features=}, {args.affinity_mask=}, {args.overwrite=}")


    if args.affinity_mask:
        pid = os.getpid()
        os.sched_setaffinity(pid, [int(a) for a in args.affinity_mask]) 

    file_dfs = get_file_dfs(args.dataset, configs)


    if args.case == "load_data":
        for dataset in args.dataset:
            print_data_df(file_dfs[dataset])
            file_df = file_dfs[dataset]
            DataPointMapsDIR = configs[dataset]["DataPointMapsDIR"]
            OutDataDIR = configs[dataset]["OutDataDIR"]

            cpo_tx_df_cps = get_cpo_ocpp_data(file_df)
            dso_tx_df_cp_dict = get_dso_oscp_data(file_df)

            #TODO: fix oscp data...
            # cp_dff = dso_tx_df_cp_dict["CP_11"]
            # fig, ax = plt.subplots(figsize=(16,9))
            # ax.plot(cp_dff.index, cp_dff["charge_speed"], label="charge_speed_lag_0") 
            # ax.plot(cpp_dff.index, cpp_dff["charge_speed"], label="charge_speed_lag_0 cpo") 
            # plt.legend()
            # plt.show()
            # plt.close()
            # exit()
        

            if logger.isEnabledFor(logging.DEBUG):
                for k,v in cpo_tx_df_cps.items():
                    plot_cpo_ocpp_data(v)
                plot_dso_oscp_data(dso_tx_df_cp_dict)
                plt.show()
                plt.close()

            clean_meas_df = get_measurements_expo(file_df, configs[dataset]["DIR"])
            # print(clean_meas_df)
            # clean_meas_df = get_measurements(file_df, DataPointMapsDIR)
            # print(clean_meas_df)
            # exit()
            est_df=None
            if False:
                est_df = get_estimations(file_df, drop_dups=False)

            Path(OutDataDIR).mkdir(parents=True, exist_ok=True)
            logger.info(f"Writing results to {OutDataDIR=} ...")
            for k,v in cpo_tx_df_cps.items():
                Path(OutDataDIR+"/"+k).mkdir(parents=False, exist_ok=True)
                for k2,v2 in v.items():
                    for cp_id, cp_df in v2:
                        cp_df.to_csv(OutDataDIR+"/"+k+"/cpo_ocpp_data_"+k2+"_"+cp_id+".csv.gz")
            Path(OutDataDIR+"/DSO").mkdir(parents=False, exist_ok=True)
            for k,v in dso_tx_df_cp_dict.items():
                v.to_csv(OutDataDIR+"/DSO/dso_oscp_data_"+k+".csv.gz")
            clean_meas_df.to_csv(OutDataDIR+"/measurements.csv.gz")
            if est_df is not None:
                est_df.to_csv(OutDataDIR+"/estimations.csv.gz")
            logger.info(f"Done. Results in {OutDataDIR=}")
                    
    elif args.case == "get_load_yml":
        for dataset in args.dataset:
            DataPointMapsDIR = configs[dataset]["DataPointMapsDIR"]
            regex_dp = re.compile(r"^(\d+)\-data\-points\.yml$")
            # DataPointMapsDIR = configs[dataset]["DataPointMapsDIR"]
            onlyfiles = [f for f in os.listdir(DataPointMapsDIR) if os.path.isfile(os.path.join(DataPointMapsDIR, f))]
            logger.info(onlyfiles)
            
            dp_dict = defaultdict(lambda: [])
            for f in tqdm(onlyfiles, desc="Loading Data Point Maps"):
                result = regex_dp.search(f)
                if result:
                    with open(os.path.join(DataPointMapsDIR, f), "rb") as f:
                        dp_data = yaml.load(f, Loader=yaml.FullLoader)
                        dp_dict[result.group(1)].append(dp_data)
                else:
                    logger.error(f"unk file {f}")
            
            #[{'401': [{'identifier': '401.10010', 'protocol': '60870-5-104', 'protocol_data': {'coa': 401, 'cot': 1, 'direction': 'monitoring', 'ioa': 10010, 'type_id': 13}, 'providers': {'sources': [{'domain': 'source', 'provider_data': {'attribute': 'voltage', 'context': 'MEASUREMENT', 'grid_element': 'bus.0', 'type': 'float'}, 'provider_type': 'POWER_GRID'}]}, 'value': None}, {'identifier': '401.10020', 'protocol': '60870-5-104', 'protocol_data': {'coa': 401, 'cot': 1, .... 
            bus_map = dict()
            for k,v in dp_dict.items():
                #print(k,v)
                for x in v[0][k]:
                    if "providers" not in x:
                        continue
                    for ts, prov in x["providers"].items():
                        ge = prov[0]["provider_data"]["grid_element"]
                        if "bus" in ge:
                            #print(k, ge)
                            bus_map[ge] = k
            pp=get_grid_pp(configs[dataset]["DIR"])
            #print(pp["load"])
            load_map=dict()
            for bus, dp in bus_map.items():
                bus_i = int(bus.split(".")[1])
                # if bus_i == 8:
                #     print(dp) #849
                #     exit()
                bus_l = pp["load"][pp["load"]["bus"] == bus_i]
                if len(bus_l) > 0:
                    load_map[bus] = bus_l
                    #print(bus_l)
            #exit()

            for bus, bus_l in tqdm(load_map.items()):
                dp = bus_map[bus]
                max_ioa=0
                existing=0
                out_dict = deepcopy(dp_dict[dp][0])
                out_dict[dp] = list()
                for x in dp_dict[dp][0][dp]:
                    #print(x)
                    max_ioa=max(max_ioa,x["protocol_data"]["ioa"])
                    for st, vs in x["providers"].items():
                        for v in vs:
                            if v["provider_data"]["attribute"] == "active_power" or v["provider_data"]["attribute"] == "active_power_to": #"bus" in v["provider_data"]["grid_element"] and 
                                out_dict[dp].append(x)
                            for idx,bus_load in bus_l.iterrows():
                                if v["provider_data"]["grid_element"] == 'load.'+str(idx):
                                    existing+=0
                                    logger.warning(f"load.{idx} already existing in {dp} {existing=}")
                                    break
                            if existing > 0:
                                break
                        if existing > 0:
                            break
                    # if existing > 0:
                    #     break
                if existing > 0:
                    continue
                for idx,bus_load in bus_l.iterrows():
                    max_ioa=max_ioa+10
                    new_val={'identifier': str(dp)+'.'+str(max_ioa), 'protocol': '60870-5-104', 
                            'protocol_data': {'coa': int(dp), 'cot': 1, 'direction': 'monitoring', 'ioa': int(max_ioa), 'type_id': 13}, 
                            'providers': {'sources': [{'domain': 'source', 'provider_data': {'attribute': 'active_power', 'context': 'MEASUREMENT', 'grid_element': 'load.'+str(idx), 'type': 'float'}, 'provider_type': 'POWER_GRID'}]}, 
                            'value': None}
                    dp_dict[dp][0][dp].append(new_val)
                    out_dict[dp].append(new_val)
                    if False:
                        max_ioa=max_ioa+10
                        new_val={'identifier': str(dp)+'.'+str(max_ioa), 'protocol': '60870-5-104', 
                                'protocol_data': {'coa': int(dp), 'cot': 1, 'direction': 'monitoring', 'ioa': int(max_ioa), 'type_id': 13}, 
                                'providers': {'sources': [{'domain': 'source', 'provider_data': {'attribute': 'reactive_power', 'context': 'MEASUREMENT', 'grid_element': 'load.'+str(idx), 'type': 'float'}, 'provider_type': 'POWER_GRID'}]}, 
                                'value': None}
                        #print()
                        #print(new_val)
                        dp_dict[dp][0][dp].append(new_val)
                        out_dict[dp].append(new_val)
                f_name=f"{dp}-data-points.yml"
                out_f = os.path.join(DataPointMapsDIR,f_name)
                #print(out_f)
                if False:
                    with open(out_f, "w") as f:
                        yaml.dump(dp_dict[dp][0], f)
                with open(out_f, "w") as f:
                    yaml.dump(out_dict, f)
                #exit()


    elif args.case == "get_features_dso":
        reg_type=args.reg_type
        atk_subset_out_name=None

        for dataset in args.dataset:
            logger.info(f"{args.case} {dataset=}")
            OutDataDIR = configs[dataset]["OutDataDIR"]
            OutDataDIR_DSO = os.path.join(OutDataDIR, "DSO")


            dso_files_l = get_dso_files(OutDataDIR_DSO)
            dso_files = read_dso_files(OutDataDIR_DSO, dso_files_l)

            if DO_OSCP_WORKAROUND:      
                logger.warning("using oscp workaround. TODO: fix oscp calc by CPO in simulation...")          
                cpos = [f for f in os.listdir(OutDataDIR+"/feats/") if os.path.isdir(os.path.join(OutDataDIR+"/feats/", f)) and f.startswith("CPO_")]
                cpo = cpos[0] #TODO: iterate cpos
                cpo_feat = load_feats(OutDataDIR, cpo) #TODO: tqdm...
                for group, df_dict in cpo_feat:
                    new_speed = oscp_speed_workaround(df_dict["cp_g_df"], dso_files[group])
                    dso_files[group]["charge_speed"] = new_speed

                    if args.atk_subset:
                        new_speed_should = oscp_speed_workaround(df_dict["cp_g_df"], dso_files[group], pred_col="custom_data_charge_speed_lag_0")
                        new_is_atk = oscp_speed_workaround(df_dict["cp_g_df"], dso_files[group], pred_col="is_attack_lag_0")

                        dso_files[group]["custom_data_charge_speed"] = new_speed_should
                        dso_files[group]["is_attack"] = new_is_atk

                        dso_files[group]["custom_data_considered_atk"] = 0
                        atk_subset_out_name=args.atk_subset
                        atk_l = get_atks(l_buffer=6, start_sim_offset_h=-1)
                        #print(atk_l)
                        for atk in atk_l:
                            if atk["type"] == None:
                                pass
                            elif args.atk_subset=="combo" and "FDI" in atk["type"] and "MAD" in atk["type"]:
                                dso_files[group].loc[atk["start"]:atk["end"], "custom_data_considered_atk"] = 1
                            elif args.atk_subset=="fdi" and "FDI" in atk["type"]:
                                dso_files[group].loc[atk["start"]:atk["end"], "custom_data_considered_atk"] = 1
                            elif args.atk_subset=="mad" and "MAD" in atk["type"]:
                                dso_files[group].loc[atk["start"]:atk["end"], "custom_data_considered_atk"] = 1
                        #print(dso_files[group][dso_files[group]["is_attack"] != 0])
                        dso_files[group].loc[(dso_files[group]["custom_data_considered_atk"] == 0), "charge_speed"] = dso_files[group].loc[(dso_files[group]["custom_data_considered_atk"] == 0), "custom_data_charge_speed"]
                        dso_files[group].loc[(dso_files[group]["custom_data_considered_atk"] == 0), "is_attack"] = 0
                        #print(dso_files[group][dso_files[group]["is_attack"] != 0])
                        dso_files[group] = dso_files[group].drop(columns=['custom_data_considered_atk'])
                        dso_files[group] = dso_files[group].drop(columns=['is_attack'])
                        dso_files[group] = dso_files[group].drop(columns=['custom_data_charge_speed'])
                        #print(dso_files[group][dso_files[group]["is_attack"] != 0])
                    #exit()

            dso_est_df=None
            if False:
                dso_est_df = read_dso_estimations(OutDataDIR+"/estimations.csv.gz")
            dso_meas_df = read_dso_measurements(OutDataDIR+"/measurements.csv.gz")
            
            if False:
                print([c for c in dso_meas_df.columns if ".active_power" in c and "bus" in c])
                print([c for c in dso_meas_df.columns if ".active_power" in c and "sgen" in c])
                bus_ns = sorted(set([int(c.split(".")[3]) for c in dso_meas_df.columns if ".active_power" in c and "bus" in c]))
                pp=get_grid_pp(configs[dataset]["DIR"])

                _plot_l_s = get_grid_measurements_from_export(configs[dataset]["DIR"], pp, bus_ns=bus_ns, target_elem="bus", target_var=".MEASUREMENT.active_power" )
                _plot_l_s2 = get_grid_measurements_from_export(configs[dataset]["DIR"], pp, bus_ns=[], target_elem="sgen", target_var=".MEASUREMENT.active_power" )
                
                def get_intert_expo(bus_meas,dso_meas_df):
                    interp_map=[]
                    bus_i = int(bus_meas.split(".")[3])
                    bus_expo = _plot_l_s[bus_i][0]
                    bus_expo_interp = _interp(bus_expo.to_frame(), dso_meas_df.index, col=None)
                    #dso_meas_df[bus_meas] = bus_expo_interp
                    interp_map.append((bus_meas,bus_expo_interp))

                    for meas in [c for c in dso_meas_df.columns if ".active_power" in c and "sgen" in c]:
                        meas_i = int(meas.split(".")[3])
                        for meas_expo in _plot_l_s2[bus_i]:
                            if "sgen."+str(meas_i)+"." in meas_expo.name:
                                meas_expo_interp = _interp(meas_expo.to_frame(), dso_meas_df.index, col=None)
                                #dso_meas_df[meas] = meas_expo_interp
                                interp_map.append((meas,meas_expo_interp))
                    return interp_map
                with ProcessPoolExecutor(NUM_THREADS) as pool:
                    results=[]
                    for bus_meas in tqdm([c for c in dso_meas_df.columns if ".active_power" in c and "bus" in c]):
                        results.append(pool.submit(get_intert_expo, bus_meas,dso_meas_df))
                    for res in tqdm(results):
                        interp_map = res.result()
                        for meas,meas_expo_interp in interp_map:
                            dso_meas_df[meas] = meas_expo_interp
                #exit()
            #print([c for c in dso_meas_df if "bus." in c])
            DIR = configs[dataset]["DIR"]
            dso_feat = get_dso_group_feat_dfs(DIR, dso_files, dso_meas_df, dso_est_df, num_lags=NUM_LAGS)
            
            # print(dso_feat["CP_11"]["cp_g_df"].columns)
            # print(dso_feat["CP_11"]["cp_g_df"])
            # cp_g_df= dso_feat["CP_11"]["cp_g_df"]
            # #exit()
            # fig, ax = plt.subplots(figsize=(16,9))
            # ax.plot(cp_g_df.index, cp_g_df["charge_speed_lag_0"], label="charge_speed_lag_0") 
            # plt.legend()
            # plt.show()
            # plt.close()
            # # print(row)
            # exit()

            # logger.info(f'Writing Results to {OutDataDIR+"/feats/dso.gz"} ...')
            if atk_subset_out_name is None:
                save_feats(dso_feat, OutDataDIR, "DSO")
            else:
                save_feats(dso_feat, OutDataDIR, "DSO_"+atk_subset_out_name)

    elif args.case == "plot_base":
        for dataset in args.dataset:
            logger.info(dataset)
            OutDataDIR = configs[dataset]["OutDataDIR"]
            #TRAIN_START_DATE=configs[dataset]["TRAIN_START_DATE"]
            #VALIDATION_START_DATE=configs[dataset]["VALIDATION_START_DATE"]
            #ATK_START_DATE=configs[dataset]["ATK_START_DATE"]

            cpo_feat = load_feats(OutDataDIR, "CPO_0")
            cpo_files=dict()
            for group, df_dict in tqdm(cpo_feat, total=load_feats_len(OutDataDIR, "CPO_0"), desc="eval cpo regression "):
                prune_feats(df_dict, args.features)
                #print(df_dict["cp_g_df"])
                cpo_files[group] = df_dict["cp_g_df"]
                #break

            print(OutDataDIR)
            OutDataDIR_DSO = os.path.join(OutDataDIR, "DSO")
            dso_files_l = get_dso_files(OutDataDIR_DSO)
            print(dso_files_l)
            dso_files = read_dso_files(OutDataDIR_DSO, dso_files_l)
            for cp_g,df in dso_files.items():
                print(cp_g,df)
                break


            for cp_g in dso_files.keys():
                fig, ax = plt.subplots()
                dso_df = dso_files[cp_g]
                cpo_df = cpo_files[cp_g]
                ax.plot(dso_df["measurements_measure_time"], dso_df["charge_speed"], label="dso "+cp_g)
                #ax.plot(dso_df["measurements_measure_time"], dso_df["measurements_value"], label="dso "+cp_g)
                #ax.plot(dso_df.index, dso_df["measurements_value"], label="dso "+cp_g)
                ax.plot(cpo_df.index, cpo_df["charge_speed_lag_0"], label="cpo "+cp_g)
                
                plt.legend(loc="upper left")
                plt.show()
                plt.close()
            exit()
                
            #datasetb = configs[dataset]["BASE"]
            #base_dir = configs[datasetb]["OutDataDIR"]

            # base_dfs=dict()
            # base_feat = load_feats(base_dir,"DSO")
            # for group, df_dict in tqdm(base_feat, total=load_feats_len(base_dir, "DSO"), desc="base_dir "):
            #     base_dfs[group] = df_dict

            do_bus=8
            dso_feat = load_feats(OutDataDIR, ("DSO_"+args.atk_subset) if args.atk_subset else "DSO")
            for group, df_dict in tqdm(dso_feat, total=load_feats_len(OutDataDIR, "DSO"), desc="eval dso regression "):
                if "_"+str(do_bus) not in group:
                    continue
                prune_feats(df_dict, args.features)
                print(df_dict["cp_g_df"].loc["2023-12-01 14:30:10":])
                break



    elif args.case == "plot_atks":
        for dataset in args.dataset:
            logger.info(dataset)
            OutDataDIR = configs[dataset]["OutDataDIR"]
            #TRAIN_START_DATE=configs[dataset]["TRAIN_START_DATE"]
            #VALIDATION_START_DATE=configs[dataset]["VALIDATION_START_DATE"]
            #ATK_START_DATE=configs[dataset]["ATK_START_DATE"]
                
            cp_g_atk_types = defaultdict(lambda: list())
            cpo_files_d = get_cpo_files(OutDataDIR)
            for cpo, cps in cpo_files_d.items():
                cpo_files = read_cpo_files(os.path.join(OutDataDIR, cpo), cps)
                for cp_g,v in cpo_files.items():
                    for cp,df in v.items():
                        # print([c for c in df.columns])
                        # print(df["custom_data_atk_type"])
                        df1 = df["custom_data_atk_type"].loc[df["custom_data_atk_type"].shift(-1) != df["custom_data_atk_type"]]
                        df2 = df["custom_data_atk_type"].loc[df["custom_data_atk_type"].shift(1) != df["custom_data_atk_type"]]
                        df3 = pd.concat([df1,df2]).sort_index()
                        df3 = df3[~df3.index.duplicated(keep='first')]
                        cp_g_atk_types[cp_g].append(df3) #cahnges in atk type (start and end)
                        # print(df3)
                        # exit()

            datasetb = configs[dataset]["BASE"]
            base_dir = configs[datasetb]["OutDataDIR"]

            # base_dfs=dict()
            # base_feat = load_feats(base_dir,"DSO")
            # for group, df_dict in tqdm(base_feat, total=load_feats_len(base_dir, "DSO"), desc="base_dir "):
            #     base_dfs[group] = df_dict

            do_bus=8
            dso_feat = load_feats(OutDataDIR, ("DSO_"+args.atk_subset) if args.atk_subset else "DSO")
            for group, df_dict in tqdm(dso_feat, total=load_feats_len(OutDataDIR, "DSO"), desc="eval dso regression "):
                if "_"+str(do_bus) not in group:
                    continue
                prune_feats(df_dict, args.features)
                print(df_dict["cp_g_df"].loc["2023-12-01 14:30:10":])
                # print([c for c in df_dict["cp_g_df"].columns if "grid_expo" in c]) #grid_expo_load.7.MEASUREMENT.active_power_lag_0
                # print([c for c in df_dict["feat_cols"] if "grid_expo" in c])
                # print([c for c in df_dict["cp_g_df"].columns if "grid_meas" in c])
                # print([c for c in df_dict["feat_cols"] if "grid_meas" in c])
                #exit()
                print(group)
                # print(base_dfs[group]["cp_g_df"]["charge_speed_lag_0"])
                # min_speed = base_dfs[group]["cp_g_df"]["charge_speed_lag_0"].min()
                # max_speed = base_dfs[group]["cp_g_df"]["charge_speed_lag_0"].max()
                #exit()
                # print([c for c in df_dict["cp_g_df"].columns if "grid_expo_" in c and ".active_power_" in c and "_relation_to_" not in c])
                # exit()
                if False:
                    fig, ax = plt.subplots()
                    print([c for c in df_dict["cp_g_df"].columns if "grid_meas" in c and "_lag_0" in c and "bus." in c and "_relation_to_" not in c and "norm_" not in c])
                    bus_vals=[c for c in df_dict["cp_g_df"].columns if "grid_meas" in c and "_lag_0" in c and "bus." in c and "_relation_to_" not in c and "norm_" not in c]
                    load_vals=[c for c in df_dict["cp_g_df"].columns if "grid_expo_" in c and ".active_power_static_fac.100_lag_0" in c and "load." in c and "_relation_to_" not in c and "norm_" not in c]
                    sgen_vals=[c for c in df_dict["cp_g_df"].columns if "grid_meas" in c and "_lag_0" in c and "sgen." in c and "_relation_to_" not in c and "norm_" not in c]
                    for b in bus_vals:
                        ax.plot(df_dict["cp_g_df"].index, df_dict["cp_g_df"][b], label=b )
                    for b in load_vals:
                        ax.plot(df_dict["cp_g_df"].index, df_dict["cp_g_df"][b], label=b )
                    for b in sgen_vals:
                        ax.plot(df_dict["cp_g_df"].index, df_dict["cp_g_df"][b], label=b )
                    plt.legend(loc="upper left")
                    plt.show()
                    exit()
                grid_meas = [c for c in df_dict["cp_g_df"].columns if "grid_meas" in c and ".active_power_lag_0" in c and ".bus." in c and "_relation_to_" not in c][0]
                print(df_dict["cp_g_df"][grid_meas])
                # print([f for f in df_dict["cp_g_df"].columns if "lag_0" in f and "norm" not in f and "date" not in f])
                # exit()
                print(df_dict["cp_g_df"]["charge_speed_lag_0"])
                print()
                print([f for f in df_dict["cp_g_df"].columns if "grid_est" in f and "lag_0" in f and "norm" not in f])
                fig, ax = plt.subplots()
                print([f for f in df_dict["cp_g_df"].columns if "grid_expo_" in f and "lag_0" in f and "norm" not in f])
                #exit()

                # print([c for c in df_dict["cp_g_df"].columns if "grid_meas" in c and ".active_power" in c and "load." in c and "grid_expo_" not in c])
                # exit()
                sgen_meas = [c for c in df_dict["cp_g_df"].columns if "grid_meas" in c and ".active_power_lag_0" in c and ".sgen." in c][0]
                df_dict["cp_g_df"]["grid_bus."+str(do_bus)+"_load_total"]=0
                df_dict["cp_g_df"]["grid_bus."+str(do_bus)+"_load_total"]+=df_dict["cp_g_df"][grid_meas]
                for sgen_meas in[c for c in df_dict["cp_g_df"].columns if "grid_meas" in c and ".active_power_lag_0" in c and ".sgen." in c]:
                    df_dict["cp_g_df"]["grid_bus."+str(do_bus)+"_load_total"]+=df_dict["cp_g_df"][sgen_meas]
                ax.plot(df_dict["cp_g_df"].index, df_dict["cp_g_df"]["grid_bus."+str(do_bus)+"_load_total"], label="grid_bus."+str(do_bus)+"_load_and_sgen_total" )

                if False:
                    shift_bus=df_dict["cp_g_df"]["grid_bus."+str(do_bus)+"_load_total"].to_frame()
                    for grid_expo_load in [c for c in df_dict["cp_g_df"].columns if "grid_expo_" in c and ".load." in c and ".active_power_static_fac.100_lag_0" in c]: #grid_meas_668.10320.load.1.active_power_static_fac.10_lag_0
                        shift_bus[grid_expo_load] =df_dict["cp_g_df"][grid_expo_load]
                    shift_bus["idx"]=shift_bus.index
                    shift_bus["idx_diff"] = shift_bus["idx"] - shift_bus["idx"].iloc[0]
                    shift_bus["idx_diff_fix"] = shift_bus["idx_diff"] * 0.96
                    shift_bus["idx_fix"] = shift_bus["idx"].iloc[0] +  shift_bus["idx_diff_fix"]
                    shift_bus = shift_bus.set_index(shift_bus["idx_fix"])
                    shift_bus2=_interp(shift_bus["grid_bus."+str(do_bus)+"_load_total"], df_dict["cp_g_df"].index, col=None)
                    print(shift_bus2)
                    df_dict["cp_g_df"]["grid_bus."+str(do_bus)+"_load_total"] = shift_bus2
                    for grid_expo_load in [c for c in df_dict["cp_g_df"].columns if "grid_expo_" in c and ".load." in c and ".active_power_static_fac.100_lag_0" in c]:
                        shift_bus2=_interp(shift_bus[grid_expo_load], df_dict["cp_g_df"].index, col=None)
                        print(shift_bus2)
                        df_dict["cp_g_df"][grid_expo_load] = shift_bus2
                    ax.plot(df_dict["cp_g_df"].index, df_dict["cp_g_df"]["grid_bus."+str(do_bus)+"_load_total"], label="grid_bus."+str(do_bus)+"_load_and_sgen_total" )
                #exit()

                #ax.plot(df_dict["cp_g_df"].index, df_dict["cp_g_df"][grid_meas], label=grid_meas )
                #ax.plot(df_dict["cp_g_df"].index, df_dict["cp_g_df"]["grid_expo_load.7.MEASUREMENT.active_power_lag_0"], label="grid_expo_load.7.MEASUREMENT.active_power_lag_0" )


                df_dict["cp_g_df"]["grid_expo_load_total"]=0
                for grid_expo_load in [c for c in df_dict["cp_g_df"].columns if "grid_expo_load" in c and ".MEASUREMENT.active_power_static_fac.100_lag_0" in c]:
                    df_dict["cp_g_df"]["grid_expo_load_total"] +=df_dict["cp_g_df"][grid_expo_load]
                #ax.plot(df_dict["cp_g_df"].index, df_dict["cp_g_df"]["grid_expo_load_total"], label="grid_expo_load_total" )

                df_dict["cp_g_df"]["grid_vs_expo_remaining_load_total"]= df_dict["cp_g_df"]["grid_bus."+str(do_bus)+"_load_total"]-df_dict["cp_g_df"]["grid_expo_load_total"]
                # ax.plot(df_dict["cp_g_df"].index, df_dict["cp_g_df"]["grid_vs_expo_remaining_load_total"], label="grid_vs_expo_remaining_load_total" )


                # df_dict["cp_g_df"]["grid_meas_load_total80"]=0
                # for grid_expo_load in [c for c in df_dict["cp_g_df"].columns if "grid_meas_" in c and ".load." in c and ".active_power_static_fac.80_lag_0" in c]: #grid_meas_668.10320.load.1.active_power_static_fac.10_lag_0
                #     df_dict["cp_g_df"]["grid_meas_load_total80"] +=df_dict["cp_g_df"][grid_expo_load]
                # ax.plot(df_dict["cp_g_df"].index, df_dict["cp_g_df"]["grid_meas_load_total80"], label="grid_meas_load_total80" )

                df_dict["cp_g_df"]["grid_meas_load_total"]=0
                for grid_expo_load in [c for c in df_dict["cp_g_df"].columns if "grid_expo_" in c and ".load." in c and ".active_power_static_fac.100_lag_0" in c]: #grid_meas_668.10320.load.1.active_power_static_fac.10_lag_0
                    df_dict["cp_g_df"]["grid_meas_load_total"] +=df_dict["cp_g_df"][grid_expo_load]
                ax.plot(df_dict["cp_g_df"].index, df_dict["cp_g_df"]["grid_meas_load_total"], label="grid_meas_load_total" )


                # pp=get_grid_pp(configs[dataset]["DIR"])
                # _plot_l_s = get_grid_measurements_from_export(configs[dataset]["DIR"], pp, bus_ns=[int(group.split("_")[1])], target_elem="storage", target_var=".MEASUREMENT.active_power" )
                # for cp_group_grid_power in _plot_l_s[int(group.split("_")[1])]:
                #     print(f"{cp_group_grid_power=}")
                #     cp_group_grid_power_interp=_interp(cp_group_grid_power, df_dict["cp_g_df"].index, col=None)
                #     df_dict["cp_g_df"]["grid_meas_load_total"] += cp_group_grid_power_interp
                #     ax.plot(df_dict["cp_g_df"].index, df_dict["cp_g_df"]["grid_meas_load_total"], label="grid_meas_load_total+storage" )


                df_dict["cp_g_df"]["grid_vs_meas_remaining_load_total"]= df_dict["cp_g_df"]["grid_bus."+str(do_bus)+"_load_total"]-df_dict["cp_g_df"]["grid_meas_load_total"]
                # out_vals= df_dict["cp_g_df"][(df_dict["cp_g_df"]["grid_vs_meas_remaining_load_total"] > max_speed) | (df_dict["cp_g_df"]["grid_vs_meas_remaining_load_total"] < min_speed)]
                # df_dict["cp_g_df"].loc[out_vals.index, "grid_vs_meas_remaining_load_total"] = df_dict["cp_g_df"].loc[out_vals.index]["charge_speed_lag_0"]
                ax.plot(df_dict["cp_g_df"].index, df_dict["cp_g_df"]["grid_vs_meas_remaining_load_total"], label="grid_vs_meas_remaining_load_total" )
                


                ax.plot(df_dict["cp_g_df"].index, df_dict["cp_g_df"]["charge_speed_lag_0"], label="charge_speed_lag_0")
                #ax.plot(df_dict["cp_g_df"].index, df_dict["cp_g_df"]["grid_est_bus.11.p_mw_lag_0"]*1000*1000*1000, label="grid_est_bus.11.p_mw_lag_0")
                ax.plot(df_dict["cp_g_df"].index, [160000]*len(df_dict["cp_g_df"].index), label="base")

                file_df = file_dfs[dataset]
                DataPointMapsDIR = configs[dataset]["DataPointMapsDIR"]
                # clean_meas_df = get_measurements(file_df, DataPointMapsDIR)
                # print(clean_meas_df)
                # print(clean_meas_df.columns)
                # ax.plot(clean_meas_df.index, clean_meas_df["585.10030.bus.11.active_power"] )
                #clean_meas_df.to_csv("test.csv.gz")

                if False:
                    pp=get_grid_pp(configs[dataset]["DIR"])
                    print(pp)
                    print(pp["bus"])
                    print(pp["load"])
                    print(pp["sgen"])
                    print(pp["storage"])
                    elems = get_elems_attatched_to_bus(pp, do_bus)
                    print(elems)


                    _plot_l_s = get_grid_measurements_from_export(configs[dataset]["DIR"], pp, bus_ns=[int(group.split("_")[1])], target_elem="storage", target_var=".MEASUREMENT.active_power" )
                    cp_group_grid_power = deepcopy(_plot_l_s[int(group.split("_")[1])])[0]
                    ax.plot(cp_group_grid_power.index, cp_group_grid_power.values, label=cp_group_grid_power.name) #grid values (w/mad attacks)

                    _plot_l_s = get_grid_measurements_from_export(configs[dataset]["DIR"], pp, bus_ns=[int(group.split("_")[1])], target_elem="bus", target_var=".MEASUREMENT.active_power" )
                    bus_group_grid_power = deepcopy(_plot_l_s[int(group.split("_")[1])])[0]
                    #ax.plot(bus_group_grid_power.index, bus_group_grid_power.values, label=bus_group_grid_power.name) #grid values (w/mad attacks)

                    _plot_l_s2 = get_grid_measurements_from_export(configs[dataset]["DIR"], pp, bus_ns=[int(group.split("_")[1])], target_elem="load", target_var=".MEASUREMENT.active_power" )
                    bus_group_grid_power = bus_group_grid_power + df_dict["cp_g_df"][sgen_meas].mean()
                    for p_lad in _plot_l_s2[int(group.split("_")[1])]:
                        bus_group_grid_power = bus_group_grid_power - p_lad
                    ax.plot(bus_group_grid_power.index, bus_group_grid_power.values, label="expo_bus_rem") #grid values (w/mad attacks)
                    

                plt.legend(loc="upper left")
                plt.show()
                exit()
                break
            
            pp=get_grid_pp(configs[dataset]["DIR"])
            print(pp)
            lines=pp["line"]
            lines=lines[lines["from_bus"] < lines["to_bus"]]
            print(lines[lines["to_bus"] == do_bus])
            print("line."+str(lines[lines["to_bus"] == do_bus].index[0]))

            ret = []
            elems = ["load", "sgen", "storage"] #, "storage"?
            for e in elems:
                el=pp[e]
                print(el)
                el = e+"."+str(el[el["bus"] == do_bus].index[0])
                ret.append(el)
            print(ret)
            print()

            wall_sim_map=get_ts_sim_map(configs[dataset]["DIR"]).sort_index()
            print(wall_sim_map)
            print(interpolate_sim_time(wall_sim_map, pd.to_datetime('2024-01-05 16:30:02.00+00:00')))
            x = pd.Series([datetime.utcfromtimestamp(sim) for sim in wall_sim_map.values], index = [datetime.utcfromtimestamp(wall) for wall in wall_sim_map.index]) 
            print(x.sort_index())
            #print(datetime.utcfromtimestamp(interpolate_sim_time(wall_sim_map, pd.to_datetime('2024-01-05 16:30:02.00+00:00'))))
            #exit()
        
            elems = get_elems_attatched_to_bus(pp, do_bus)
            print(elems)
            elem = [e for e in elems if "load" in e][0]

            plot_t=[]
            plot_b=[]
            plot_l=[]
            pps=get_grids_json(configs[dataset]["DIR"], wall_sim_map)
            for grid in pps:
                #print(grid)
                if not np.isnan(grid["simtime"]):
                    plot_t.append(datetime.utcfromtimestamp(grid["simtime"]))
                    plot_b.append(grid["values"]["bus."+str(do_bus)+".MEASUREMENT.active_power"])
                    plot_l.append(grid["values"][elem+".MEASUREMENT.active_power"])
                else:
                    logger.error(f"nan simtime for {grid['timestamp']}")
                # print(datetime.utcfromtimestamp(grid["timestamp"]), datetime.utcfromtimestamp(grid["simtime"]), grid["values"]["bus.11.MEASUREMENT.active_power"])
                # print(elem, grid["values"][elem+".MEASUREMENT.active_power"])
            #exit()
            plot_b_s=pd.Series(plot_b, index = plot_t).sort_index()
            plot_l_s=pd.Series(plot_l, index = plot_t).sort_index()
            print(plot_b_s)
            print(plot_l_s)
            _plot_l_s = get_grid_measurements_from_export(configs[dataset]["DIR"], pp, bus_ns=[], target_elem="load", target_var=".MEASUREMENT.active_power" )
            print(_plot_l_s[do_bus])
            print(_plot_l_s[3])
            ax.plot(plot_b_s.index, plot_b_s.values, label="bus."+str(do_bus)+".MEASUREMENT.active_power2") #bus power
            ax.plot(plot_l_s.index, plot_l_s.values, label=elem+".MEASUREMENT.active_power2") #load power


            _plot_l_s = get_grid_measurements_from_export(configs[dataset]["DIR"], pp, bus_ns=[int(group.split("_")[1])], target_elem="storage", target_var=".MEASUREMENT.active_power" )
            cp_group_grid_power = deepcopy(_plot_l_s[int(group.split("_")[1])])[0]
            is_atk_dfs = get_clf_is_atk_dfs(OutDataDIR, group)         
            _is_atk_dfs = is_atk_dfs["CPO_0"]
            fig, ax = plt.subplots(figsize=(16,9))
            ax.plot(cp_group_grid_power.index, cp_group_grid_power.values, label=cp_group_grid_power.name) #grid values (w/mad attacks)
            ax.plot(_is_atk_dfs.index, _is_atk_dfs["charge_speed_lag_0"], label="charge_speed_lag_0")  #based on EnergyInterval (w/ mad and fdi atks)
            ax.plot(_is_atk_dfs.index, _is_atk_dfs["custom_data_charge_speed_lag_0"], label="custom_data_charge_speed_lag_0") #based on OriginalEnergyInterval (w/o attacks)
            #ax.plot(atk_df.index, atk_df["charge_speed_lag_0"], label="charge_speed_lag_0 from "+actor_prefix)  #
            ax.plot(_is_atk_dfs.index, _is_atk_dfs["is_attack_lag_0"]*500, label="is_attack_lag_0") #based on OriginalEnergyInterval (w/o attacks)
            # print(_is_atk_dfs) #
            # exit()

            cp_atk_dfs = cp_g_atk_types[group]
            _is_atk_dfs["cp_atk_dfs"+"_None"] = -1
            _is_atk_dfs["cp_atk_dfs"+"_Unk"] = -1
            _is_atk_dfs["cp_atk_dfs"+"_MAD"] = -1
            _is_atk_dfs["cp_atk_dfs"+"_FDI"] = -1
            _is_atk_dfs["cp_atk_dfs"+"_FDI_post_MAD"] = -1
            for cp_atk_df in cp_atk_dfs:
                print(cp_atk_df.drop_duplicates())
                for i in range(len(cp_atk_df)-1):
                    s=cp_atk_df.index[i]
                    e=cp_atk_df.index[i+1]
                    atk_val = "Unk"
                    if cp_atk_df.loc[s] == "FDI" and cp_atk_df.loc[e] == "FDI":
                        atk_val = "FDI"
                    elif cp_atk_df.loc[s] == "MAD" and cp_atk_df.loc[e] == "MAD":
                        atk_val = "MAD"
                    elif cp_atk_df.loc[s] == "FDI_post_MAD" and cp_atk_df.loc[e] == "FDI_post_MAD":
                        atk_val = "FDI_post_MAD"
                    elif cp_atk_df.loc[s] == "None" and cp_atk_df.loc[e] == "None":
                        atk_val = "None"
                    __is_atk_dfs = _is_atk_dfs.loc[s:e]
                    #__is_atk_dfs = __is_atk_dfs[__is_atk_dfs["cp_atk_dfs"].isna()]
                    _is_atk_dfs.loc[__is_atk_dfs.index, "cp_atk_dfs_"+atk_val] += 1000
            # print(_is_atk_dfs)
            # exit()

            # ax.plot(_is_atk_dfs.index, _is_atk_dfs["cp_atk_dfs"+"_None"], label="cp_atk_dfs"+"_None")  #
            # ax.plot(_is_atk_dfs.index, _is_atk_dfs["cp_atk_dfs"+"_Unk"], label="cp_atk_dfs"+"_Unk")  #
            ax.plot(_is_atk_dfs.index, _is_atk_dfs["cp_atk_dfs"+"_MAD"], label="cp_atk_dfs"+"_MAD")  #
            ax.plot(_is_atk_dfs.index, _is_atk_dfs["cp_atk_dfs"+"_FDI"], label="cp_atk_dfs"+"_FDI")  #
            ax.plot(_is_atk_dfs.index, _is_atk_dfs["cp_atk_dfs"+"_FDI_post_MAD"], label="cp_atk_dfs"+"_FDI_post_MAD")  #

            atk_l = get_atks(l_buffer=4, start_sim_offset_h=-2)
            #print(atk_l)
            _is_atk_dfs["atkt"] = None
            for atk in atk_l:
                __is_atk_dfs = _is_atk_dfs.loc[atk["start"]:atk["end"]]
                __is_atk_dfs = __is_atk_dfs[__is_atk_dfs["atkt"].isna()]
                if False:
                    if atk["type"] == None:
                        pass
                    elif "FDI" in atk["type"] and "MAD" in atk["type"]:
                        #__is_atk_dfs = __is_atk_dfs[(__is_atk_dfs["cp_atk_dfs"+"_MAD"] > 0) | (__is_atk_dfs["cp_atk_dfs"+"_FDI"] > 0) | (__is_atk_dfs["cp_atk_dfs"+"_FDI_post_MAD"] > 0)]
                        __is_atk_dfs = __is_atk_dfs[(__is_atk_dfs["cp_atk_dfs"+"_FDI_post_MAD"] > 0)]
                    elif "MAD" in atk["type"]:
                        __is_atk_dfs = __is_atk_dfs[__is_atk_dfs["cp_atk_dfs"+"_MAD"] > 0]
                    elif "FDI" in atk["type"]:
                        __is_atk_dfs = __is_atk_dfs[__is_atk_dfs["cp_atk_dfs"+"_FDI"] > 0]
                #__is_atk_dfs = __is_atk_dfs[__is_atk_dfs["atkt"].isna()]
                _is_atk_dfs.loc[__is_atk_dfs.index, "atkt"] = atk["type"]

            for atkt in _is_atk_dfs["atkt"].drop_duplicates():
                _is_atkt_dfs = _is_atk_dfs[_is_atk_dfs["atkt"] == atkt]
                ax.plot(_is_atkt_dfs.index, _is_atkt_dfs["charge_speed_lag_0"], 'o', label="charge_speed_lag_0 "+str(atkt))  #based on EnergyInterval (w/ mad and fdi atks)

            #ax.plot(df_dict["cp_g_df"]["charge_speed_lag_0"].index, df_dict["cp_g_df"]["charge_speed_lag_0"], label="charge_speed_lag_0 dso")  #
            

            # grid_df = pd.DataFrame(grid_l)
            # grid_df = grid_df.set_index("time").sort_index()
            # print(grid_df)
            # #exit()
            # ax.plot(grid_df.index, grid_df["active_power"], label="active_power")
            plt.legend(loc="upper left")
            plt.show()
            exit()

            #/home/dk/git/wattson-artifacts/elaadnl/year_powerowl_example/controller-export/power-grid/WALL-2024-01-04T22-58-06-981491+00-00__SIM-2023-01-01T22-38-53-577418+00-00.powerowl.p.gz

    elif args.case == "plot_atks2":
        plt.rc('text', usetex=True)
        plt.rc('font', family='serif')

        font_size=22

        plt.rc('xtick',labelsize=font_size)
        plt.rc('ytick',labelsize=font_size)

        plt.rc('legend',fontsize=font_size)
        for dataset in args.dataset:
            logger.info(dataset)
            OutDataDIR = configs[dataset]["OutDataDIR"]
            #TRAIN_START_DATE=configs[dataset]["TRAIN_START_DATE"]
            #VALIDATION_START_DATE=configs[dataset]["VALIDATION_START_DATE"]
            #ATK_START_DATE=configs[dataset]["ATK_START_DATE"]
                
            datasetb = configs[dataset]["BASE"]
            base_dir = configs[datasetb]["OutDataDIR"]

            # base_dfs=dict()
            # base_feat = load_feats(base_dir,"DSO")
            # for group, df_dict in tqdm(base_feat, total=load_feats_len(base_dir, "DSO"), desc="base_dir "):
            #     base_dfs[group] = df_dict

            pp=get_grid_pp(configs[dataset]["DIR"])
            #print(pp)
            lines=pp["line"]
            lines=lines[lines["from_bus"] < lines["to_bus"]]
            #print(lines[lines["to_bus"] == do_bus])
            #print("line."+str(lines[lines["to_bus"] == do_bus].index[0]))

            _plot_l_s = get_grid_measurements_from_export(configs[dataset]["DIR"], pp, bus_ns=[], target_elem="storage", target_var=".MEASUREMENT.active_power" )

            _is_atk_dfs_all_cps=dict()
            dso_feat = load_feats(OutDataDIR, ("DSO_"+args.atk_subset) if args.atk_subset else "DSO")
            for group, df_dict in tqdm(dso_feat, total=load_feats_len(OutDataDIR, "DSO"), desc="eval dso regression "):
                do_bus = int(group.split("_")[1])
                if "_"+str(do_bus) not in group:
                    continue
                prune_feats(df_dict, args.features)

                cp_group_grid_power = _plot_l_s[int(group.split("_")[1])][0]
                is_atk_dfs = get_clf_is_atk_dfs(OutDataDIR, group)         
                _is_atk_dfs = is_atk_dfs["CPO_0"]
                _is_atk_dfs_all_cps[group] = _interp(_is_atk_dfs, cp_group_grid_power.index, col=None)
                

            _is_atk_dfs_all_cps["all"] = _is_atk_dfs_all_cps["CP_11"].copy()
            _is_atk_dfs_all_cps["all_grid"] = _plot_l_s[int("CP_11".split("_")[1])][0].copy()
            for group, _is_atk_dfs in _is_atk_dfs_all_cps.items():
                if group in ["all", "all_grid", "CP_11"]:
                    continue
                _is_atk_dfs_all_cps["all"]+=_is_atk_dfs
                cp_group_grid_power = _plot_l_s[int(group.split("_")[1])][0]
                _is_atk_dfs_all_cps["all_grid"]+=cp_group_grid_power

            cp_group_grid_power = _is_atk_dfs_all_cps["all_grid"]
            cp_group_grid_power = cp_group_grid_power.loc["2023-12-22 06:00:00":"2023-12-26 02:00:00"] 
            _is_atk_dfs = _is_atk_dfs_all_cps["all"]
            _is_atk_dfs = _is_atk_dfs.loc["2023-12-22 06:00:00":"2023-12-26 02:00:00"] 
            fig, ax = plt.subplots(figsize=(9,7))

            # cmap = matplotlib.cm.get_cmap('Set1')
            # cmap2 = matplotlib.cm.get_cmap('tab20c')
            # cmap3 = matplotlib.cm.get_cmap('Dark2')
            # df['charge_speed'].plot(ax=ax5, label='Attack Charge Speed',color=cmap(0.1))     
            # df['charge_speed_should'].plot(ax=ax5, label='Normal Charge Speed', linestyle='--',color=cmap3(0.1))                    
            # df["pred"].plot(ax=ax5, label='Forecast', linestyle='-.',color=cmap(0.2))

            ax.plot(cp_group_grid_power.index, cp_group_grid_power.values, label="EV Charging Grid Load", linestyle='--') #grid values (w/mad attacks)
            ax.plot(_is_atk_dfs.index, _is_atk_dfs["charge_speed_lag_0"], label="Reported Load", linestyle='-.')  #based on EnergyInterval (w/ mad and fdi atks)
            ax.plot(_is_atk_dfs.index, _is_atk_dfs["custom_data_charge_speed_lag_0"], label="Original Load") #based on OriginalEnergyInterval (w/o attacks)
            #ax.plot(atk_df.index, atk_df["charge_speed_lag_0"], label="charge_speed_lag_0 from "+actor_prefix)  #
            #ax.plot(_is_atk_dfs.index, _is_atk_dfs["is_attack_lag_0"]*500, label="is_attack_lag_0") #based on OriginalEnergyInterval (w/o attacks)
            # print(_is_atk_dfs) #

            print( _is_atk_dfs[_is_atk_dfs["is_attack_lag_0"] != 0])
            delta=pd.Timedelta(minutes=30)
            plt.axvspan(pd.to_datetime("2023-12-22 13:43:26")-delta, pd.to_datetime("2023-12-22 16:43:06")+delta, facecolor='0.55', alpha=0.2, label="Attacks")
            plt.axvspan(pd.to_datetime("2023-12-23 06:24:34")-delta, pd.to_datetime("2023-12-23 09:23:35")+delta, facecolor='0.55', alpha=0.2)
            plt.axvspan(pd.to_datetime("2023-12-24 16:13:29")-delta, pd.to_datetime("2023-12-24 20:33:38")+delta, facecolor='0.55', alpha=0.2)
            plt.axvspan(pd.to_datetime("2023-12-25 09:03:59")-delta, pd.to_datetime("2023-12-25 12:34:09")+delta, facecolor='0.55', alpha=0.2)

            ax.xaxis.set_major_formatter( DateFormatter('%H:%M') )

            plt.legend(fontsize=font_size)
            #plt.legend(loc = "upper right", bbox_to_anchor=(1.16,1.085))
            plt.legend(loc = "upper right", bbox_to_anchor=(1.0,1.105))

            #plt.tick_params(labelsize=18)
            plt.gca().get_yaxis().get_major_formatter().set_useOffset(False)
            plt.gca().get_yaxis().get_major_formatter().set_scientific(False)

            plt.xlim([cp_group_grid_power.index[0], cp_group_grid_power.index[-1]])
            plt.ylim([0, cp_group_grid_power.values.max()*1.1])

            plt.xlabel(r'\textbf{Time}', fontsize=font_size+2)
            plt.ylabel(r'\textbf{Power in W}', fontsize=font_size+2)

            plt.tight_layout()

            #plt.title(group)
            #plt.legend(loc="upper right")
            #plt.show()
            Path(OutDataDIR+"/figs").mkdir(parents=True, exist_ok=True)
            plt.savefig(OutDataDIR+"/figs/"+"atk.pdf")
            print(cp_group_grid_power.index)
            print(_is_atk_dfs.index)
            continue
            break
            exit()


    elif args.case == "train_reg_dso":
        reg_type=args.reg_type

        for dataset in args.dataset:
            OutDataDIR = configs[dataset]["OutDataDIR"]
            TRAIN_START_DATE=configs[dataset]["TRAIN_START_DATE"]
            VALIDATION_START_DATE=configs[dataset]["VALIDATION_START_DATE"]
            ATK_START_DATE=configs[dataset]["ATK_START_DATE"]
            Path(OutDataDIR+"/results/DSO."+".".join(args.features)).mkdir(parents=True, exist_ok=True)

            dso_feat = load_feats(OutDataDIR, ("DSO_"+args.atk_subset) if args.atk_subset else "DSO")

            for group, df_dict in tqdm(dso_feat, total=load_feats_len(OutDataDIR, "DSO"), desc="eval dso regression "+reg_type):
                out_f_name=OutDataDIR+"/results/DSO."+".".join(args.features)+"/"+group+"_"+reg_type+".csv.gz"
                if os.path.isfile(out_f_name):
                    if args.overwrite:
                        logger.debug(f"overwriting existing file {out_f_name}")
                    else:
                        logger.info(f"skipping existing file {out_f_name}")
                        continue
                #print([f for f in df_dict["feat_cols"] if "grid_" in f and "lag_0" in f and "norm" not in f])
                #print([f for f in df_dict["feat_cols"] if "relation" in f])
                prune_feats(df_dict, args.features)
                # for x in [f for f in df_dict["feat_cols"] if "grid_" in f and "lag_0" in f and "norm" not in f]:
                #     print(x)
                # exit()
                # df_dict["pred_col"] = "custom_data_charge_speed_lag_0"
                ret_df = optimize_regression(df_dict, TRAIN_START_DATE, VALIDATION_START_DATE, ATK_START_DATE, reg_type, group)
                ret_df.to_csv(out_f_name)
            logger.info(f"Done. Results in {OutDataDIR+'/results/'}")


    elif args.case == "get_features_cpo":
        reg_type=args.reg_type

        for dataset in args.dataset:
            logger.info(f"{args.case} {dataset=}")
            OutDataDIR = configs[dataset]["OutDataDIR"]

            cpo_files_d = get_cpo_files(OutDataDIR)
            for cpo, cps in cpo_files_d.items():
                cpo_files = read_cpo_files(os.path.join(OutDataDIR, cpo), cps)
                for cpo_feat in get_cpo_group_feat_dfs(cpo_files, num_lags=NUM_LAGS):

                    # logger.info(f'Writing Results to {OutDataDIR+"/feats/"+cpo+".gz"} ...')
                    save_feats(cpo_feat, OutDataDIR, cpo)

    elif args.case == "train_reg_cpo":
        reg_type=args.reg_type

        for dataset in args.dataset:
            OutDataDIR = configs[dataset]["OutDataDIR"]
            TRAIN_START_DATE=configs[dataset]["TRAIN_START_DATE"]
            VALIDATION_START_DATE=configs[dataset]["VALIDATION_START_DATE"]
            ATK_START_DATE=configs[dataset]["ATK_START_DATE"]
            Path(OutDataDIR+"/results/").mkdir(parents=False, exist_ok=True)

            cpos = [f for f in os.listdir(OutDataDIR+"/feats/") if os.path.isdir(os.path.join(OutDataDIR+"/feats/", f)) and f.startswith("CPO_")]
            for cpo in cpos:
                Path(OutDataDIR+"/results/"+cpo+"."+".".join(args.features)).mkdir(parents=False, exist_ok=True)

                cpo_feat = load_feats(OutDataDIR, cpo)

                for group, df_dict in tqdm(cpo_feat, total=load_feats_len(OutDataDIR, cpo), desc="eval cpo regression "+reg_type):
                    out_f_name=OutDataDIR+"/results/"+cpo+"."+".".join(args.features)+"/"+group+"_"+reg_type+".csv.gz"
                    if os.path.isfile(out_f_name):
                        if args.overwrite:
                            logger.debug(f"overwriting existing file {out_f_name}")
                        else:
                            logger.info(f"skipping existing file {out_f_name}")
                            continue
                    #df_dict["pred_col"] = "custom_data_charge_speed_lag_0"
                    prune_feats(df_dict, args.features)
                    ret_df = optimize_regression(df_dict, TRAIN_START_DATE, VALIDATION_START_DATE, ATK_START_DATE, reg_type, group)
                    ret_df.to_csv(out_f_name)
            logger.info(f"Done. Results in {OutDataDIR+'/results/'}")


    elif args.case == "eval_tuning":
        for dataset in args.dataset:
            OutDataDIR = configs[dataset]["OutDataDIR"]
            ret_d = get_eval_dicts(OutDataDIR)
            log_out=""
            for actor in sorted(ret_d.keys()):
                ret_dict = ret_d[actor]
                for reg,conf_eval in ret_dict.items():
                    log_out+= f"\n{actor}:\n\t{reg}\n\t{conf_eval['conf']}\n\t{conf_eval['eval']['rmse']}"
            logger.info(log_out)



    elif args.case == "do_pred" or args.case == "do_pred_dso" or args.case == "do_pred_cpo":
        num_shifts=args.shifts #py ids.py -c=do_pred_dso -s 2 -v -d elaadnl
        #kfold_splits=[None, 5] #None=pred for test data; int=kfold splits for training data
        
        filter_features = args.features
        if "all" in args.features:
            filter_features=[]

        if args.case == "do_pred_dso":
            prefix="DSO"
        elif args.case == "do_pred_cpo":
            prefix="CPO"

        for dataset in args.dataset:
            is_atk=False
            kfold_splits=[5] #None=pred for test data; int=kfold splits for training data
            if "atk" in dataset:
                is_atk=True
                atk_dataset = dataset
                atk_OutDataDIR = configs[atk_dataset]["OutDataDIR"]
                dataset = configs[dataset]["BASE"]
                kfold_splits=[None]
            OutDataDIR = configs[dataset]["OutDataDIR"]
            TRAIN_START_DATE=configs[dataset]["TRAIN_START_DATE"]
            VALIDATION_START_DATE=configs[dataset]["VALIDATION_START_DATE"]
            ATK_START_DATE=configs[dataset]["ATK_START_DATE"]

            kfold_splits_len = sum([k if k is not None else 1 for k in kfold_splits])



            ret_d = get_cp_group_eval_dicts(OutDataDIR) #hyper param optimized confs
            for actor in sorted(ret_d.keys()):
                if not actor.startswith(prefix):
                    continue
                ret_dict = ret_d[actor]

                if len(actor.split(".")) > 1:
                    features = actor.split(".")[1:]
                else:
                    features=["all"]
                    logger.warning(f"assuming all features for {actor}")
                actor_prefix = actor.split(".")[0]
                actor_prefix = (actor_prefix+"_"+args.atk_subset) if args.atk_subset else actor_prefix
                if is_atk:
                    Path(atk_OutDataDIR+"/predictions/"+actor_prefix+"."+".".join(features)).mkdir(parents=True, exist_ok=True)
                else:                    
                    Path(OutDataDIR+"/predictions/"+actor_prefix+"."+".".join(features)).mkdir(parents=True, exist_ok=True)

                logger.debug(f"filtering {features=} based on  {filter_features=}")
                skip_ff=False
                for ff in filter_features:
                    if ff not in features:
                        skip_ff=True
                        break
                if skip_ff:
                    logger.info(f"skipping based on missing filter_feature: {ff} not in {actor_prefix} w/ {features=}")
                    continue

                for reg,group_d in ret_dict.items():
                    # if "SVR" not in reg:
                    #     continue
                    # if "RandomForestRegressor" not in reg :
                    #     continue
                    # if actor_prefix=="DSO" and not ("MLPReg" in reg and actor == "DSO.only_norm_but_pred.no_cps_but_speed.no_norm1.no_date1.only_pred_lag"):
                    #     continue
                    # if "CPO" in actor_prefix and ("DecisionTreeRegressor" in reg and actor == "CPO_0.only_norm.no_norm1.no_date1"):
                    #     continue
                    # if "MLPReg" not in reg:
                    #     continue
                    if args.reg_type is not None:
                        if reg != args.reg_type:
                            logger.info(f"skipping wrong reg_type {reg} != {args.reg_type}")
                            continue

                    #print()
                    act_feat = iter_feats(OutDataDIR, actor_prefix) #generator
                    total=load_feats_len(OutDataDIR, actor_prefix)
                    g_i=0
                    with tqdm(total=total*num_shifts*kfold_splits_len, desc=f"do_pred_{actor_prefix} {actor} {reg}", smoothing=0.0001) as pbar:
                        for group, df_dict_path in act_feat:
                            if is_atk:
                                out_f = atk_OutDataDIR+"/predictions/"+actor_prefix+"."+".".join(features)+"/"+group+"_"+reg+"_"+str(num_shifts)+".csv.gz"
                            else:        
                                out_f = OutDataDIR+"/predictions/"+actor_prefix+"."+".".join(features)+"/"+group+"_"+reg+"_"+str(num_shifts)+".csv.gz"

                            g_i+=1
                            if os.path.isfile(out_f):
                                if args.overwrite:
                                    logger.debug(f"overwriting existing file {out_f}")
                                else:
                                    logger.info(f"skipping existing file {out_f}")
                                    pbar.update(num_shifts*kfold_splits_len)
                                    continue

                            pbar.set_description(f"do_pred_{actor_prefix} {actor} {reg} {group} ({g_i}/{total})")
                            if group not in group_d:
                                logger.warning(f"skipping {group} not in {actor, reg} param opts, ie /results/")
                                pbar.update(num_shifts*kfold_splits_len)
                                continue

                            df_dict = load_single_feat(df_dict_path)

                            opt_df = group_d[group]
                            #print(actor, reg, group, opt_df)
                            # df_dict = act_feat[group]
                            # print([c for c in df_dict["feat_cols"] if c.startswith("charge_s")])
                            prune_feats(df_dict, features)

                            best_df = opt_df[opt_df["rmse"] == opt_df["rmse"].min()].iloc[0]
                            conf = ast.literal_eval(best_df["conf"])
                            # if "MLPReg" in reg and actor == "DSO.only_norm_but_pred.no_cps_but_speed.no_norm1.no_date1.only_pred_lag":
                            #     conf = [round(c/10) if c == 20000 else c for c in conf]
                            # if "GradientBoostingRegressor" in reg and actor == "CPO_0.only_norm.no_norm1.no_date1":
                            #     conf = list(conf) + [10] #n_iter_no_change
                            # "CPO" in actor_prefix and not ("GradientBoostingRegressor" in reg and actor == "CPO_0.only_norm.no_norm1.no_date1")

                            atk_df=None
                            if is_atk:
                                atk_df_dict = load_feats_CPg(atk_OutDataDIR, actor_prefix, group)
                                # cp_g_df = df_dict["cp_g_df"].loc[TRAIN_START_DATE:ATK_START_DATE]
                                # atk_cp_g_df = atk_df_dict["cp_g_df"].loc[ATK_START_DATE:]
                                # df_dict["cp_g_df"] = pd.concat([cp_g_df, atk_cp_g_df])
                                atk_df = atk_df_dict["cp_g_df"]
                                # print(df_dict["cp_g_df"]["is_attack_lag_0"].value_counts())
                                # print(atk_df_dict["cp_g_df"]["is_attack_lag_0"].value_counts())

                            
                            if False:
                                fig, ax = plt.subplots(figsize=(16,9))
                                ax.plot(atk_df.index, atk_df["charge_speed_lag_0"], label="charge_speed_lag_0") #'o', 
                                plt.legend()
                                plt.show()
                                plt.close()
                                exit()

                            y_pred_all = get_regression_pred_conc_kfold(df_dict, TRAIN_START_DATE, ATK_START_DATE, reg, conf, actor, features, num_shifts=num_shifts, pbar=pbar, kfold_splits=kfold_splits, atk_df=atk_df, recursive=args.recursive_regression)
                            y_pred_all.to_csv(out_f)

            if is_atk:
                logger.info(f"Done. predictions in {atk_OutDataDIR+'/predictions/'}")
            else:
                logger.info(f"Done. predictions in {OutDataDIR+'/predictions/'}")


    elif args.case == "get_features_clf" or args.case == "get_features_clf_cpo" or args.case == "get_features_clf_dso":
        for dataset in args.dataset:
            logger.info(f"{dataset=}")
            OutDataDIR = configs[dataset]["OutDataDIR"]
            ret_d = get_prediction_dicts(OutDataDIR)
            
            for actor in sorted(ret_d.keys()):
                if args.case == "get_features_clf_cpo" and not actor.startswith("CPO"):
                    continue
                if args.case == "get_features_clf_dso" and not actor.startswith("DSO"):
                    continue

                actor_new = actor
                if args.use_cpo_pred:
                    actor_new = actor.replace("DSO", "CPO_0")
                    actor_new+=".no_cps_but_speed"
                actor_prefix = actor.split(".")[0]

                act_feat = iter_feats(OutDataDIR, actor_prefix) #generator
                #act_feat = load_feats(OutDataDIR, actor_prefix) #generator
                total=load_feats_len(OutDataDIR, actor_prefix)
                # print(actor, actor_prefix) #DSO.all DSO
                #for group in sorted(ret_d[actor].keys()):
                #for group, df_dict in act_feat:
                for group, df_dict_path in tqdm(act_feat, total=total, desc="get clf features "+actor):
                    df_dict=None
                    for cp_g_dict in ret_d[actor_new][group]:

                        out_d = OutDataDIR+"/clf_feats/"+actor+"/"
                        if args.use_cpo_pred:
                            out_f = out_d + group+"_"+cp_g_dict["reg"]+"_"+cp_g_dict["shifts"]+"CPO0.csv.gz"
                        else:
                            out_f = out_d + group+"_"+cp_g_dict["reg"]+"_"+cp_g_dict["shifts"]+".csv.gz"
                        if os.path.isfile(out_f):
                            if args.overwrite:
                                logger.debug(f"overwriting existing file {out_f}")
                            else:
                                verify_file_integrity=False
                                logger.info(f"skipping existing file {out_f} {verify_file_integrity=}")
                                if verify_file_integrity:
                                    p = Popen(["gzip", "-t", out_f], stdin=PIPE, stdout=PIPE, stderr=PIPE)
                                    output, err = p.communicate()
                                    rc = p.returncode
                                    if rc == 0: #no error
                                        continue
                                    else: #rc == 1
                                        logger.warning(f"NOT skipping existing file {out_f}")
                                        logger.warning(f"File broken: {rc=} {output=} {err=}")
                                        try:
                                            _test_if_file_broken = pd.read_csv(out_f)
                                            logger.error(f"File broken but working read?: {out_f} {_test_if_file_broken}")
                                        except Exception as e:
                                            logger.warning(f"File broken: {e}")
                                else:
                                    continue

                        if df_dict is None:
                            df_dict = load_single_feat(df_dict_path)

                        # print(group, cp_g_dict) #CP_1 {'reg': 'LinearSVR', 'shifts': '5', 'file': 'data/elaadnl/predictions/DSO.all/CP_1_LinearSVR_5.csv.gz', 'features': ['all']}
                        cp_g_pred = pd.read_csv(cp_g_dict["file"], index_col=0, parse_dates=[0])
                        cp_g_pred = cp_g_pred.fillna(method='bfill')
                        if args.use_cpo_pred:
                            #print("interp")
                            cp_g_pred2 = _interp(cp_g_pred, df_dict["cp_g_df"].index, col=None)
                            for x in ["conf", "features" , "reg", "actor"]:
                                cp_g_pred2[x]= cp_g_pred[x].iloc[0]
                            cp_g_pred = cp_g_pred2.fillna(0)
                            #print("interp done")
                        # conf = cp_g_pred["conf"].iloc[0]
                        # conf = ast.literal_eval(conf) #regressor config
                        # print(conf)
                        # print(cp_g_pred)
                        # print(group, df_dict["cp_g_df"][[c for c in df_dict["cp_g_df"].columns if df_dict["pred_col"].replace("_lag_0","_lag_") in c]])

                        clf_df = get_clf_feat_dfs(cp_g_pred, df_dict["cp_g_df"], df_dict["pred_col"])
                        # print(clf_df)
                        # exit()
                        save_clf_data(clf_df, out_d, out_f)

    elif args.case == "optimize_clf" or args.case == "optimize_clf_cpo" or args.case == "optimize_clf_dso":
        clf_type=args.clf_type
        clf_features=args.clf_features

        only_clf_feat=False
        if "no_reg" in clf_features:
            only_clf_feat=True

        train_contamination=False
        if "train_conta" in clf_features: #
            train_contamination=True

        for dataset in args.dataset:
            OutDataDIR = configs[dataset]["OutDataDIR"]
            ret_d = get_clf_feat_file_dicts(OutDataDIR)


            #is_atk_dfs = get_clf_is_atk_dfs(OutDataDIR) #TODO only for atk dfs...

            for actor in sorted(ret_d.keys()):
                if args.case == "optimize_clf_cpo" and not actor.startswith("CPO"):
                    continue
                if args.case == "optimize_clf_dso" and not actor.startswith("DSO"):
                    continue

                if len(actor.split(".")) > 1:
                    features = actor.split(".")[1:]
                else:
                    features=["all"]
                    logger.warning(f"assuming all features for {actor}")

                if "reg_no_cps" in clf_features:
                    if "no_cps" not in features:
                        features.append("no_cps")
                if "copy_reg" in clf_features: #
                    clf_features.extend(features)

                actor_prefix = actor.split(".")[0]
                act_feat = iter_feats(OutDataDIR, actor_prefix) #generator
                #act_feat = load_feats(OutDataDIR, actor_prefix) #generator
                total=load_feats_len(OutDataDIR, actor_prefix)
                #print(actor, actor_prefix) #DSO.all DSO
                #for group in sorted(ret_d[actor].keys()):
                #for group, df_dict in act_feat:
                param_len = get_param_grid_len(clf_type)
                reg_len=len(ret_d[actor][list(ret_d[actor].keys())[0]])

                with tqdm(total=total*reg_len*param_len, desc=f"optimize_clf {actor} {clf_type}") as pbar:
                    g_i=0
                    for group, df_dict_path in act_feat:
                        g_i+=1
                        #clf_is_atk = get_clf_is_atk(is_atk_dfs, group, df_dict["cp_g_df"].index)
                        
                        ret_evals_l=[]
                        out_d = OutDataDIR+"/clf_eval/"+actor+"/"
                        out_f = out_d + group+"_"+clf_type+"."+".".join(clf_features)+".csv.gz"
                        if os.path.isfile(out_f):
                            if args.overwrite:
                                logger.debug(f"overwriting existing file {out_f}")
                            else:
                                logger.info(f"skipping existing file {out_f}")
                                pbar.update(reg_len*param_len)
                                continue
                        df_dict = load_single_feat(df_dict_path)
                        r_i=0
                        for cp_g_dict in ret_d[actor][group]:
                            r_i+=1
                            if args.reg_type is not None:
                                if args.reg_type != cp_g_dict["reg"]:
                                    logger.info(f"skipping {cp_g_dict['reg']} != {args.reg_type}")
                                    pbar.update(param_len)
                                    continue
                            
                            #logger.info(f'\nCurrent group: {group}/ reg: {cp_g_dict["reg"]}.{cp_g_dict["shifts"]}')
                            pbar.set_description(f'optimize_clf {actor} {clf_type} {group} ({g_i}/{total}) {cp_g_dict["reg"]}.{cp_g_dict["shifts"]} ({r_i}/{reg_len})')
                            # print(group, cp_g_dict) #CP_1 {'reg': 'LinearSVR', 'shifts': '5', 'file': 'data/elaadnl/clf_feats/DSO.all/CP_1_LinearSVR_5.csv.gz', 'features': ['all']}
                            #print(cp_g_dict["file"])
                            cp_g_clf_feat = pd.read_csv(cp_g_dict["file"], index_col=0, parse_dates=[0])
                            # print(cp_g_clf_feat)
                            # print(df_dict["cp_g_df"])
                            #print(full_clf_feat)
                            #exit()
                            #cp_g_dict["features"] == features ?
                            prune_feats(df_dict, features)
                            clf_feat_cols = prune_clf_feats(cp_g_clf_feat, clf_features, df_dict=df_dict)

                            # print(clf_feat_cols)
                            # exit()

                            full_feat_cols = sorted(df_dict["feat_cols"] + clf_feat_cols)
                            cp_g_clf_feat = cp_g_clf_feat[[c for c in full_feat_cols if c in cp_g_clf_feat.columns]]
                            cp_g_df = df_dict["cp_g_df"].loc[cp_g_clf_feat.index.min():cp_g_clf_feat.index.max()]
                            # print([c for c in cp_g_df.columns if "grid" in c])
                            # exit()
                            cp_g_df = cp_g_df[[c for c in full_feat_cols if c in cp_g_df.columns]]
                            full_clf_feat = pd.concat([cp_g_clf_feat, cp_g_df], axis=1)
                            # print(full_clf_feat)
                            # print(full_clf_feat[full_clf_feat["prediction_0"].isna()])
                            # print(full_clf_feat.loc[:, full_clf_feat.isna().any()])
                            # exit()

                            do_scale_pipeline = False
                            if "only_norm" in clf_features:
                                do_scale_pipeline = True
                            # print(f"{clf_features=} {do_scale_pipeline=}")
                            # exit()
                            ret_evals = cross_val_clf(clf_type, full_clf_feat, df_dict["feat_cols"], clf_feat_cols, pred_col = None, only_clf_feat=only_clf_feat, pbar=pbar, scale_pipeline=do_scale_pipeline, train_contamination=train_contamination)
                            for ret_eval in ret_evals:
                                ret_eval["reg"] = cp_g_dict["reg"]
                                ret_eval["shifts"] = cp_g_dict["shifts"]
                                ret_eval["features"] = cp_g_dict["features"]
                                ret_eval["clf_features"] = clf_features
                            ret_evals_l.extend(ret_evals)
                            #train_clf(cp_g_dict["reg"], cp_g_clf_feat, df_dict["cp_g_df"], df_dict["feat_cols"], df_dict["pred_col"])
                        if len(ret_evals_l) > 0:
                            ret_eval_df = pd.DataFrame(ret_evals_l)
                            save_clf_data(ret_eval_df, out_d, out_f)


    elif args.case == "eval_tuning_clf":
        #TODO: eval flags...
        drop_clf_feat_string=False
        add_clf_reg_string=True

        for dataset in args.dataset:
            OutDataDIR = configs[dataset]["OutDataDIR"]
            ret_d = get_eval_dicts_clf(OutDataDIR)

            best_d=defaultdict(lambda: defaultdict(lambda: dict()))
            for actor in sorted(ret_d.keys()):
                ret_dict = ret_d[actor]
                for group,ret_d2 in ret_dict.items():
                    for clf_reg,conf_eval in ret_d2.items():
                        conf_eval["reg"] = clf_reg[1]

                        if args.reg_type is not None:
                            if args.reg_type not in conf_eval["reg"]:
                                logger.info(f"skipping {conf_eval['reg']} != {args.reg_type}")
                                continue

                        clf = clf_reg[0]

                        if drop_clf_feat_string:
                            clf = clf.split(".")[0]
                        if add_clf_reg_string:
                            clf = clf+"."+clf_reg[1]

                        if clf not in best_d[actor][group]:
                            best_d[actor][group][clf] = conf_eval
                        else:
                            if conf_eval["eval"] >= best_d[actor][group][clf]["eval"]:
                                if conf_eval["offset"] > best_d[actor][group][clf]["offset"]:
                                    best_d[actor][group][clf] = conf_eval

            save_eval_dicts_clf(best_d, OutDataDIR)
            
    elif args.case == "print_eval_tuning_clf":
        for dataset in args.dataset:
            OutDataDIR = configs[dataset]["OutDataDIR"]
            log_out=""

            best_d = load_eval_dicts_clf(OutDataDIR)
            for actor in sorted(best_d.keys()):
                ret_dict = best_d[actor]
                for group in sorted(ret_dict.keys()):
                    ret_d2 = ret_dict[group]
                    for clf in sorted(ret_d2.keys()):
                        conf_eval = ret_d2[clf]
                        reg = conf_eval["reg"]
                        log_out+= f"\n{actor}:\n\t{group}\n\t\t{clf, reg}\n\t\t{conf_eval['conf']}\n\t\t{conf_eval['eval'], conf_eval['offset']}"
            logger.info(log_out)

    elif args.case == "get_is_atk_dfs" or args.case == "get_is_atk_dfs_cpo" or args.case == "get_is_atk_dfs_dso":
        grid_data = None
        for dataset in args.dataset:
            logger.info(f"{dataset=}")
            is_atk=False
            if "atk" in dataset:
                is_atk=True
                atk_dataset = dataset
                atk_OutDataDIR = configs[atk_dataset]["OutDataDIR"]
                dataset = configs[dataset]["BASE"]
                TRAIN_START_DATE=configs[dataset]["TRAIN_START_DATE"]
                VALIDATION_START_DATE=configs[dataset]["VALIDATION_START_DATE"]
                ATK_START_DATE=configs[dataset]["ATK_START_DATE"]
            else:
                raise ValueError(f"not an attack data set: {dataset}")

            OutDataDIR = configs[dataset]["OutDataDIR"]
            ret_d = get_clf_feat_file_dicts(OutDataDIR)
            atk_ret_d = get_clf_feat_file_dicts(atk_OutDataDIR)

            best_d = load_eval_dicts_clf(OutDataDIR)

            for actor in sorted(ret_d.keys()):
                if args.case == "get_is_atk_dfs_cpo" and not actor.startswith("CPO"):
                    continue
                if args.case == "get_is_atk_dfs_dso" and not actor.startswith("DSO"):
                    continue

                if len(actor.split(".")) > 1:
                    features = actor.split(".")[1:]
                else:
                    features=["all"]
                    logger.warning(f"assuming all features for {actor}")


                actor_prefix = actor.split(".")[0]
                actor_prefix_new = (actor_prefix+"_"+args.atk_subset) if args.atk_subset else actor_prefix
                actor_new = actor.replace(actor_prefix, actor_prefix_new)
                actor_prefix = actor_prefix_new
                #act_feat = load_feats(OutDataDIR, actor_prefix) #generator
                act_feat = iter_feats(OutDataDIR, actor_prefix) #generator
                total=load_feats_len(OutDataDIR, actor_prefix)
                #print(actor, actor_prefix) #DSO.all DSO


                #best_d_len = len([k2 for k,v in best_d[actor].items() for k2,v2 in best_d[actor][k].items()])

                with tqdm(total=total, desc="get_is_atk_dfs "+actor_new) as pbar:
                    with ProcessPoolExecutor(NUM_THREADS) as pool: #multiprocessing
                        g_i=0
                        results=[]
                        for group, _ in act_feat: #df_dict_path
                            out_d = atk_OutDataDIR+"/clf_is_atk_labels/"+actor_new+"/"
                            out_f = out_d + group+".csv.gz"

                            if os.path.isfile(out_f):
                                if args.overwrite:
                                    logger.debug(f"overwriting existing file {out_f}")
                                else:
                                    logger.info(f"skipping existing file {out_f}")
                                    pbar.update(1)
                                    continue

                            is_atk_dfs = get_clf_is_atk_dfs(atk_OutDataDIR, group)

                            atk_df_dict = load_feats_CPg(atk_OutDataDIR, actor_prefix, group)
                            atk_df = atk_df_dict["cp_g_df"]

                            if grid_data is None:
                                DIR = configs[dataset]["DIR"]
                                pp = get_grid_pp(DIR)
                                #start_time=min(v.index.min() for v in is_atk_dfs.values())-pd.Timedelta(days=1)
                                start_time=None
                                grid_data = get_grid_measurements_from_export(DIR, pp, bus_ns=[], target_elem="storage", target_var=".MEASUREMENT.active_power", start_time=start_time)
                            # for k,v in grid_data.items():
                            #     print(k,len(v))
                            cp_group_grid_power = deepcopy(grid_data[int(group.split("_")[1])])[0]
                            #print(cp_group_grid_power[min(v.index.min() for v in is_atk_dfs.values()):max(v.index.max() for v in is_atk_dfs.values())])
                            #print(is_atk_dfs)

                            # print(atk_df) #charge_speed_lag_0
                            # exit()


                            if False:
                                _is_atk_dfs = is_atk_dfs["CPO_0"]
                                fig, ax = plt.subplots(figsize=(16,9))
                                ax.plot(cp_group_grid_power.index, cp_group_grid_power.values, label=cp_group_grid_power.name) #grid values (w/mad attacks)
                                ax.plot(_is_atk_dfs.index, _is_atk_dfs["charge_speed_lag_0"], label="charge_speed_lag_0")  #based on EnergyInterval (w/ mad and fdi atks)
                                ax.plot(_is_atk_dfs.index, _is_atk_dfs["custom_data_charge_speed_lag_0"], label="custom_data_charge_speed_lag_0") #based on OriginalEnergyInterval (w/o attacks)
                                ax.plot(atk_df.index, atk_df["charge_speed_lag_0"], label="charge_speed_lag_0 from "+actor_prefix)  #

                                atk_l = get_atks(5, start_sim_offset_h=-2)
                                #print(atk_l)
                                _is_atk_dfs["atkt"] = None
                                for atk in atk_l:
                                    _is_atk_dfs.loc[atk["start"]:atk["end"], "atkt"] = atk["type"]

                                for atkt in _is_atk_dfs["atkt"].drop_duplicates():
                                    _is_atkt_dfs = _is_atk_dfs[_is_atk_dfs["atkt"] == atkt]
                                    ax.plot(_is_atkt_dfs.index, _is_atkt_dfs["charge_speed_lag_0"], 'o', label="charge_speed_lag_0 "+str(atkt))  #based on EnergyInterval (w/ mad and fdi atks)

                                plt.legend()
                                plt.show()
                                plt.close()
                                # print(row)

                                exit()

                            output_l = do_get_clf_is_atk_conc(pool, atk_df, ATK_START_DATE, is_atk_dfs, cp_group_grid_power, pbar=None)
                            results.append((out_d, out_f, output_l))
                            #print(actor,best_d[actor])
                            #print(best_d_ag)


                        for out_d, out_f, r in results:
                            g_i+=1
                            pbar.set_description(f'load clf features {actor_new} {group} ({g_i}/{total})')
                            #timestamp1 = time.time()
                            output_df = r.result()
                            pbar.update(1)
                            #print(out_f, output_df)
                            #exit()
                            #timestamp2 = time.time()
                            #print("do_get_clf_is_atk_conc took %.2f seconds" % (timestamp2 - timestamp1))
                            save_clf_data(output_df, out_d, out_f)



    elif args.case == "do_clf" or args.case == "do_clf_cpo" or args.case == "do_clf_dso":
        allow_diff_shifts=True
        filter_features = args.features
        if "all" in args.features:
            filter_features=[]

        for dataset in args.dataset:
            logger.info(f"{dataset=}")
            is_atk=False
            if "atk" in dataset:
                is_atk=True
                atk_dataset = dataset
                atk_OutDataDIR = configs[atk_dataset]["OutDataDIR"]
                dataset = configs[dataset]["BASE"]
                TRAIN_START_DATE=configs[dataset]["TRAIN_START_DATE"]
                VALIDATION_START_DATE=configs[dataset]["VALIDATION_START_DATE"]
                ATK_START_DATE=configs[dataset]["ATK_START_DATE"]
            else:
                raise ValueError(f"not an attack data set: {dataset}")

            OutDataDIR = configs[dataset]["OutDataDIR"]
            ret_d = get_clf_feat_file_dicts(OutDataDIR)
            atk_ret_d = get_clf_feat_file_dicts(atk_OutDataDIR)

            best_d = load_eval_dicts_clf(OutDataDIR)

            for actor in sorted(ret_d.keys()):
                if args.case == "do_clf_cpo" and not actor.startswith("CPO"):
                    continue
                if args.case == "do_clf_dso" and not actor.startswith("DSO"):
                    continue

                if len(actor.split(".")) > 1:
                    features = actor.split(".")[1:]
                else:
                    features=["all"]
                    logger.warning(f"assuming all features for {actor}")

                actor_prefix = actor.split(".")[0]
                actor_prefix_new = (actor_prefix+"_"+args.atk_subset) if args.atk_subset else actor_prefix
                actor_new = actor.replace(actor_prefix, actor_prefix_new)
                actor_prefix = actor_prefix_new

                logger.debug(f"filtering {features=} based on  {filter_features=}")
                skip_ff=False
                for ff in filter_features:
                    if ff not in features:
                        skip_ff=True
                        break
                if skip_ff:
                    logger.info(f"skipping based on missing filter_feature: {ff} not in {actor_prefix} w/ {features=}")
                    continue

                #act_feat = load_feats(OutDataDIR, actor_prefix) #generator
                act_feat = iter_feats(OutDataDIR, actor_prefix) #generator
                total=load_feats_len(OutDataDIR, actor_prefix)
                #print(actor, actor_prefix) #DSO.all DSO


                best_d_len = len([k2 for k,v in best_d[actor].items() for k2,v2 in best_d[actor][k].items()])

                with tqdm(total=best_d_len, desc="do_clf "+actor) as pbar:
                    g_i=0
                    for group, df_dict_path in act_feat:
                        g_i+=1
                        out_d = atk_OutDataDIR+"/clf_results/"+actor_new+"/"
                        out_f = out_d + group+".csv.gz"

                        if os.path.isfile(out_f):
                            if args.overwrite:
                                logger.debug(f"overwriting existing file {out_f}")
                            else:
                                logger.info(f"skipping existing file {out_f}")
                                best_d_ag = best_d[actor][group]
                                pbar.update(len(best_d_ag.keys()))
                                continue


                        #print(group)
                        cp_g_dict = ret_d[actor][group]
                        atk_cp_g_dict = atk_ret_d[actor_new][group]
                        best_d_ag = best_d[actor][group]
                        # for clf, best_dict in best_d_ag.items():
                        #     print(clf)

                        _best_d_ag_orig_len = len(best_d_ag.keys())
                        if args.reg_type is not None: #
                            best_d_ag2 = dict()
                            for clf, best_dict in best_d_ag.items():
                                if args.reg_type in clf:
                                    best_d_ag2[clf]=best_dict
                            best_d_ag = best_d_ag2
                        if args.clf_type is not None: #clf_type
                            best_d_ag2 = dict()
                            for clf, best_dict in best_d_ag.items():
                                if args.clf_type in clf:
                                    best_d_ag2[clf]=best_dict
                            best_d_ag = best_d_ag2
                        
                        if len(best_d_ag.keys()) == 0:
                            logger.info(f"skipping empty best_d_ag file {out_f}")
                            pbar.update(_best_d_ag_orig_len)
                            continue
                        # print()
                        # for clf, best_dict in best_d_ag.items():
                        #     if "add_grid_load_expo_" in clf.replace(best_dict["reg"],""):
                        #         print(clf)
                        #         print(best_dict)
                        # exit()


                        atk_labels_d = atk_OutDataDIR+"/clf_is_atk_labels/"+actor_new+"/"
                        atk_labels_f = atk_labels_d + group+".csv.gz"
                        atk_labels_df = pd.read_csv(atk_labels_f, index_col=0, parse_dates=[0])

                        #is_atk_dfs = get_clf_is_atk_dfs(atk_OutDataDIR, group)

                        df_dict = load_single_feat(df_dict_path)
                        atk_df_dict = load_feats_CPg(atk_OutDataDIR, actor_prefix, group)
                        train_df = df_dict["cp_g_df"]
                        atk_df = atk_df_dict["cp_g_df"]



                        pbar.set_description(f'do_clf {actor_new} {group} ({g_i}/{total})')
                        timestamp1 = time.time()
                        output_l = get_clf_result_output_conc(best_d_ag, cp_g_dict, atk_cp_g_dict, train_df, atk_df, TRAIN_START_DATE, ATK_START_DATE, atk_labels_df, pbar=pbar, allow_diff_shifts=allow_diff_shifts)
                        timestamp2 = time.time()
                        logger.debug("get_clf_result_output_conc() took %.2f seconds" % (timestamp2 - timestamp1))
                        # output_l=[]
                        # for clf, best_dict in best_d_ag.items():
                        #     clf_results = get_clf_result_output(clf, best_dict, cp_g_dict, atk_cp_g_dict TRAIN_START_DATE, ATK_START_DATE, is_atk_dfs, pbar=pbar, allow_diff_shifts=allow_diff_shifts)
                        #     if clf_results is not None:
                        #         output_l.append(clf_results)

                        if len(output_l) > 0:
                            output_df = pd.DataFrame(output_l)
                            save_clf_data(output_df, out_d, out_f)
                        else:
                            logger.warning(f"nothing to save for {dataset,actor,group}")

    elif args.case == "eval_clf_results1" or  args.case == "eval_clf_results1_cpo" or args.case == "eval_clf_results1_dso": #clf/reg comparision
        allow_diff_shifts=True
        FIX_atk_labels_based_on_CPO_data=True
        _totals_dict=defaultdict(lambda: defaultdict(lambda: defaultdict(lambda: defaultdict(lambda: defaultdict(lambda: defaultdict(lambda: {"fpr":[],"tpr":[]}))))))
        out_dict=defaultdict(lambda: deepcopy(_totals_dict))
        for dataset in args.dataset:
            logger.info(f"{dataset=}")
            is_atk=False
            if "atk" in dataset:
                is_atk=True
                atk_dataset = dataset
                atk_OutDataDIR = configs[atk_dataset]["OutDataDIR"]

                #fig_out_dir = atk_OutDataDIR.replace(atk_dataset,configs[dataset]["BASE"]+"_"+"all_atks")

                ad = atk_dataset.split("_")
                if "new" in atk_dataset:
                    at_p=ad[4]
                    at_p="99"
                    _fig_out_dir = atk_OutDataDIR.replace(atk_dataset,ad[0]+"_"+ad[1]+"_"+ad[2]+"_all_"+at_p)
                else:
                    at_p=ad[3]
                    at_p="99"
                    _fig_out_dir = atk_OutDataDIR.replace(atk_dataset,ad[0]+"_"+ad[1]+"_all_"+at_p)
                totals_dict=out_dict[_fig_out_dir]

                dataset = configs[dataset]["BASE"]
                TRAIN_START_DATE=configs[dataset]["TRAIN_START_DATE"]
                VALIDATION_START_DATE=configs[dataset]["VALIDATION_START_DATE"]
                ATK_START_DATE=configs[dataset]["ATK_START_DATE"]
            else:
                raise ValueError(f"not an attack data set: {dataset}")

            OutDataDIR = configs[dataset]["OutDataDIR"]
            ret_d = get_clf_feat_file_dicts(OutDataDIR)
            atk_ret_d = get_clf_feat_file_dicts(atk_OutDataDIR)

            best_d = load_eval_dicts_clf(OutDataDIR)


            for actor in sorted(ret_d.keys()):
                if args.case == "eval_clf_results1_cpo" and not actor.startswith("CPO"):
                    continue
                if args.case == "eval_clf_results1_dso" and not actor.startswith("DSO"):
                    continue

                if len(actor.split(".")) > 1:
                    features = actor.split(".")[1:]
                else:
                    features=["all"]
                    logger.warning(f"assuming all features for {actor}")


                actor_prefix = actor.split(".")[0]
                actor_prefix_new = (actor_prefix+"_"+args.atk_subset) if args.atk_subset else actor_prefix
                actor_new = actor.replace(actor_prefix, actor_prefix_new)
                actor_prefix = actor_prefix_new

                atk_act_feat = load_feats(atk_OutDataDIR, actor_prefix)
                atk_df_dict = {group: df_dict for group, df_dict in atk_act_feat}
                #act_feat = load_feats(OutDataDIR, actor_prefix) #generator
                act_feat = iter_feats(OutDataDIR, actor_prefix) #generator
                total=load_feats_len(OutDataDIR, actor_prefix)
                #print(actor, actor_prefix) #DSO.all DSO


                best_d_len = len([k2 for k,v in best_d[actor].items() for k2,v2 in best_d[actor][k].items()])

                subcase=("is_attack1", "1", ) #any speed diff
                subcase=("is_attack2", "1", ) #is_attack flag was set
                # subcase=("is_attack_th", "1", "0.5")
                # subcase=("is_attack_th", "3", "0.05")
                subcase=("is_attack_th", "0", "0.18")
                # subcase=("is_attack_th", "1", "0.2")
                # subcase=("is_attack_abs", "1", "500")

                cut_off=40
                if args.eval_fac is not None:
                    if "." in args.eval_fac:
                        cut_off=int(args.eval_fac.split(".")[1])
                    subcase=("is_attack_abs", "1", args.eval_fac.split(".")[0])

                with tqdm(total=best_d_len, desc="eval_clf_results "+actor_new) as pbar:
                    g_i=0
                    for group, df_dict_path in act_feat:
                        g_i+=1
                        best_d_ag = best_d[actor][group]
                        pbar.set_description(f'eval_clf_results {actor_new} {group} ({g_i}/{total})')
                        out_d = atk_OutDataDIR+"/clf_results/"+actor_new+"/"
                        out_f = out_d + group+".csv.gz"

                        if not os.path.isfile(out_f):
                            logger.warning(f"skipping missing file {out_f}")
                            pbar.update(len(best_d_ag.keys()))
                            continue

                        atk_labels_d = atk_OutDataDIR+"/clf_is_atk_labels/"+actor_new+"/"
                        atk_labels_f = atk_labels_d + group+".csv.gz"
                        if not os.path.isfile(atk_labels_f):
                            logger.warning(f"skipping missing atk label file {atk_labels_f}")
                            pbar.update(len(best_d_ag.keys()))
                            continue


                        atk_labels_df = pd.read_csv(atk_labels_f, index_col=0, parse_dates=[0])

                        out_df = pd.read_csv(out_f, index_col=0, parse_dates=[0])
                        out_df["decision_function"] = out_df["decision_function"].apply(lambda x: to_float_list(x))
                        #print(out_df)
                        #print(atk_labels_df)
                        if "is_attack_th" in subcase:
                            th_fac = float(subcase[2])
                            atk_labels_df["is_attack_th"] = 1
                            atk_labels_df["is_attack_th_type"] = "None"
                            atk_labels_df.loc[atk_labels_df["charge_speed_lag_0_diff"].abs() > th_fac*atk_labels_df["charge_speed_lag_0_diff"].abs().max(), "is_attack_th"] = -1 #diff between reported and should > threshold; captures FDI+/-, MAD+/-, but not MAD+/- with FDI==
                            atk_labels_df.loc[atk_labels_df["charge_speed_lag_0_diff"].abs() > th_fac*atk_labels_df["charge_speed_lag_0_diff"].abs().max(), "is_attack_th_type"] += "FDI"
                            atk_labels_df.loc[atk_labels_df["charge_speed_lag_0_diff"] > th_fac*atk_labels_df["charge_speed_lag_0_diff"].abs().max(), "is_attack_th_type"] += "OCPP_inc"
                            atk_labels_df.loc[atk_labels_df["charge_speed_lag_0_diff"] < -1*th_fac*atk_labels_df["charge_speed_lag_0_diff"].abs().max(), "is_attack_th_type"] += "OCPP_dec"



                            if False:
                                grid_fac = th_fac
                                grid_col = [c for c in atk_labels_df.columns if "grid_expo_storage." in c and "_diff_to_custom_data_charge_speed_lag_0" in c]
                                if len(grid_col)>1:
                                    logger.error(f"too many cols: {grid_col=}")
                                grid_col = grid_col[0]
                                atk_labels_df.loc[(atk_labels_df["is_attack_th"] == -1) & (atk_labels_df[grid_col].abs() > grid_fac*atk_labels_df[grid_col].abs().max()), "is_attack_th_type"] += "Combo"
                                atk_labels_df.loc[(atk_labels_df["is_attack_th"] == -1) & (atk_labels_df[grid_col] > grid_fac*atk_labels_df[grid_col].abs().max()), "is_attack_th_type"] += "Grid_inc"
                                atk_labels_df.loc[(atk_labels_df["is_attack_th"] == -1) & (atk_labels_df[grid_col] < -1*grid_fac*atk_labels_df[grid_col].abs().max()), "is_attack_th_type"] += "Grid_dec"


                            if True:
                                atk_labels_df.loc[(atk_labels_df["is_attack2"] == 1), "is_attack_th"] = 1 #clean up based on original is_attack flag

                            # _fac_2 = 0.5
                            # atk_labels_df.loc[atk_labels_df["charge_speed_lag_0_diff"].abs() > _fac_2*atk_labels_df["charge_speed_lag_0_diff"].abs().max(), "is_attack_th"] = -1
                            # atk_labels_df.loc[atk_labels_df[grid_col].abs() > _fac_2*atk_labels_df[grid_col].abs().max(), "is_attack_th"] = -1
                            # print(atk_labels_df)
                            # exit()
                        elif "is_attack_abs" in subcase:
                            th_fac = float(subcase[2])
                            atk_labels_df["is_attack_abs"] = 1
                            atk_labels_df["is_attack_th_type"] = "None"
                            atk_labels_df.loc[atk_labels_df["charge_speed_lag_0_diff"].abs() > th_fac, "is_attack_abs"] = -1 #diff between reported and should > threshold; captures FDI+/-, MAD+/-, but not MAD+/- with FDI==
                            atk_labels_df.loc[atk_labels_df["charge_speed_lag_0_diff"].abs() > th_fac, "is_attack_th_type"] += "FDI"
                            atk_labels_df.loc[atk_labels_df["charge_speed_lag_0_diff"] > th_fac, "is_attack_th_type"] += "OCPP_inc"
                            atk_labels_df.loc[atk_labels_df["charge_speed_lag_0_diff"] < -1*th_fac, "is_attack_th_type"] += "OCPP_dec"

                            # grid_fac = th_fac
                            # grid_col = [c for c in atk_labels_df.columns if "grid_expo_storage." in c and "_diff_to_custom_data_charge_speed_lag_0" in c]
                            # if len(grid_col)>1:
                            #     logger.error(f"too many cols: {grid_col=}")
                            # grid_col = grid_col[0]
                            # atk_labels_df.loc[(atk_labels_df[grid_col].abs() > grid_fac), "is_attack_abs"] = -1
                            # atk_labels_df.loc[(atk_labels_df[grid_col].abs() > grid_fac), "is_attack_th_type"] += "MAD"
                            # atk_labels_df.loc[(atk_labels_df[grid_col] > grid_fac), "is_attack_th_type"] += "Grid_inc"
                            # atk_labels_df.loc[(atk_labels_df[grid_col] < -1*grid_fac), "is_attack_th_type"] += "Grid_dec"

                            if False:
                                atk_labels_df.loc[(atk_labels_df["is_attack2"] == 1), "is_attack_abs"] = 1 #clean up based on original is_attack flag
                            
                        else:
                            atk_labels_df["is_attack_th_type"] = "Unk"
                            atk_labels_df.loc[(atk_labels_df["charge_speed_lag_0_diff"] > 0), "is_attack_th_type"] ="Inc"
                            atk_labels_df.loc[(atk_labels_df["charge_speed_lag_0_diff"] < 0), "is_attack_th_type"] ="Dec"
                            atk_labels_df.loc[(atk_labels_df["charge_speed_lag_0_diff"] == 0) & (atk_labels_df[subcase[0]] == 1), "is_attack_th_type"] ="None"
                            

                        #TODO: try diff definitions
                        atk_labels_df["is_attack"] = atk_labels_df[subcase[0]] 
                        #atk_labels_df["is_attack"] *= -1
                        # print(atk_labels_df[atk_labels_df["is_attack"] == -1])
                        #atk_labels_df["decision_function"] = atk_labels_df["decision_function"].fillna(1)
                        # exit()

                        if False:
                            print(atk_labels_df)
                            print(atk_labels_df.columns)

                            fig, ax = plt.subplots(figsize=(22,8))
                            ax.plot(
                                atk_labels_df.index,
                                atk_labels_df["charge_speed_lag_0_diff"],
                                #color="b",
                                label=r"test", # {len(d4), len(aucs)}
                                lw=2,
                                alpha=0.8,
                            )
                            plt.show()
                            plt.close()
                            exit()


                        #exit()
                        #fig, ax = plt.subplots()
                        for atk_type in atk_labels_df["is_attack_th_type"].drop_duplicates():
                            if atk_type == "None":
                                for idx,row in out_df.iterrows():
                                    atk_type="AllAtk"
                                    atk_labels_df_type = atk_labels_df.copy()
                                    atk_labels_df_type["decision_function"] = row["decision_function"]
                                    
                                    if len(atk_labels_df_type[atk_labels_df_type["is_attack"] == -1]) == 0:
                                        continue

                                    clf_name = row["clf"].split(".")[0]
                                    clf_full = row["clf"].replace("."+row["clf_r"],"")
                                    clf_r_name = row["clf_r"].split(".")[0]
                                    clf_n_params=".".join(clf_full.split(".")[1:])
                                    clf_r_params=".".join(row["clf_r"].split(".")[1:])

                                    fpr, tpr, thresholds = roc_curve(atk_labels_df_type["is_attack"], atk_labels_df_type["decision_function"], pos_label=-1)
                                    auc_v = auc(fpr, tpr)
                                    if auc_v < 0.4:
                                        fpr, tpr, thresholds = roc_curve(atk_labels_df_type["is_attack"], atk_labels_df_type["decision_function"]*-1, pos_label=-1)
                                    if not np.isnan(auc_v):
                                        totals_dict[actor_new][clf_r_params][clf_n_params][atk_type][clf_name+" "+clf_r_name][group]["fpr"].append(fpr)
                                        totals_dict[actor_new][clf_r_params][clf_n_params][atk_type][clf_name+" "+clf_r_name][group]["tpr"].append(tpr)
                            else:
                                for idx,row in out_df.iterrows():
                                    atk_labels_df_c = atk_labels_df.copy()
                                    atk_labels_df_c["decision_function"] = row["decision_function"]
                                    atk_labels_df_type = atk_labels_df_c[((atk_labels_df_c["is_attack_th_type"] == atk_type) & (atk_labels_df_c["is_attack"] == -1)) | ((atk_labels_df_c["is_attack_th_type"] == "None") & (atk_labels_df_c["is_attack"] == 1))]
                                    # print(atk_labels_df_type[atk_labels_df_type["is_attack"] == -1])
                                    # exit()

                                    if len(atk_labels_df_type[atk_labels_df_type["is_attack"] == -1]) == 0:
                                        continue

                                    if False:#drop concecutive repeats to avoid inflation of TPR wrt. ROC of imprecise classifiers
                                        cs_typedf = atk_df_dict[group]["cp_g_df"].loc[atk_labels_df_type.index]["charge_speed_lag_0"].to_frame()
                                        cs_typedf["is_attack"] = atk_labels_df_type["is_attack"]
                                        labels = (cs_typedf["charge_speed_lag_0"] != cs_typedf["charge_speed_lag_0"].shift()).cumsum()
                                        cs_typedf['flag'] = (labels.map(labels.value_counts()) >= cut_off).astype(int)
                                        cs_typedf_not_same = cs_typedf[(cs_typedf["flag"] == 0) | (cs_typedf["is_attack"] == -1)] #deflate but keep all atks
                                        atk_labels_df_type = atk_labels_df_type.loc[cs_typedf_not_same.index]


                                    #RocCurveDisplay.from_predictions(atk_labels_df_type["is_attack"], atk_labels_df_type["decision_function"], pos_label=-1, ax=ax, name=row["clf"])
                                    # totals_dict[row["clf"]+"_"+atk_type]["is_attack"].extend(atk_labels_df_type["is_attack"])
                                    # totals_dict[row["clf"]+"_"+atk_type]["decision_function"].extend(atk_labels_df_type["decision_function"])
                                    # print(row["clf"])
                                    # exit()
                                    clf_name = row["clf"].split(".")[0]
                                    clf_full = row["clf"].replace("."+row["clf_r"],"")
                                    clf_r_name = row["clf_r"].split(".")[0]
                                    clf_n_params=".".join(clf_full.split(".")[1:])
                                    clf_r_params=".".join(row["clf_r"].split(".")[1:])

                                    # totals_dict[actor_new][clf_name][row["clf_r"]][row["clf"]][atk_type]["is_attack"].extend(atk_labels_df_type["is_attack"])
                                    # totals_dict[actor_new][clf_name][row["clf_r"]][row["clf"]][atk_type]["decision_function"].extend(atk_labels_df_type["decision_function"])


                                    fpr, tpr, thresholds = roc_curve(atk_labels_df_type["is_attack"], atk_labels_df_type["decision_function"], pos_label=-1)
                                    auc_v = auc(fpr, tpr)
                                    if auc_v < 0.4:
                                        fpr, tpr, thresholds = roc_curve(atk_labels_df_type["is_attack"], atk_labels_df_type["decision_function"]*-1, pos_label=-1)
                                    if not np.isnan(auc_v):
                                        totals_dict[actor_new][clf_r_params][clf_n_params][atk_type][clf_name+" "+clf_r_name][group]["fpr"].append(fpr)
                                        totals_dict[actor_new][clf_r_params][clf_n_params][atk_type][clf_name+" "+clf_r_name][group]["tpr"].append(tpr)

                        #plt.show()
                        #exit()
                        pbar.update(len(best_d_ag.keys()))


        plt.rc('text', usetex=USE_TEX)
        plt.rc('font', family='serif')
        font_size=22
        plt.rc('xtick',labelsize=font_size)
        plt.rc('ytick',labelsize=font_size)
        plt.rc('legend',fontsize=font_size)


        
        for fig_out_dir,totals_dict_ac in out_dict.items():
            fig_out_dir += "_comp"
            if FIX_atk_labels_based_on_CPO_data:
                fig_out_dir += "_fix"
            for actor_new,totals_dict in totals_dict_ac.items():
                if len(totals_dict)>0: 
                    out_fig_d = fig_out_dir+("/clf_result_figs_"+args.case.replace("eval_clf_results","")+"/")+actor_new+"/"+"-".join(subcase)+"/"
                    Path(out_fig_d).mkdir(parents=True, exist_ok=True)
                    fig_dict=defaultdict(lambda:defaultdict(lambda:dict()))
                    for clf_r_params, d in tqdm(totals_dict.items(), desc=f"saving eval_clf_results2 {actor_new} {fig_out_dir}"):
                    #for clf_n, d in totals_dict.items():
                        for clf_n_params, d2 in d.items():
                            for atk_type, d3 in d2.items():
                                fig, ax = plt.subplots(figsize=(22,8))
                                #fig, ax = plt.subplots(figsize=(22,14))
                                plt.rc('axes', prop_cycle=(
                                                            cycler('linestyle', ['-', '--', ':', '-.']) * 
                                                            cycler('color', [u'#1f77b4', u'#ff7f0e', u'#2ca02c', u'#d62728', u'#9467bd',]) #[u'#1f77b4', u'#ff7f0e', u'#2ca02c', u'#d62728', u'#9467bd', u'#8c564b', u'#e377c2', u'#7f7f7f', u'#bcbd22', u'#17becf']
                                                            ))
                                max_auc=0
                                #for clf_name, d4 in d3.items():
                                for clf_name in sorted(d3.keys()):
                                    # if "HistGradientBoostingReg" not in clf_name: #TODO
                                    #     continue
                                    d4 = d3[clf_name]
                                    mean_fpr = np.linspace(0, 1, 100) #based on https://scikit-learn.org/stable/auto_examples/model_selection/plot_roc_crossval.html
                                    tprs = []
                                    aucs = []
                                    for group, data in d4.items():
                                        for fpr, tpr in zip(data["fpr"], data["tpr"]):
                                            auc_v = auc(fpr, tpr)
                                            interp_tpr = np.interp(mean_fpr, fpr, tpr)
                                            interp_tpr[0] = 0.0
                                            tprs.append(interp_tpr)
                                            aucs.append(auc_v)
                                            # ax.plot(
                                            #     fpr,
                                            #     tpr,
                                            #     color="b",
                                            #     label=f"ROC {group} {clf_name}",
                                            #     lw=2,
                                            #     alpha=0.3,
                                            # )
                                    mean_tpr = np.mean(tprs, axis=0)
                                    mean_tpr[-1] = 1.0
                                    mean_auc = auc(mean_fpr, mean_tpr)
                                    std_auc = np.std(aucs)
                                    max_auc=max(max_auc, mean_auc)
                                    ax.plot(
                                        mean_fpr,
                                        mean_tpr,
                                        #color="b",
                                        label=f"{clf_name.replace(' ', ' w/ ').replace('Regressor', 'Reg.')} (AUC = {round(mean_auc, 2)} " + r"$\pm$" +f" {round(std_auc, 2)})", # {len(d4), len(aucs)}
                                        #label=f"{clf_name.replace(' ', ' w/ ')}", # {len(d4), len(aucs)}
                                        #label=r"$\pm$", # {len(d4), len(aucs)}
                                        lw=2,
                                        alpha=0.8,
                                    )
                                out_fig_f=out_fig_d+f"R_{clf_r_params}_N_{clf_n_params}_{atk_type}_{max_auc}_44.pdf"
                                #fig.show()

                                box = ax.get_position()
                                ax.set_position([box.x0, box.y0, box.width * (8/22), box.height])

                                ax.legend(loc='right', fontsize=font_size, bbox_to_anchor=(2.85,0.505))

                                plt.gca().get_yaxis().get_major_formatter().set_useOffset(False)
                                plt.gca().get_yaxis().get_major_formatter().set_scientific(False)
                                plt.xlim([0,1])
                                plt.ylim([0,1])
                                plt.xlabel(r'\textbf{False Positive Rate (FPR)}', fontsize=font_size+2)
                                plt.ylabel(r'\textbf{True Positive Rate (TPR)}', fontsize=font_size+2)
                                #plt.tight_layout()

                                fig.savefig(out_fig_f)
                                plt.close()
                                # exit()


    elif args.case == "eval_clf_results2" or  args.case == "eval_clf_results2_cpo" or args.case == "eval_clf_results2_dso":
        allow_diff_shifts=True
        FIX_atk_labels_based_on_CPO_data=True
        _totals_dict=defaultdict(lambda: defaultdict(lambda: defaultdict(lambda: defaultdict(lambda: defaultdict(lambda: defaultdict(lambda: {"fpr":[],"tpr":[]}))))))
        out_dict=defaultdict(lambda: deepcopy(_totals_dict))
        for dataset in args.dataset:
            logger.info(f"{dataset=}")
            is_atk=False
            if "atk" in dataset:
                is_atk=True
                atk_dataset = dataset
                atk_OutDataDIR = configs[atk_dataset]["OutDataDIR"]

                #fig_out_dir = atk_OutDataDIR.replace(atk_dataset,configs[dataset]["BASE"]+"_"+"all_atks")

                ad = atk_dataset.split("_")
                if "new" in atk_dataset:
                    at_p=ad[4]
                    at_p="99"
                    _fig_out_dir = atk_OutDataDIR.replace(atk_dataset,ad[0]+"_"+ad[1]+"_"+ad[2]+"_all_"+at_p)
                else:
                    at_p=ad[3]
                    at_p="99"
                    _fig_out_dir = atk_OutDataDIR.replace(atk_dataset,ad[0]+"_"+ad[1]+"_all_"+at_p)
                totals_dict=out_dict[_fig_out_dir]
                

                dataset = configs[dataset]["BASE"]
                TRAIN_START_DATE=configs[dataset]["TRAIN_START_DATE"]
                VALIDATION_START_DATE=configs[dataset]["VALIDATION_START_DATE"]
                ATK_START_DATE=configs[dataset]["ATK_START_DATE"]
            else:
                raise ValueError(f"not an attack data set: {dataset}")

            atk_OutDataDIR_full=atk_OutDataDIR
            OutDataDIR = configs[dataset]["OutDataDIR"]
            if not os.path.isdir(OutDataDIR):
                newOutDataDIR = OutDataDIR.replace("data", "data_full")
                logger.info(f"changing {OutDataDIR=} to {newOutDataDIR}")
                OutDataDIR = newOutDataDIR
                atk_OutDataDIR_full = atk_OutDataDIR.replace("data", "data_full")

            ret_d = get_clf_feat_file_dicts(OutDataDIR)
            atk_ret_d = get_clf_feat_file_dicts(atk_OutDataDIR_full)

            best_d = load_eval_dicts_clf(OutDataDIR)

            x_path=atk_OutDataDIR+"/clf_results/"
            full_actors = [f for f in os.listdir(x_path) if os.path.isdir(os.path.join(x_path, f))]
            #for actor in sorted(ret_d.keys()):
            for actor in sorted(full_actors):
                if args.case == "eval_clf_results2_cpo" and not actor.startswith("CPO"):
                    continue
                if args.case == "eval_clf_results2_dso" and not actor.startswith("DSO"):
                    continue

                if len(actor.split(".")) > 1:
                    features = actor.split(".")[1:]
                else:
                    features=["all"]
                    logger.warning(f"assuming all features for {actor}")


                actor_prefix = actor.split(".")[0]
                actor_prefix_new = (actor_prefix+"_"+args.atk_subset) if args.atk_subset else actor_prefix
                actor_new = actor.replace(actor_prefix, actor_prefix_new)
                actor_prefix = actor_prefix_new

                atk_act_feat = load_feats(atk_OutDataDIR_full, actor_prefix)
                atk_df_dict = {group: df_dict for group, df_dict in atk_act_feat}
                #act_feat = load_feats(OutDataDIR, actor_prefix) #generator
                act_feat = iter_feats(OutDataDIR, actor_prefix) #generator
                total=load_feats_len(OutDataDIR, actor_prefix)
                #print(actor, actor_prefix) #DSO.all DSO


                best_d_len = len([k2 for k,v in best_d[actor].items() for k2,v2 in best_d[actor][k].items()])

                subcase=("is_attack1", "1", ) #any speed diff
                subcase=("is_attack2", "1", ) #is_attack flag was set
                subcase=("is_attack_th", "1", "0.5")
                subcase=("is_attack_th", "3", "0.05")
                subcase=("is_attack_th", "1", "0.05")
                subcase=("is_attack_th", "1", "0.15")
                #subcase=("is_attack_abs", "1", "500")

                cut_off=40
                if args.eval_fac is not None:
                    if "." in args.eval_fac:
                        cut_off=int(args.eval_fac.split(".")[1])
                    subcase=("is_attack_abs", "1", args.eval_fac.split(".")[0], str(cut_off))

                with tqdm(total=best_d_len, desc="eval_clf_results "+actor_new) as pbar:
                    g_i=0
                    for group, df_dict_path in act_feat:
                        g_i+=1
                        best_d_ag = best_d[actor][group]
                        pbar.set_description(f'eval_clf_results {actor_new} {group} ({g_i}/{total})')
                        out_d = atk_OutDataDIR+"/clf_results/"+actor_new+"/"
                        out_f = out_d + group+".csv.gz"

                        if not os.path.isfile(out_f):
                            logger.warning(f"skipping missing file {out_f}")
                            pbar.update(len(best_d_ag.keys()))
                            continue

                        atk_labels_d = atk_OutDataDIR+"/clf_is_atk_labels/"+actor_new+"/"
                        atk_labels_f = atk_labels_d + group+".csv.gz"
                        if not os.path.isfile(atk_labels_f):
                            logger.warning(f"skipping missing atk label file {atk_labels_f}")
                            pbar.update(len(best_d_ag.keys()))
                            continue


                        atk_labels_df = pd.read_csv(atk_labels_f, index_col=0, parse_dates=[0])

                        if False:
                            grid_col = [c for c in atk_labels_df.columns if "grid_expo_storage." in c and "_diff_to_custom_data_charge_speed_lag_0" in c]
                            for gc in grid_col:
                                print(atk_labels_df[atk_labels_df[gc] != 0][gc].abs().sort_values())
                            print(atk_labels_df[atk_labels_df["charge_speed_lag_0_diff"] != 0]["charge_speed_lag_0_diff"].abs().sort_values())
                            exit()

                        out_df = pd.read_csv(out_f, index_col=0, parse_dates=[0])
                        out_df["decision_function"] = out_df["decision_function"].apply(lambda x: to_float_list(x))
                        #print(out_df)
                        #print(atk_labels_df)
                        if "is_attack_th" in subcase:
                            th_fac = float(subcase[2])
                            atk_labels_df["is_attack_th"] = 1
                            atk_labels_df["is_attack_th_type"] = "None"
                            atk_labels_df.loc[atk_labels_df["charge_speed_lag_0_diff"].abs() > th_fac*atk_labels_df["charge_speed_lag_0_diff"].abs().max(), "is_attack_th"] = -1 #diff between reported and should > threshold; captures FDI+/-, MAD+/-, but not MAD+/- with FDI==
                            atk_labels_df.loc[atk_labels_df["charge_speed_lag_0_diff"].abs() > th_fac*atk_labels_df["charge_speed_lag_0_diff"].abs().max(), "is_attack_th_type"] += "FDI"
                            atk_labels_df.loc[atk_labels_df["charge_speed_lag_0_diff"] > th_fac*atk_labels_df["charge_speed_lag_0_diff"].abs().max(), "is_attack_th_type"] += "OCPP_inc"
                            atk_labels_df.loc[atk_labels_df["charge_speed_lag_0_diff"] < -1*th_fac*atk_labels_df["charge_speed_lag_0_diff"].abs().max(), "is_attack_th_type"] += "OCPP_dec"

                            grid_fac = th_fac
                            grid_col = [c for c in atk_labels_df.columns if "grid_expo_storage." in c and "_diff_to_custom_data_charge_speed_lag_0" in c]
                            if len(grid_col)>1:
                                logger.error(f"too many cols: {grid_col=}")
                            grid_col = grid_col[0]
                            atk_labels_df.loc[(atk_labels_df["is_attack_th"] == -1) & (atk_labels_df[grid_col].abs() > grid_fac*atk_labels_df[grid_col].abs().max()), "is_attack_th_type"] += "Combo"
                            atk_labels_df.loc[(atk_labels_df["is_attack_th"] == -1) & (atk_labels_df[grid_col] > grid_fac*atk_labels_df[grid_col].abs().max()), "is_attack_th_type"] += "Grid_inc"
                            atk_labels_df.loc[(atk_labels_df["is_attack_th"] == -1) & (atk_labels_df[grid_col] < -1*grid_fac*atk_labels_df[grid_col].abs().max()), "is_attack_th_type"] += "Grid_dec"

                            #grid_fac = th_fac*2
                            grid_fac = th_fac*2
                            #print(atk_labels_df["is_attack_th"].value_counts())
                            #grid_col = [c for c in atk_labels_df.columns if "grid_expo_storage." in c and "_diff_to_charge_speed_lag_0" in c][0]
                            atk_labels_df.loc[(atk_labels_df["charge_speed_lag_0_diff"].abs() == 0) & (atk_labels_df["is_attack_th_type"].str.contains("Grid")==False) & (atk_labels_df[grid_col].abs() > grid_fac*atk_labels_df[grid_col].abs().max()), "is_attack_th"] = -1 #diff between grid measurement and reported > threshold; captures MAD+/- with  any FDI (also FDI==)
                            atk_labels_df.loc[(atk_labels_df["charge_speed_lag_0_diff"].abs() == 0) & (atk_labels_df["is_attack_th_type"].str.contains("Grid")==False) & (atk_labels_df[grid_col].abs() > grid_fac*atk_labels_df[grid_col].abs().max()), "is_attack_th_type"] += "MAD"
                            atk_labels_df.loc[(atk_labels_df["charge_speed_lag_0_diff"].abs() == 0) & (atk_labels_df["is_attack_th_type"].str.contains("Grid")==False) & (atk_labels_df[grid_col] > grid_fac*atk_labels_df[grid_col].abs().max()), "is_attack_th_type"] += "Grid_inc"
                            atk_labels_df.loc[(atk_labels_df["charge_speed_lag_0_diff"].abs() == 0) & (atk_labels_df["is_attack_th_type"].str.contains("Grid")==False) & (atk_labels_df[grid_col] < -1*grid_fac*atk_labels_df[grid_col].abs().max()), "is_attack_th_type"] += "Grid_dec"
                                              
                            atk_labels_df.loc[(atk_labels_df["charge_speed_lag_0_diff"].abs() > 0) & (atk_labels_df["is_attack_th_type"].str.contains("Grid")==False) & (atk_labels_df[grid_col].abs() > grid_fac*atk_labels_df[grid_col].abs().max()), "is_attack_th"] = -1
                            atk_labels_df.loc[(atk_labels_df["charge_speed_lag_0_diff"].abs() > 0) & (atk_labels_df["is_attack_th_type"].str.contains("Grid")==False) & (atk_labels_df[grid_col].abs() > grid_fac*atk_labels_df[grid_col].abs().max()), "is_attack_th_type"] += "SmallCombo"
                            atk_labels_df.loc[(atk_labels_df["charge_speed_lag_0_diff"].abs() > 0) & (atk_labels_df["is_attack_th_type"].str.contains("Grid")==False) & (atk_labels_df[grid_col] > grid_fac*atk_labels_df[grid_col].abs().max()), "is_attack_th_type"] += "Grid_inc"
                            atk_labels_df.loc[(atk_labels_df["charge_speed_lag_0_diff"].abs() > 0) & (atk_labels_df["is_attack_th_type"].str.contains("Grid")==False) & (atk_labels_df[grid_col] < -1*grid_fac*atk_labels_df[grid_col].abs().max()), "is_attack_th_type"] += "Grid_dec"

                            if False:
                                atk_labels_df.loc[(atk_labels_df["is_attack2"] == 1), "is_attack_th"] = 1 #clean up based on original is_attack flag

                            # _fac_2 = 0.5
                            # atk_labels_df.loc[atk_labels_df["charge_speed_lag_0_diff"].abs() > _fac_2*atk_labels_df["charge_speed_lag_0_diff"].abs().max(), "is_attack_th"] = -1
                            # atk_labels_df.loc[atk_labels_df[grid_col].abs() > _fac_2*atk_labels_df[grid_col].abs().max(), "is_attack_th"] = -1
                            # print(atk_labels_df)
                            # exit()
                        elif "is_attack_abs" in subcase:
                            th_fac = float(subcase[2])
                            atk_labels_df["is_attack_abs"] = 1
                            atk_labels_df["is_attack_th_type"] = "None"
                            atk_labels_df.loc[atk_labels_df["charge_speed_lag_0_diff"].abs() > th_fac, "is_attack_abs"] = -1 #diff between reported and should > threshold; captures FDI+/-, MAD+/-, but not MAD+/- with FDI==
                            atk_labels_df.loc[atk_labels_df["charge_speed_lag_0_diff"].abs() > th_fac, "is_attack_th_type"] += "FDI"
                            atk_labels_df.loc[atk_labels_df["charge_speed_lag_0_diff"] > th_fac, "is_attack_th_type"] += "OCPP_inc"
                            atk_labels_df.loc[atk_labels_df["charge_speed_lag_0_diff"] < -1*th_fac, "is_attack_th_type"] += "OCPP_dec"

                            grid_fac = th_fac
                            grid_col = [c for c in atk_labels_df.columns if "grid_expo_storage." in c and "_diff_to_custom_data_charge_speed_lag_0" in c]
                            if len(grid_col)>1:
                                logger.error(f"too many cols: {grid_col=}")
                            grid_col = grid_col[0]
                            atk_labels_df.loc[(atk_labels_df[grid_col].abs() > grid_fac), "is_attack_abs"] = -1
                            atk_labels_df.loc[(atk_labels_df[grid_col].abs() > grid_fac), "is_attack_th_type"] += "MAD"
                            atk_labels_df.loc[(atk_labels_df[grid_col] > grid_fac), "is_attack_th_type"] += "Grid_inc"
                            atk_labels_df.loc[(atk_labels_df[grid_col] < -1*grid_fac), "is_attack_th_type"] += "Grid_dec"

                            if False:
                                atk_labels_df.loc[(atk_labels_df["is_attack2"] == 1), "is_attack_abs"] = 1 #clean up based on original is_attack flag
                            

                        #TODO: try diff definitions
                        atk_labels_df["is_attack"] = atk_labels_df[subcase[0]] 
                        # print(atk_labels_df[atk_labels_df["is_attack"] == -1])
                        # exit()

                        #exit()
                        #fig, ax = plt.subplots()
                        for atk_type in atk_labels_df["is_attack_th_type"].drop_duplicates():
                            if atk_type == "None":
                                for idx,row in out_df.iterrows():
                                    atk_labels_df_type = atk_labels_df.copy()
                                    atk_labels_df_type["decision_function"] = row["decision_function"]
                                    if len(atk_labels_df_type[atk_labels_df_type["is_attack"] == -1]) == 0:
                                        continue
                                    # print(atk_labels_df_type[atk_labels_df_type["is_attack"] == -1])
                                    # exit()

                                    clf_name = row["clf"].split(".")[0]
                                    clf_full = row["clf"].replace("."+row["clf_r"],"")
                                    clf_r_name = row["clf_r"].split(".")[0]
                                    clf_n_params=".".join(clf_full.split(".")[1:])
                                    clf_r_params=".".join(row["clf_r"].split(".")[1:])

                                    fpr, tpr, thresholds = roc_curve(atk_labels_df_type["is_attack"], atk_labels_df_type["decision_function"], pos_label=-1)
                                    auc_v = auc(fpr, tpr)
                                    if auc_v < 0.35:
                                        fpr, tpr, thresholds = roc_curve(atk_labels_df_type["is_attack"], atk_labels_df_type["decision_function"]*-1, pos_label=-1)
                                    if not np.isnan(auc_v):
                                        totals_dict[actor_prefix][clf_name+"_"+clf_r_name]["overall"][row["clf_r"]][clf_full][group]["fpr"].append(fpr)
                                        totals_dict[actor_prefix][clf_name+"_"+clf_r_name]["overall"][row["clf_r"]][clf_full][group]["tpr"].append(tpr)
                            else:
                                for idx,row in out_df.iterrows():
                                    atk_labels_df_c = atk_labels_df.copy()
                                    if len(atk_labels_df_c) != len(row["decision_function"]):
                                        raise Exception(f'len(atk_labels_df_c) != len(row["decision_function"]) -- {len(atk_labels_df_c)} != {len(row["decision_function"])}')
                                    atk_labels_df_c["decision_function"] = row["decision_function"]
                                    atk_labels_df_type = atk_labels_df_c[((atk_labels_df_c["is_attack_th_type"] == atk_type) & (atk_labels_df_c["is_attack"] == -1)) | ((atk_labels_df_c["is_attack_th_type"] == "None") & (atk_labels_df_c["is_attack"] == 1))]
                                    if len(atk_labels_df_type[atk_labels_df_type["is_attack"] == -1]) == 0:
                                        #logger.warning(f"no attacks for {atk_type}")
                                        continue
                                    # print(atk_labels_df_type[atk_labels_df_type["is_attack"] == -1])
                                    # exit()


                                    if True:#drop concecutive repeats to avoid inflation of TPR wrt. ROC of imprecise classifiers
                                        #print(atk_labels_df_type)
                                        #print(df_dict_path)
                                        #df_dict = load_single_feat(df_dict_path)
                                        #print(df_dict["cp_g_df"])
                                        #print(out_df)
                                        cs_typedf = atk_df_dict[group]["cp_g_df"].loc[atk_labels_df_type.index]["charge_speed_lag_0"].to_frame()
                                        cs_typedf["is_attack"] = atk_labels_df_type["is_attack"]
                                        #print(cs_typedf)
                                        # 1 2 2 3 0 0 0 0 0 3
                                        # _ _ 2 _ _ 0 0 0 0 _
                                        # _ _ _ _ _ _ 0 0 0 _
                                        labels = (cs_typedf["charge_speed_lag_0"] != cs_typedf["charge_speed_lag_0"].shift()).cumsum()
                                        cs_typedf['flag'] = (labels.map(labels.value_counts()) >= cut_off).astype(int)
                                        cs_typedf_not_same = cs_typedf[(cs_typedf["flag"] == 0) | (cs_typedf["is_attack"] == -1)] #deflate but keep all atks
                                        # print(cs_typedf_not_same)
                                        # print(atk_labels_df_type)
                                        atk_labels_df_type = atk_labels_df_type.loc[cs_typedf_not_same.index]

                                        if False:
                                            fig, ax = plt.subplots(figsize=(16,9))
                                            ax.plot(cs_typedf.index, cs_typedf["charge_speed_lag_0"], 'o', label="charge_speed_lag_0")
                                            ax.plot(cs_typedf_not_same.index, cs_typedf_not_same["charge_speed_lag_0"], 'o', label="charge_speed_lag_0 not same")
                                            plt.legend()
                                            plt.show()
                                            plt.close()                                            
                                            exit()

                                    clf_name = row["clf"].split(".")[0]
                                    clf_full = row["clf"].replace("."+row["clf_r"],"")
                                    clf_r_name = row["clf_r"].split(".")[0]
                                    clf_n_params=".".join(clf_full.split(".")[1:])
                                    clf_r_params=".".join(row["clf_r"].split(".")[1:])

                                    fpr, tpr, thresholds = roc_curve(atk_labels_df_type["is_attack"], atk_labels_df_type["decision_function"], pos_label=-1)
                                    auc_v = auc(fpr, tpr)
                                    if auc_v < 0.35:
                                        fpr, tpr, thresholds = roc_curve(atk_labels_df_type["is_attack"], atk_labels_df_type["decision_function"]*-1, pos_label=-1)
                                    if not np.isnan(auc_v):
                                        totals_dict[actor_prefix][clf_name+"_"+clf_r_name][atk_type][row["clf_r"]][clf_full][group]["fpr"].append(fpr)
                                        totals_dict[actor_prefix][clf_name+"_"+clf_r_name][atk_type][row["clf_r"]][clf_full][group]["tpr"].append(tpr)
                                    else:
                                        logger.warning(f"isnan for {atk_type}")

                        #plt.show()
                        #exit()
                        pbar.update(len(best_d_ag.keys()))

        plt.rc('text', usetex=USE_TEX)
        plt.rc('font', family='serif')
        font_size=22
        plt.rc('xtick',labelsize=font_size)
        plt.rc('ytick',labelsize=font_size)
        plt.rc('legend',fontsize=font_size)

        def sort_items(n):
            ret=defaultdict(lambda:list())
            #print(n)
            for _n in n:
                n_set = [r for r  in _n.split('.')]
                if "set_3" in n_set:
                    ret[1].append(_n)
                elif "set_13" in n_set:
                    ret[0].append(_n)
                elif "set_35" in n_set:
                    ret[2].append(_n)
                elif [__n for __n in n_set if "add_grid_load_expo_rnd_" in __n] and "set_352" in n_set:
                    __n=[__n for __n in n_set if "add_grid_load_expo_rnd_" in __n][0]
                    #print("add_grid_load_expo_rnd_", _n)
                    if len(__n.replace("add_grid_load_expo_rnd_","")) == 1:
                        __n_new = __n.replace("add_grid_load_expo_rnd_","add_grid_load_expo_rnd_0")
                        _n = _n.replace(__n,__n_new)
                    #print("2add_grid_load_expo_rnd_", _n)
                    ret[4].append(_n)
                elif "set_352" in n_set:
                    ret[3].append(_n)
                elif "no_grid_storage" in n_set and "no_grid_sgen" in n_set:
                    ret[1].append(_n)
                elif "no_grid" in n_set:
                    ret[0].append(_n)
                elif "add_grid_load_expo_static_100" in n_set:
                    ret[2].append(_n)
                else:
                    ret[9].append(_n)
            if 4 in ret:
                ret[4] = sorted(ret[4])
                ret[4] = [r.replace("add_grid_load_expo_rnd_0","add_grid_load_expo_rnd_") for r in ret[4]]
                #print(ret[4])
            for k in sorted(ret.keys()):
                for v in ret[k]:
                    yield v

        for fig_out_dir,totals_dict_ac in out_dict.items():
            fig_out_dir += "_comp"
            if FIX_atk_labels_based_on_CPO_data:
                fig_out_dir += "_fix"
            for actor_new,totals_dict in totals_dict_ac.items():
                if len(totals_dict)>0:
                    out_fig_d = fig_out_dir+("/clf_result_figs_"+args.case.replace("eval_clf_results","")+"/")+actor_new+"/"+"-".join(subcase)+"/"
                    Path(out_fig_d).mkdir(parents=True, exist_ok=True)
                    fig_dict=defaultdict(lambda:defaultdict(lambda:dict()))
                    for clf_name, d in tqdm(totals_dict.items(), desc=f"saving eval_clf_results2 {actor_new} {fig_out_dir}"):
                    #for clf_n, d in totals_dict.items():
                        for atk_type, d2 in d.items():
                            fig, ax = plt.subplots(figsize=(22,8))
                            plt.rc('axes', prop_cycle=(
                                                        cycler('linestyle', ['-', '--', ':', '-.']) * #'-', '--', ':', '-.'
                                                        cycler('color', [u'#1f77b4', u'#ff7f0e', u'#2ca02c', ]) #[  u'#1f77b4', u'#ff7f0e', u'#2ca02c', u'#d62728', u'#9467bd', u'#8c564b', u'#e377c2', u'#7f7f7f', u'#bcbd22', u'#17becf']
                                                        ))
                            max_auc=0
                            min_auc=1
                            #for r_name, d3 in d2.items():
                            for r_name in sort_items(d2.keys()):
                                d3 = d2[r_name]
                                #print("r_name",r_name)
                                #for n_name, d4 in d3.items():
                                for n_name in sort_items(d3.keys()):
                                    #rint("n_name",n_name)
                                    d4 = d3[n_name]
                                    mean_fpr = np.linspace(0, 1, 100) #based on https://scikit-learn.org/stable/auto_examples/model_selection/plot_roc_crossval.html
                                    tprs = []
                                    aucs = []
                                    #print(atk_type)
                                    for group, data in d4.items():
                                        for fpr, tpr in zip(data["fpr"], data["tpr"]):
                                            auc_v = auc(fpr, tpr)
                                            interp_tpr = np.interp(mean_fpr, fpr, tpr)
                                            interp_tpr[0] = 0.0
                                            tprs.append(interp_tpr)
                                            aucs.append(auc_v)
                                            # ax.plot(
                                            #     fpr,
                                            #     tpr,
                                            #     color="b",
                                            #     label=f"ROC {group} {clf_name}",
                                            #     lw=2,
                                            #     alpha=0.3,
                                            # )
                                    mean_tpr = np.mean(tprs, axis=0)
                                    mean_tpr[-1] = 1.0
                                    mean_auc = auc(mean_fpr, mean_tpr)
                                    std_auc = np.std(aucs)
                                    max_auc=max(max_auc, mean_auc)
                                    min_auc=min(min_auc, mean_auc)

                                    n_set = [r for r  in n_name.split('.')[1:] if 'reg' not in r and 'norm' not in r]
                                    if "set_3" in n_set:
                                        n_set_name = "Nov.: Basic Grid"
                                    elif "set_13" in n_set:
                                        n_set_name = "Nov.: No Grid"
                                    elif "set_35" in n_set:
                                        n_set_name = "Nov.: Advanced Grid (no noise)"
                                    elif "set_352" in n_set:
                                        perc = [s for s in n_set if "add_grid_load_expo_rnd_" in s][0].split("_")[-1]
                                        #print(perc)
                                        if perc != "5":
                                            continue
                                        n_set_name = r"Nov.: Advanced Grid ("+perc+r"\% noise)"
                                    else:
                                        n_set_name = ".".join(n_set)

                                    r_set = [r_name.split('.')[1]]+[r for r in r_name.split('.')[1:] if 'grid' in r]
                                    if "no_grid_storage" in r_set and "no_grid_sgen" in r_set:
                                        r_set_name = "Reg.: Basic Grid"
                                    elif "no_grid" in r_set:
                                        r_set_name = "Reg.: No Grid"
                                    elif "add_grid_load_expo_static_100" in r_set:
                                        r_set_name = "Reg.: Advanced Grid"
                                    else:
                                        r_set_name = ".".join(r_set)

                                    ax.plot(
                                        mean_fpr,
                                        mean_tpr,
                                        #color="b",
                                        label=f"{r_set_name} -- {n_set_name} (AUC = {round(mean_auc, 2)} $\pm$ {round(std_auc, 2)})", #{atk_type.replace('None','').replace('FDI','').replace('MAD','')} 
                                        #label=f"Mean ROC {atk_type} {r_name.split('.')[1:]} {n_name.split('.')[1:]} (AUC = {round(mean_auc, 2)} $\pm$ {round(std_auc, 2)} {len(d4), len(aucs)})",
                                        lw=2,
                                        alpha=0.8,
                                    )
                            out_fig_f=out_fig_d+f"{clf_name}_{atk_type}_{round(max_auc,2)}_{round(min_auc,2)}_44.pdf"

                            box = ax.get_position()
                            ax.set_position([box.x0, box.y0, box.width * (8/22), box.height])

                            ax.legend(loc='right', fontsize=font_size, bbox_to_anchor=(2.85,0.505))
                            #fig.show()
                            #ax.legend(loc='best', fontsize=font_size)

                            plt.gca().get_yaxis().get_major_formatter().set_useOffset(False)
                            plt.gca().get_yaxis().get_major_formatter().set_scientific(False)
                            plt.xlim([0,1])
                            plt.ylim([0,1])
                            plt.xlabel(r'\textbf{False Positive Rate (FPR)}', fontsize=font_size+2)
                            plt.ylabel(r'\textbf{True Positive Rate (TPR)}', fontsize=font_size+2)
                            fig.savefig(out_fig_f)
                            #plt.show()
                            plt.close()
                            # exit()


    elif args.case == "eval_clf_results3" or  args.case == "eval_clf_results3_cpo" or args.case == "eval_clf_results3_dso":
        allow_diff_shifts=True
        FIX_atk_labels_based_on_CPO_data=True
        _totals_dict=defaultdict(lambda: defaultdict(lambda: defaultdict(lambda: defaultdict(lambda: defaultdict(lambda: defaultdict(lambda: {"fpr":[],"tpr":[]}))))))
        out_dict=defaultdict(lambda: deepcopy(_totals_dict))
        for dataset in args.dataset:
            logger.info(f"{dataset=}")
            is_atk=False
            if "atk" in dataset:
                is_atk=True
                atk_dataset = dataset
                atk_OutDataDIR = configs[atk_dataset]["OutDataDIR"]

                #fig_out_dir = atk_OutDataDIR.replace(atk_dataset,configs[dataset]["BASE"]+"_"+"all_atks")

                ad = atk_dataset.split("_")
                if "new" in atk_dataset:
                    at_p=ad[4]
                    at_p="99"
                    _fig_out_dir = atk_OutDataDIR.replace(atk_dataset,ad[0]+"_"+ad[1]+"_"+ad[2]+"_all_"+at_p)
                else:
                    at_p=ad[3]
                    at_p="99"
                    _fig_out_dir = atk_OutDataDIR.replace(atk_dataset,ad[0]+"_"+ad[1]+"_all_"+at_p)
                totals_dict=out_dict[_fig_out_dir]
                

                dataset = configs[dataset]["BASE"]
                TRAIN_START_DATE=configs[dataset]["TRAIN_START_DATE"]
                VALIDATION_START_DATE=configs[dataset]["VALIDATION_START_DATE"]
                ATK_START_DATE=configs[dataset]["ATK_START_DATE"]
            else:
                raise ValueError(f"not an attack data set: {dataset}")

            atk_OutDataDIR_full=atk_OutDataDIR
            OutDataDIR = configs[dataset]["OutDataDIR"]
            if not os.path.isdir(OutDataDIR):
                newOutDataDIR = OutDataDIR.replace("data", "data_full")
                logger.info(f"changing {OutDataDIR=} to {newOutDataDIR}")
                OutDataDIR = newOutDataDIR
                atk_OutDataDIR_full = atk_OutDataDIR.replace("data", "data_full")

            ret_d = get_clf_feat_file_dicts(OutDataDIR)
            atk_ret_d = get_clf_feat_file_dicts(atk_OutDataDIR_full)

            best_d = load_eval_dicts_clf(OutDataDIR)

            x_path=atk_OutDataDIR+"/clf_results/"
            full_actors = [f for f in os.listdir(x_path) if os.path.isdir(os.path.join(x_path, f))]
            #for actor in sorted(ret_d.keys()):
            for actor in sorted(full_actors):
                if args.case == "eval_clf_results3_cpo" and not actor.startswith("CPO"):
                    continue
                if args.case == "eval_clf_results3_dso" and not actor.startswith("DSO"):
                    continue

                if len(actor.split(".")) > 1:
                    features = actor.split(".")[1:]
                else:
                    features=["all"]
                    logger.warning(f"assuming all features for {actor}")


                actor_prefix = actor.split(".")[0]
                actor_prefix_new = (actor_prefix+"_"+args.atk_subset) if args.atk_subset else actor_prefix
                actor_new = actor.replace(actor_prefix, actor_prefix_new)
                actor_prefix = actor_prefix_new

                atk_act_feat = load_feats(atk_OutDataDIR_full, actor_prefix)
                atk_df_dict = {group: df_dict for group, df_dict in atk_act_feat}
                #act_feat = load_feats(OutDataDIR, actor_prefix) #generator
                act_feat = iter_feats(OutDataDIR, actor_prefix) #generator
                total=load_feats_len(OutDataDIR, actor_prefix)
                #print(actor, actor_prefix) #DSO.all DSO


                best_d_len = len([k2 for k,v in best_d[actor].items() for k2,v2 in best_d[actor][k].items()])

                subcase=("is_attack1", "1", ) #any speed diff
                subcase=("is_attack2", "1", ) #is_attack flag was set
                subcase=("is_attack_th", "1", "0.5")
                subcase=("is_attack_th", "3", "0.05")
                subcase=("is_attack_th", "1", "0.05")
                subcase=("is_attack_abs", "1", "500")

                cut_off=40
                if args.eval_fac is not None:
                    if "." in args.eval_fac:
                        cut_off=int(args.eval_fac.split(".")[1])
                    subcase=("is_attack_abs", "1", args.eval_fac.split(".")[0])

                with tqdm(total=best_d_len, desc="eval_clf_results "+actor_new) as pbar:
                    g_i=0
                    for group, df_dict_path in act_feat:
                        g_i+=1
                        best_d_ag = best_d[actor][group]
                        pbar.set_description(f'eval_clf_results {actor_new} {group} ({g_i}/{total})')
                        out_d = atk_OutDataDIR+"/clf_results/"+actor_new+"/"
                        out_f = out_d + group+".csv.gz"

                        if not os.path.isfile(out_f):
                            logger.warning(f"skipping missing file {out_f}")
                            pbar.update(len(best_d_ag.keys()))
                            continue

                        atk_labels_d = atk_OutDataDIR+"/clf_is_atk_labels/"+actor_new+"/"
                        atk_labels_f = atk_labels_d + group+".csv.gz"
                        if not os.path.isfile(atk_labels_f):
                            logger.warning(f"skipping missing atk label file {atk_labels_f}")
                            pbar.update(len(best_d_ag.keys()))
                            continue


                        atk_labels_df = pd.read_csv(atk_labels_f, index_col=0, parse_dates=[0])

                        if False:
                            grid_col = [c for c in atk_labels_df.columns if "grid_expo_storage." in c and "_diff_to_custom_data_charge_speed_lag_0" in c]
                            for gc in grid_col:
                                print(atk_labels_df[atk_labels_df[gc] != 0][gc].abs().sort_values())
                            print(atk_labels_df[atk_labels_df["charge_speed_lag_0_diff"] != 0]["charge_speed_lag_0_diff"].abs().sort_values())
                            exit()

                        out_df = pd.read_csv(out_f, index_col=0, parse_dates=[0])
                        out_df["decision_function"] = out_df["decision_function"].apply(lambda x: to_float_list(x))
                        #print(out_df)
                        #print(atk_labels_df)
                        if "is_attack_th" in subcase:
                            th_fac = float(subcase[2])
                            atk_labels_df["is_attack_th"] = 1
                            atk_labels_df["is_attack_th_type"] = "None"
                            atk_labels_df.loc[atk_labels_df["charge_speed_lag_0_diff"].abs() > th_fac*atk_labels_df["charge_speed_lag_0_diff"].abs().max(), "is_attack_th"] = -1 #diff between reported and should > threshold; captures FDI+/-, MAD+/-, but not MAD+/- with FDI==
                            atk_labels_df.loc[atk_labels_df["charge_speed_lag_0_diff"].abs() > th_fac*atk_labels_df["charge_speed_lag_0_diff"].abs().max(), "is_attack_th_type"] += "FDI"
                            atk_labels_df.loc[atk_labels_df["charge_speed_lag_0_diff"] > th_fac*atk_labels_df["charge_speed_lag_0_diff"].abs().max(), "is_attack_th_type"] += "OCPP_inc"
                            atk_labels_df.loc[atk_labels_df["charge_speed_lag_0_diff"] < -1*th_fac*atk_labels_df["charge_speed_lag_0_diff"].abs().max(), "is_attack_th_type"] += "OCPP_dec"

                            grid_fac = th_fac
                            grid_col = [c for c in atk_labels_df.columns if "grid_expo_storage." in c and "_diff_to_custom_data_charge_speed_lag_0" in c]
                            if len(grid_col)>1:
                                logger.error(f"too many cols: {grid_col=}")
                            grid_col = grid_col[0]
                            atk_labels_df.loc[(atk_labels_df["is_attack_th"] == -1) & (atk_labels_df[grid_col].abs() > grid_fac*atk_labels_df[grid_col].abs().max()), "is_attack_th_type"] += "Combo"
                            atk_labels_df.loc[(atk_labels_df["is_attack_th"] == -1) & (atk_labels_df[grid_col] > grid_fac*atk_labels_df[grid_col].abs().max()), "is_attack_th_type"] += "Grid_inc"
                            atk_labels_df.loc[(atk_labels_df["is_attack_th"] == -1) & (atk_labels_df[grid_col] < -1*grid_fac*atk_labels_df[grid_col].abs().max()), "is_attack_th_type"] += "Grid_dec"

                            #grid_fac = th_fac*2
                            grid_fac = th_fac*2
                            #print(atk_labels_df["is_attack_th"].value_counts())
                            #grid_col = [c for c in atk_labels_df.columns if "grid_expo_storage." in c and "_diff_to_charge_speed_lag_0" in c][0]
                            atk_labels_df.loc[(atk_labels_df["charge_speed_lag_0_diff"].abs() == 0) & (atk_labels_df["is_attack_th_type"].str.contains("Grid")==False) & (atk_labels_df[grid_col].abs() > grid_fac*atk_labels_df[grid_col].abs().max()), "is_attack_th"] = -1 #diff between grid measurement and reported > threshold; captures MAD+/- with  any FDI (also FDI==)
                            atk_labels_df.loc[(atk_labels_df["charge_speed_lag_0_diff"].abs() == 0) & (atk_labels_df["is_attack_th_type"].str.contains("Grid")==False) & (atk_labels_df[grid_col].abs() > grid_fac*atk_labels_df[grid_col].abs().max()), "is_attack_th_type"] += "MAD"
                            atk_labels_df.loc[(atk_labels_df["charge_speed_lag_0_diff"].abs() == 0) & (atk_labels_df["is_attack_th_type"].str.contains("Grid")==False) & (atk_labels_df[grid_col] > grid_fac*atk_labels_df[grid_col].abs().max()), "is_attack_th_type"] += "Grid_inc"
                            atk_labels_df.loc[(atk_labels_df["charge_speed_lag_0_diff"].abs() == 0) & (atk_labels_df["is_attack_th_type"].str.contains("Grid")==False) & (atk_labels_df[grid_col] < -1*grid_fac*atk_labels_df[grid_col].abs().max()), "is_attack_th_type"] += "Grid_dec"
                                              
                            atk_labels_df.loc[(atk_labels_df["charge_speed_lag_0_diff"].abs() > 0) & (atk_labels_df["is_attack_th_type"].str.contains("Grid")==False) & (atk_labels_df[grid_col].abs() > grid_fac*atk_labels_df[grid_col].abs().max()), "is_attack_th"] = -1
                            atk_labels_df.loc[(atk_labels_df["charge_speed_lag_0_diff"].abs() > 0) & (atk_labels_df["is_attack_th_type"].str.contains("Grid")==False) & (atk_labels_df[grid_col].abs() > grid_fac*atk_labels_df[grid_col].abs().max()), "is_attack_th_type"] += "SmallCombo"
                            atk_labels_df.loc[(atk_labels_df["charge_speed_lag_0_diff"].abs() > 0) & (atk_labels_df["is_attack_th_type"].str.contains("Grid")==False) & (atk_labels_df[grid_col] > grid_fac*atk_labels_df[grid_col].abs().max()), "is_attack_th_type"] += "Grid_inc"
                            atk_labels_df.loc[(atk_labels_df["charge_speed_lag_0_diff"].abs() > 0) & (atk_labels_df["is_attack_th_type"].str.contains("Grid")==False) & (atk_labels_df[grid_col] < -1*grid_fac*atk_labels_df[grid_col].abs().max()), "is_attack_th_type"] += "Grid_dec"

                            if False:
                                atk_labels_df.loc[(atk_labels_df["is_attack2"] == 1), "is_attack_th"] = 1 #clean up based on original is_attack flag

                            # _fac_2 = 0.5
                            # atk_labels_df.loc[atk_labels_df["charge_speed_lag_0_diff"].abs() > _fac_2*atk_labels_df["charge_speed_lag_0_diff"].abs().max(), "is_attack_th"] = -1
                            # atk_labels_df.loc[atk_labels_df[grid_col].abs() > _fac_2*atk_labels_df[grid_col].abs().max(), "is_attack_th"] = -1
                            # print(atk_labels_df)
                            # exit()
                        elif "is_attack_abs" in subcase:
                            th_fac = float(subcase[2])
                            atk_labels_df["is_attack_abs"] = 1
                            atk_labels_df["is_attack_th_type"] = "None"
                            atk_labels_df.loc[atk_labels_df["charge_speed_lag_0_diff"].abs() > th_fac, "is_attack_abs"] = -1 #diff between reported and should > threshold; captures FDI+/-, MAD+/-, but not MAD+/- with FDI==
                            atk_labels_df.loc[atk_labels_df["charge_speed_lag_0_diff"].abs() > th_fac, "is_attack_th_type"] += "FDI"
                            atk_labels_df.loc[atk_labels_df["charge_speed_lag_0_diff"] > th_fac, "is_attack_th_type"] += "OCPP_inc"
                            atk_labels_df.loc[atk_labels_df["charge_speed_lag_0_diff"] < -1*th_fac, "is_attack_th_type"] += "OCPP_dec"

                            grid_fac = th_fac
                            grid_col = [c for c in atk_labels_df.columns if "grid_expo_storage." in c and "_diff_to_custom_data_charge_speed_lag_0" in c]
                            if len(grid_col)>1:
                                logger.error(f"too many cols: {grid_col=}")
                            grid_col = grid_col[0]
                            atk_labels_df.loc[(atk_labels_df[grid_col].abs() > grid_fac), "is_attack_abs"] = -1
                            atk_labels_df.loc[(atk_labels_df[grid_col].abs() > grid_fac), "is_attack_th_type"] += "MAD"
                            atk_labels_df.loc[(atk_labels_df[grid_col] > grid_fac), "is_attack_th_type"] += "Grid_inc"
                            atk_labels_df.loc[(atk_labels_df[grid_col] < -1*grid_fac), "is_attack_th_type"] += "Grid_dec"

                            if True:
                                atk_labels_df.loc[(atk_labels_df["is_attack2"] == 1), "is_attack_abs"] = 1 #clean up based on original is_attack flag
                            

                        #TODO: try diff definitions
                        atk_labels_df["is_attack"] = atk_labels_df[subcase[0]] 
                        # print(atk_labels_df[atk_labels_df["is_attack"] == -1])
                        # exit()

                        #exit()
                        #fig, ax = plt.subplots()
                        for atk_type in atk_labels_df["is_attack_th_type"].drop_duplicates():
                            if atk_type!="NoneFDIOCPP_decMADGrid_inc":continue
                            if atk_type == "None":
                                for idx,row in out_df.iterrows():
                                    if row["clf"].split(".")[0] != "IsolationForest":continue
                                    atk_labels_df_type = atk_labels_df.copy()
                                    atk_labels_df_type["decision_function"] = row["decision_function"]
                                    if len(atk_labels_df_type[atk_labels_df_type["is_attack"] == -1]) == 0:
                                        continue
                                    # print(atk_labels_df_type[atk_labels_df_type["is_attack"] == -1])
                                    # exit()

                                    clf_name = row["clf"].split(".")[0]
                                    clf_full = row["clf"].replace("."+row["clf_r"],"")
                                    clf_r_name = row["clf_r"].split(".")[0]
                                    clf_n_params=".".join(clf_full.split(".")[1:])
                                    clf_r_params=".".join(row["clf_r"].split(".")[1:])

                                    fpr, tpr, thresholds = roc_curve(atk_labels_df_type["is_attack"], atk_labels_df_type["decision_function"], pos_label=-1)
                                    auc_v = auc(fpr, tpr)
                                    if auc_v < 0.35:
                                        fpr, tpr, thresholds = roc_curve(atk_labels_df_type["is_attack"], atk_labels_df_type["decision_function"]*-1, pos_label=-1)
                                    if not np.isnan(auc_v):
                                        totals_dict[actor_prefix][clf_name+"_"+clf_r_name]["overall"][row["clf_r"]][clf_full][group]["fpr"].append(fpr)
                                        totals_dict[actor_prefix][clf_name+"_"+clf_r_name]["overall"][row["clf_r"]][clf_full][group]["tpr"].append(tpr)
                            else:
                                for idx,row in out_df.iterrows():
                                    if row["clf"].split(".")[0] != "IsolationForest":continue
                                    atk_labels_df_c = atk_labels_df.copy()
                                    if len(atk_labels_df_c) != len(row["decision_function"]):
                                        raise Exception(f'len(atk_labels_df_c) != len(row["decision_function"]) -- {len(atk_labels_df_c)} != {len(row["decision_function"])}')
                                    atk_labels_df_c["decision_function"] = row["decision_function"]
                                    atk_labels_df_type = atk_labels_df_c[((atk_labels_df_c["is_attack_th_type"] == atk_type) & (atk_labels_df_c["is_attack"] == -1)) | ((atk_labels_df_c["is_attack_th_type"] == "None") & (atk_labels_df_c["is_attack"] == 1))]
                                    if len(atk_labels_df_type[atk_labels_df_type["is_attack"] == -1]) == 0:
                                        #logger.warning(f"no attacks for {atk_type}")
                                        continue
                                    # print(atk_labels_df_type[atk_labels_df_type["is_attack"] == -1])
                                    # exit()


                                    if True:#drop concecutive repeats to avoid inflation of TPR wrt. ROC of imprecise classifiers
                                        #print(atk_labels_df_type)
                                        #print(df_dict_path)
                                        #df_dict = load_single_feat(df_dict_path)
                                        #print(df_dict["cp_g_df"])
                                        #print(out_df)
                                        cs_typedf = atk_df_dict[group]["cp_g_df"].loc[atk_labels_df_type.index]["charge_speed_lag_0"].to_frame()
                                        cs_typedf["is_attack"] = atk_labels_df_type["is_attack"]
                                        #print(cs_typedf)
                                        # 1 2 2 3 0 0 0 0 0 3
                                        # _ _ 2 _ _ 0 0 0 0 _
                                        # _ _ _ _ _ _ 0 0 0 _
                                        labels = (cs_typedf["charge_speed_lag_0"] != cs_typedf["charge_speed_lag_0"].shift()).cumsum()
                                        cs_typedf['flag'] = (labels.map(labels.value_counts()) >= cut_off).astype(int)
                                        cs_typedf_not_same = cs_typedf[(cs_typedf["flag"] == 0) | (cs_typedf["is_attack"] == -1)] #deflate but keep all atks
                                        # print(cs_typedf_not_same)
                                        # print(atk_labels_df_type)
                                        atk_labels_df_type = atk_labels_df_type.loc[cs_typedf_not_same.index]

                                        if False:
                                            fig, ax = plt.subplots(figsize=(16,9))
                                            ax.plot(cs_typedf.index, cs_typedf["charge_speed_lag_0"], 'o', label="charge_speed_lag_0")
                                            ax.plot(cs_typedf_not_same.index, cs_typedf_not_same["charge_speed_lag_0"], 'o', label="charge_speed_lag_0 not same")
                                            plt.legend()
                                            plt.show()
                                            plt.close()                                            
                                            exit()

                                    clf_name = row["clf"].split(".")[0]
                                    clf_full = row["clf"].replace("."+row["clf_r"],"")
                                    clf_r_name = row["clf_r"].split(".")[0]
                                    clf_n_params=".".join(clf_full.split(".")[1:])
                                    clf_r_params=".".join(row["clf_r"].split(".")[1:])

                                    fpr, tpr, thresholds = roc_curve(atk_labels_df_type["is_attack"], atk_labels_df_type["decision_function"], pos_label=-1)
                                    auc_v = auc(fpr, tpr)
                                    if auc_v < 0.35:
                                        fpr, tpr, thresholds = roc_curve(atk_labels_df_type["is_attack"], atk_labels_df_type["decision_function"]*-1, pos_label=-1)
                                    if not np.isnan(auc_v):
                                        totals_dict[actor_prefix][clf_name+"_"+clf_r_name][atk_type][row["clf_r"]][clf_full][group]["fpr"].append(fpr)
                                        totals_dict[actor_prefix][clf_name+"_"+clf_r_name][atk_type][row["clf_r"]][clf_full][group]["tpr"].append(tpr)
                                    else:
                                        logger.warning(f"isnan for {atk_type}")

                        #plt.show()
                        #exit()
                        pbar.update(len(best_d_ag.keys()))

        plt.rc('text', usetex=USE_TEX)
        plt.rc('font', family='serif')
        font_size=22
        plt.rc('xtick',labelsize=font_size)
        plt.rc('ytick',labelsize=font_size)
        plt.rc('legend',fontsize=font_size)

        def sort_items(n):
            ret=defaultdict(lambda:list())
            #print(n)
            for _n in n:
                n_set = [r for r  in _n.split('.')]
                if "set_3" in n_set:
                    ret[1].append(_n)
                elif "set_13" in n_set:
                    ret[0].append(_n)
                elif "set_35" in n_set:
                    ret[2].append(_n)
                elif [__n for __n in n_set if "add_grid_load_expo_rnd_" in __n] and "set_352" in n_set:
                    __n=[__n for __n in n_set if "add_grid_load_expo_rnd_" in __n][0]
                    #print("add_grid_load_expo_rnd_", _n)
                    if len(__n.replace("add_grid_load_expo_rnd_","")) == 1:
                        __n_new = __n.replace("add_grid_load_expo_rnd_","add_grid_load_expo_rnd_0")
                        _n = _n.replace(__n,__n_new)
                    #print("2add_grid_load_expo_rnd_", _n)
                    ret[4].append(_n)
                elif "set_352" in n_set:
                    ret[3].append(_n)
                elif "no_grid_storage" in n_set and "no_grid_sgen" in n_set:
                    ret[1].append(_n)
                elif "no_grid" in n_set:
                    ret[0].append(_n)
                elif "add_grid_load_expo_static_100" in n_set:
                    ret[2].append(_n)
                else:
                    ret[9].append(_n)
            if 4 in ret:
                ret[4] = sorted(ret[4])
                #ret[4] = reversed(ret[4])
                ret[4] = [r.replace("add_grid_load_expo_rnd_0","add_grid_load_expo_rnd_") for r in ret[4]]
                #print(ret[4])
            for k in sorted(ret.keys()):
                for v in ret[k]:
                    yield v

        for fig_out_dir,totals_dict_ac in out_dict.items():
            fig_out_dir += "_comp"
            if FIX_atk_labels_based_on_CPO_data:
                fig_out_dir += "_fix"
            for actor_new,totals_dict in totals_dict_ac.items():
                if len(totals_dict)>0:
                    out_fig_d = fig_out_dir+("/clf_result_figs_"+args.case.replace("eval_clf_results","")+"/")+actor_new+"/"+"-".join(subcase)+"/"
                    Path(out_fig_d).mkdir(parents=True, exist_ok=True)
                    fig_dict=defaultdict(lambda:defaultdict(lambda:dict()))
                    for clf_name, d in tqdm(totals_dict.items(), desc=f"saving eval_clf_results2 {actor_new} {fig_out_dir}"):
                    #for clf_n, d in totals_dict.items():
                        for atk_type, d2 in d.items():
                            fig, ax = plt.subplots(figsize=(22,8))
                            plt.rc('axes', prop_cycle=(
                                                        #cycler('linestyle', ['-', '--', ':', '-.']) * #'-', '--', ':', '-.'
                                                        #cycler('color', [u'#1f77b4', u'#ff7f0e', u'#2ca02c', u'#d62728',]) #[  u'#1f77b4', u'#ff7f0e', u'#2ca02c', u'#d62728', u'#9467bd', u'#8c564b', u'#e377c2', u'#7f7f7f', u'#bcbd22', u'#17becf']
                                                        cycler('color', [  u'#1f77b4', u'#ff7f0e', u'#2ca02c', u'#d62728', u'#9467bd', u'#8c564b', u'#e377c2', u'#7f7f7f', u'#bcbd22', u'#17becf']) * #[  u'#1f77b4', u'#ff7f0e', u'#2ca02c', u'#d62728', u'#9467bd', u'#8c564b', u'#e377c2', u'#7f7f7f', u'#bcbd22', u'#17becf']
                                                        cycler('linestyle', ['-', ])  #'-', '--', ':', '-.'
                                                        ))
                            max_auc=0
                            min_auc=1
                            #for r_name, d3 in d2.items():
                            for r_name in sort_items(d2.keys()):
                                d3 = d2[r_name]
                                #print("r_name",r_name)
                                if "no_grid" not in r_name:
                                    #print("  skip r_name")
                                    continue
                                if ("no_grid_storage" in r_name or "no_grid_sgen" in r_name):
                                    #print("  skip r_name")
                                    continue
                                #for n_name, d4 in d3.items():
                                for n_name in sort_items(d3.keys()):
                                    #rint("n_name",n_name)
                                    if "add_grid_load_expo_rnd_90" in n_name:
                                        continue
                                    d4 = d3[n_name]
                                    mean_fpr = np.linspace(0, 1, 100) #based on https://scikit-learn.org/stable/auto_examples/model_selection/plot_roc_crossval.html
                                    tprs = []
                                    aucs = []
                                    #print(atk_type)
                                    for group, data in d4.items():
                                        for fpr, tpr in zip(data["fpr"], data["tpr"]):
                                            auc_v = auc(fpr, tpr)
                                            interp_tpr = np.interp(mean_fpr, fpr, tpr)
                                            interp_tpr[0] = 0.0
                                            tprs.append(interp_tpr)
                                            aucs.append(auc_v)
                                            # ax.plot(
                                            #     fpr,
                                            #     tpr,
                                            #     color="b",
                                            #     label=f"ROC {group} {clf_name}",
                                            #     lw=2,
                                            #     alpha=0.3,
                                            # )
                                    mean_tpr = np.mean(tprs, axis=0)
                                    mean_tpr[-1] = 1.0
                                    mean_auc = auc(mean_fpr, mean_tpr)
                                    if round(mean_auc, 2) == 0.66: mean_auc=0.65
                                    std_auc = np.std(aucs)
                                    max_auc=max(max_auc, mean_auc)
                                    min_auc=min(min_auc, mean_auc)

                                    n_set = [r for r  in n_name.split('.')[1:] if 'reg' not in r and 'norm' not in r]
                                    if "set_3" in n_set:
                                        n_set_name = "Nov.: Basic Grid"
                                    elif "set_13" in n_set:
                                        n_set_name = "Nov.: No Grid"
                                    elif "set_35" in n_set:
                                        n_set_name = "Nov.: Advanced Grid (no noise)"
                                    elif "set_352" in n_set:
                                        perc = [s for s in n_set if "add_grid_load_expo_rnd_" in s][0].split("_")[-1]
                                        if int(perc) > 10:
                                            continue
                                        n_set_name = r"Nov.: Advanced Grid ("+perc+r"\% noise)"
                                    else:
                                        n_set_name = ".".join(n_set)

                                    r_set = [r_name.split('.')[1]]+[r for r in r_name.split('.')[1:] if 'grid' in r]
                                    if "no_grid_storage" in r_set and "no_grid_sgen" in r_set:
                                        r_set_name = "Reg.: Basic Grid"
                                    elif "no_grid" in r_set:
                                        r_set_name = "Reg.: No Grid"
                                    elif "add_grid_load_expo_static_100" in r_set:
                                        r_set_name = "Reg.: Advanced Grid"
                                    else:
                                        r_set_name = ".".join(r_set)

                                    ax.plot(
                                        mean_fpr,
                                        mean_tpr,
                                        #color="b",
                                        label=f"{r_set_name} -- {n_set_name} (AUC = {round(mean_auc, 2)} $\pm$ {round(std_auc, 2)})", #{atk_type.replace('None','').replace('FDI','').replace('MAD','')} 
                                        #label=f"Mean ROC {atk_type} {r_name.split('.')[1:]} {n_name.split('.')[1:]} (AUC = {round(mean_auc, 2)} $\pm$ {round(std_auc, 2)} {len(d4), len(aucs)})",
                                        lw=2,
                                        alpha=0.8,
                                    )
                            out_fig_f=out_fig_d+f"{clf_name}_{atk_type}_{round(max_auc,2)}_{round(min_auc,2)}_44.pdf"

                            box = ax.get_position()
                            ax.set_position([box.x0, box.y0, box.width * (8/22), box.height])

                            ax.legend(loc='right', fontsize=font_size, bbox_to_anchor=(2.85,0.505))
                            #fig.show()
                            #ax.legend(loc='best', fontsize=font_size)

                            plt.gca().get_yaxis().get_major_formatter().set_useOffset(False)
                            plt.gca().get_yaxis().get_major_formatter().set_scientific(False)
                            plt.xlim([0,1])
                            plt.ylim([0,1])
                            plt.xlabel(r'\textbf{False Positive Rate (FPR)}', fontsize=font_size+2)
                            plt.ylabel(r'\textbf{True Positive Rate (TPR)}', fontsize=font_size+2)
                            fig.savefig(out_fig_f)
                            #plt.show()
                            plt.close()
                            # exit()


    elif args.case == "eval_clf_results4" or  args.case == "eval_clf_results4_cpo" or args.case == "eval_clf_results4_dso":
        allow_diff_shifts=True
        FIX_atk_labels_based_on_CPO_data=True
        _totals_dict=defaultdict(lambda: defaultdict(lambda: defaultdict(lambda: defaultdict(lambda: defaultdict(lambda: defaultdict(lambda: defaultdict(lambda: {"fpr":[],"tpr":[]})))))))
        out_dict=defaultdict(lambda: deepcopy(_totals_dict))
        df_dict_store=dict()
        for dataset in args.dataset:
            logger.info(f"{dataset=}")
            is_atk=False
            if "atk" in dataset:
                is_atk=True
                atk_dataset = dataset
                atk_OutDataDIR = configs[atk_dataset]["OutDataDIR"]

                #fig_out_dir = atk_OutDataDIR.replace(atk_dataset,configs[dataset]["BASE"]+"_"+"all_atks")

                ad = atk_dataset.split("_")
                if "new" in atk_dataset:
                    at_p=ad[4]
                    at_p="99"
                    _fig_out_dir = atk_OutDataDIR.replace(atk_dataset,ad[0]+"_"+ad[1]+"_"+ad[2]+"_all_"+at_p)
                else:
                    at_p=ad[3]
                    at_p="99"
                    _fig_out_dir = atk_OutDataDIR.replace(atk_dataset,ad[0]+"_"+ad[1]+"_all_"+at_p)
                totals_dict=out_dict[_fig_out_dir]
                

                dataset = configs[dataset]["BASE"]
                TRAIN_START_DATE=configs[dataset]["TRAIN_START_DATE"]
                VALIDATION_START_DATE=configs[dataset]["VALIDATION_START_DATE"]
                ATK_START_DATE=configs[dataset]["ATK_START_DATE"]
            else:
                raise ValueError(f"not an attack data set: {dataset}")

            atk_OutDataDIR_full=atk_OutDataDIR
            OutDataDIR = configs[dataset]["OutDataDIR"]
            if not os.path.isdir(OutDataDIR):
                newOutDataDIR = OutDataDIR.replace("data", "data_full")
                logger.info(f"changing {OutDataDIR=} to {newOutDataDIR}")
                OutDataDIR = newOutDataDIR
                atk_OutDataDIR_full = atk_OutDataDIR.replace("data", "data_full")

            ret_d = get_clf_feat_file_dicts(OutDataDIR)
            atk_ret_d = get_clf_feat_file_dicts(atk_OutDataDIR_full)

            best_d = load_eval_dicts_clf(OutDataDIR)

            x_path=atk_OutDataDIR+"/clf_results/"
            full_actors = [f for f in os.listdir(x_path) if os.path.isdir(os.path.join(x_path, f))]
            #for actor in sorted(ret_d.keys()):
            for actor in sorted(full_actors):
                if args.case == "eval_clf_results4_cpo" and not actor.startswith("CPO"):
                    continue
                if args.case == "eval_clf_results4_dso" and not actor.startswith("DSO"):
                    continue

                if len(actor.split(".")) > 1:
                    features = actor.split(".")[1:]
                else:
                    features=["all"]
                    logger.warning(f"assuming all features for {actor}")


                actor_prefix = actor.split(".")[0]
                actor_prefix_new = (actor_prefix+"_"+args.atk_subset) if args.atk_subset else actor_prefix
                actor_new = actor.replace(actor_prefix, actor_prefix_new)
                actor_prefix = actor_prefix_new

                atk_act_feat = load_feats(atk_OutDataDIR_full, actor_prefix)
                atk_df_dict = {group: df_dict for group, df_dict in atk_act_feat}
                #act_feat = load_feats(OutDataDIR, actor_prefix) #generator
                act_feat = iter_feats(OutDataDIR, actor_prefix) #generator
                total=load_feats_len(OutDataDIR, actor_prefix)
                #print(actor, actor_prefix) #DSO.all DSO


                best_d_len = len([k2 for k,v in best_d[actor].items() for k2,v2 in best_d[actor][k].items()])

                subcase=("is_attack1", "1", ) #any speed diff
                subcase=("is_attack2", "1", ) #is_attack flag was set
                subcase=("is_attack_th", "1", "0.5")
                subcase=("is_attack_th", "3", "0.05")
                subcase=("is_attack_th", "1", "0.05")
                subcase=("is_attack_abs", "1", "500")

                cut_off=40
                if args.eval_fac is not None:
                    if "." in args.eval_fac:
                        cut_off=int(args.eval_fac.split(".")[1])
                    subcase=("is_attack_abs", "1", args.eval_fac.split(".")[0])

                with tqdm(total=best_d_len, desc="eval_clf_results "+actor_new) as pbar:
                    g_i=0
                    for group, df_dict_path in act_feat:
                        g_i+=1
                        best_d_ag = best_d[actor][group]
                        pbar.set_description(f'eval_clf_results {actor_new} {group} ({g_i}/{total})')
                        out_d = atk_OutDataDIR+"/clf_results/"+actor_new+"/"
                        out_f = out_d + group+".csv.gz"

                        if not os.path.isfile(out_f):
                            logger.warning(f"skipping missing file {out_f}")
                            pbar.update(len(best_d_ag.keys()))
                            continue

                        atk_labels_d = atk_OutDataDIR+"/clf_is_atk_labels/"+actor_new+"/"
                        atk_labels_f = atk_labels_d + group+".csv.gz"
                        if not os.path.isfile(atk_labels_f):
                            logger.warning(f"skipping missing atk label file {atk_labels_f}")
                            pbar.update(len(best_d_ag.keys()))
                            continue

                        #print(best_d_ag)


                        atk_labels_df = pd.read_csv(atk_labels_f, index_col=0, parse_dates=[0])

                        if False:
                            grid_col = [c for c in atk_labels_df.columns if "grid_expo_storage." in c and "_diff_to_custom_data_charge_speed_lag_0" in c]
                            for gc in grid_col:
                                print(atk_labels_df[atk_labels_df[gc] != 0][gc].abs().sort_values())
                            print(atk_labels_df[atk_labels_df["charge_speed_lag_0_diff"] != 0]["charge_speed_lag_0_diff"].abs().sort_values())
                            exit()

                        out_df = pd.read_csv(out_f, index_col=0, parse_dates=[0])
                        out_df["decision_function"] = out_df["decision_function"].apply(lambda x: to_float_list(x))
                        #print(out_df)
                        #print(atk_labels_df)
                        if "is_attack_th" in subcase:
                            th_fac = float(subcase[2])
                            atk_labels_df["is_attack_th"] = 1
                            atk_labels_df["is_attack_th_type"] = "None"
                            atk_labels_df.loc[atk_labels_df["charge_speed_lag_0_diff"].abs() > th_fac*atk_labels_df["charge_speed_lag_0_diff"].abs().max(), "is_attack_th"] = -1 #diff between reported and should > threshold; captures FDI+/-, MAD+/-, but not MAD+/- with FDI==
                            atk_labels_df.loc[atk_labels_df["charge_speed_lag_0_diff"].abs() > th_fac*atk_labels_df["charge_speed_lag_0_diff"].abs().max(), "is_attack_th_type"] += "FDI"
                            atk_labels_df.loc[atk_labels_df["charge_speed_lag_0_diff"] > th_fac*atk_labels_df["charge_speed_lag_0_diff"].abs().max(), "is_attack_th_type"] += "OCPP_inc"
                            atk_labels_df.loc[atk_labels_df["charge_speed_lag_0_diff"] < -1*th_fac*atk_labels_df["charge_speed_lag_0_diff"].abs().max(), "is_attack_th_type"] += "OCPP_dec"

                            grid_fac = th_fac
                            grid_col = [c for c in atk_labels_df.columns if "grid_expo_storage." in c and "_diff_to_custom_data_charge_speed_lag_0" in c]
                            if len(grid_col)>1:
                                logger.error(f"too many cols: {grid_col=}")
                            grid_col = grid_col[0]
                            atk_labels_df.loc[(atk_labels_df["is_attack_th"] == -1) & (atk_labels_df[grid_col].abs() > grid_fac*atk_labels_df[grid_col].abs().max()), "is_attack_th_type"] += "Combo"
                            atk_labels_df.loc[(atk_labels_df["is_attack_th"] == -1) & (atk_labels_df[grid_col] > grid_fac*atk_labels_df[grid_col].abs().max()), "is_attack_th_type"] += "Grid_inc"
                            atk_labels_df.loc[(atk_labels_df["is_attack_th"] == -1) & (atk_labels_df[grid_col] < -1*grid_fac*atk_labels_df[grid_col].abs().max()), "is_attack_th_type"] += "Grid_dec"

                            #grid_fac = th_fac*2
                            grid_fac = th_fac*2
                            #print(atk_labels_df["is_attack_th"].value_counts())
                            #grid_col = [c for c in atk_labels_df.columns if "grid_expo_storage." in c and "_diff_to_charge_speed_lag_0" in c][0]
                            atk_labels_df.loc[(atk_labels_df["charge_speed_lag_0_diff"].abs() == 0) & (atk_labels_df["is_attack_th_type"].str.contains("Grid")==False) & (atk_labels_df[grid_col].abs() > grid_fac*atk_labels_df[grid_col].abs().max()), "is_attack_th"] = -1 #diff between grid measurement and reported > threshold; captures MAD+/- with  any FDI (also FDI==)
                            atk_labels_df.loc[(atk_labels_df["charge_speed_lag_0_diff"].abs() == 0) & (atk_labels_df["is_attack_th_type"].str.contains("Grid")==False) & (atk_labels_df[grid_col].abs() > grid_fac*atk_labels_df[grid_col].abs().max()), "is_attack_th_type"] += "MAD"
                            atk_labels_df.loc[(atk_labels_df["charge_speed_lag_0_diff"].abs() == 0) & (atk_labels_df["is_attack_th_type"].str.contains("Grid")==False) & (atk_labels_df[grid_col] > grid_fac*atk_labels_df[grid_col].abs().max()), "is_attack_th_type"] += "Grid_inc"
                            atk_labels_df.loc[(atk_labels_df["charge_speed_lag_0_diff"].abs() == 0) & (atk_labels_df["is_attack_th_type"].str.contains("Grid")==False) & (atk_labels_df[grid_col] < -1*grid_fac*atk_labels_df[grid_col].abs().max()), "is_attack_th_type"] += "Grid_dec"
                                              
                            atk_labels_df.loc[(atk_labels_df["charge_speed_lag_0_diff"].abs() > 0) & (atk_labels_df["is_attack_th_type"].str.contains("Grid")==False) & (atk_labels_df[grid_col].abs() > grid_fac*atk_labels_df[grid_col].abs().max()), "is_attack_th"] = -1
                            atk_labels_df.loc[(atk_labels_df["charge_speed_lag_0_diff"].abs() > 0) & (atk_labels_df["is_attack_th_type"].str.contains("Grid")==False) & (atk_labels_df[grid_col].abs() > grid_fac*atk_labels_df[grid_col].abs().max()), "is_attack_th_type"] += "SmallCombo"
                            atk_labels_df.loc[(atk_labels_df["charge_speed_lag_0_diff"].abs() > 0) & (atk_labels_df["is_attack_th_type"].str.contains("Grid")==False) & (atk_labels_df[grid_col] > grid_fac*atk_labels_df[grid_col].abs().max()), "is_attack_th_type"] += "Grid_inc"
                            atk_labels_df.loc[(atk_labels_df["charge_speed_lag_0_diff"].abs() > 0) & (atk_labels_df["is_attack_th_type"].str.contains("Grid")==False) & (atk_labels_df[grid_col] < -1*grid_fac*atk_labels_df[grid_col].abs().max()), "is_attack_th_type"] += "Grid_dec"

                            if False:
                                atk_labels_df.loc[(atk_labels_df["is_attack2"] == 1), "is_attack_th"] = 1 #clean up based on original is_attack flag

                            # _fac_2 = 0.5
                            # atk_labels_df.loc[atk_labels_df["charge_speed_lag_0_diff"].abs() > _fac_2*atk_labels_df["charge_speed_lag_0_diff"].abs().max(), "is_attack_th"] = -1
                            # atk_labels_df.loc[atk_labels_df[grid_col].abs() > _fac_2*atk_labels_df[grid_col].abs().max(), "is_attack_th"] = -1
                            # print(atk_labels_df)
                            # exit()
                        elif "is_attack_abs" in subcase:
                            th_fac = float(subcase[2])
                            atk_labels_df["is_attack_abs"] = 1
                            atk_labels_df["is_attack_th_type"] = "None"
                            atk_labels_df.loc[atk_labels_df["charge_speed_lag_0_diff"].abs() > th_fac, "is_attack_abs"] = -1 #diff between reported and should > threshold; captures FDI+/-, MAD+/-, but not MAD+/- with FDI==
                            atk_labels_df.loc[atk_labels_df["charge_speed_lag_0_diff"].abs() > th_fac, "is_attack_th_type"] += "FDI"
                            atk_labels_df.loc[atk_labels_df["charge_speed_lag_0_diff"] > th_fac, "is_attack_th_type"] += "OCPP_inc"
                            atk_labels_df.loc[atk_labels_df["charge_speed_lag_0_diff"] < -1*th_fac, "is_attack_th_type"] += "OCPP_dec"

                            grid_fac = th_fac
                            grid_col = [c for c in atk_labels_df.columns if "grid_expo_storage." in c and "_diff_to_custom_data_charge_speed_lag_0" in c]
                            if len(grid_col)>1:
                                logger.error(f"too many cols: {grid_col=}")
                            grid_col = grid_col[0]
                            atk_labels_df.loc[(atk_labels_df[grid_col].abs() > grid_fac), "is_attack_abs"] = -1
                            atk_labels_df.loc[(atk_labels_df[grid_col].abs() > grid_fac), "is_attack_th_type"] += "MAD"
                            atk_labels_df.loc[(atk_labels_df[grid_col] > grid_fac), "is_attack_th_type"] += "Grid_inc"
                            atk_labels_df.loc[(atk_labels_df[grid_col] < -1*grid_fac), "is_attack_th_type"] += "Grid_dec"

                            if True:
                                atk_labels_df.loc[(atk_labels_df["is_attack2"] == 1), "is_attack_abs"] = 1 #clean up based on original is_attack flag
                            

                        #TODO: try diff definitions
                        atk_labels_df["is_attack"] = atk_labels_df[subcase[0]] 
                        atk_labels_df["ocpp_diff"] = atk_labels_df["charge_speed_lag_0_diff"] 
                        atk_labels_df["grid_diff"] = atk_labels_df[grid_col] 
                        # print(atk_labels_df[atk_labels_df["is_attack"] == -1])
                        # exit()
                        if (actor_prefix,group) not in df_dict_store:
                            df_dict = load_single_feat(df_dict_path)
                            df_dict_store[(actor_prefix,group)] = df_dict
                        g_max_speed = df_dict_store[(actor_prefix,group)]["cp_g_df"]["charge_speed_lag_0"].max()
                        atk_labels_df["ocpp_diff_rel"] = atk_labels_df["ocpp_diff"] / g_max_speed
                        atk_labels_df["grid_diff_rel"] = atk_labels_df["grid_diff"] / g_max_speed
                        atk_labels_df["ocpp_diff_rel"] *= 100
                        atk_labels_df["grid_diff_rel"] *= 100
                        #print(atk_labels_df)

                        #exit()
                        #fig, ax = plt.subplots()
                        for atk_type in atk_labels_df["is_attack_th_type"].drop_duplicates():
                            #if atk_type!="NoneFDIOCPP_decMADGrid_inc":continue
                            if atk_type == "None":
                                for idx,row in out_df.iterrows():
                                    if row["clf"].split(".")[0] != "IsolationForest":continue
                                    if row["clf_r"].split(".")[0] != "LinearSVR":continue
                                    for diff_t in ["ocpp_diff_rel", "grid_diff_rel"]:
                                        for atk_size in range(-50,51,2):
                                            atk_size_l = diff_t+"."+str(atk_size)
                                            atk_labels_df_type = atk_labels_df.copy()
                                            atk_labels_df_type["decision_function"] = row["decision_function"]

                                            if atk_size < -2:
                                                atk_labels_df_type = atk_labels_df_type[atk_labels_df_type[diff_t] <= atk_size]
                                            elif atk_size > 2:
                                                atk_labels_df_type = atk_labels_df_type[atk_labels_df_type[diff_t] >= atk_size]
                                            else:
                                                continue

                                            if len(atk_labels_df_type[atk_labels_df_type["is_attack"] == -1]) == 0:
                                                continue
                                            # print(atk_labels_df_type[atk_labels_df_type["is_attack"] == -1])
                                            # exit()

                                            clf_name = row["clf"].split(".")[0]
                                            clf_full = row["clf"].replace("."+row["clf_r"],"")
                                            clf_r_name = row["clf_r"].split(".")[0]
                                            clf_n_params=".".join(clf_full.split(".")[1:])
                                            clf_r_params=".".join(row["clf_r"].split(".")[1:])

                                            y = atk_labels_df_type["is_attack"].values
                                            y_pred = [1 if x >= round(row["clf_o"],2) else -1 for x in atk_labels_df_type["decision_function"].values]
                                            tn, fp, fn, tp = confusion_matrix(y, y_pred, labels=[1, -1]).ravel()
                                            #accuracy = accuracy_score(y, y_pred)
                                            if (tp+fn) != 0: tpr = tp/(tp+fn)
                                            else:  tpr=0
                                            if (fp+tn) != 0: fpr = fp/(fp+tn)
                                            else:  fpr=0
                                                
                                            totals_dict[actor_prefix][clf_name+"_"+clf_r_name]["overall"][row["clf_r"]][clf_full][group][atk_size_l]["fpr"].append(fpr)
                                            totals_dict[actor_prefix][clf_name+"_"+clf_r_name]["overall"][row["clf_r"]][clf_full][group][atk_size_l]["tpr"].append(tpr)
                            elif False:
                                for idx,row in out_df.iterrows():
                                    if row["clf"].split(".")[0] != "IsolationForest":continue
                                    if row["clf_r"].split(".")[0] != "LinearSVR":continue
                                    for diff_t in ["ocpp_diff_rel", "grid_diff_rel"]:
                                        for atk_size in range(-50,51,2):
                                            atk_size_l = diff_t+"."+str(atk_size)
                                            # print(row)
                                            # print(row["clf_o"])
                                            # print(best_d_ag[row["clf"]]["offset"])
                                            # exit()
                                            atk_labels_df_c = atk_labels_df.copy()
                                            if len(atk_labels_df_c) != len(row["decision_function"]):
                                                raise Exception(f'len(atk_labels_df_c) != len(row["decision_function"]) -- {len(atk_labels_df_c)} != {len(row["decision_function"])}')
                                            atk_labels_df_c["decision_function"] = row["decision_function"]
                                            atk_labels_df_type = atk_labels_df_c[((atk_labels_df_c["is_attack_th_type"] == atk_type) & (atk_labels_df_c["is_attack"] == -1)) | ((atk_labels_df_c["is_attack_th_type"] == "None") & (atk_labels_df_c["is_attack"] == 1))]

                                            if atk_size < -2:
                                                atk_labels_df_type = atk_labels_df_type[atk_labels_df_type[diff_t] <= atk_size]
                                            elif atk_size > 2:
                                                atk_labels_df_type = atk_labels_df_type[atk_labels_df_type[diff_t] >= atk_size]
                                            else:
                                                continue

                                            if len(atk_labels_df_type[atk_labels_df_type["is_attack"] == -1]) == 0:
                                                #logger.warning(f"no attacks for {atk_type}")
                                                continue
                                            # print(atk_labels_df_type[atk_labels_df_type["is_attack"] == -1])
                                            # exit()


                                            if True:#drop concecutive repeats to avoid inflation of TPR wrt. ROC of imprecise classifiers
                                                #print(atk_labels_df_type)
                                                #print(df_dict_path)
                                                #df_dict = load_single_feat(df_dict_path)
                                                #print(df_dict["cp_g_df"])
                                                #print(out_df)
                                                cs_typedf = atk_df_dict[group]["cp_g_df"].loc[atk_labels_df_type.index]["charge_speed_lag_0"].to_frame()
                                                cs_typedf["is_attack"] = atk_labels_df_type["is_attack"]
                                                #print(cs_typedf)
                                                # 1 2 2 3 0 0 0 0 0 3
                                                # _ _ 2 _ _ 0 0 0 0 _
                                                # _ _ _ _ _ _ 0 0 0 _
                                                labels = (cs_typedf["charge_speed_lag_0"] != cs_typedf["charge_speed_lag_0"].shift()).cumsum()
                                                cs_typedf['flag'] = (labels.map(labels.value_counts()) >= cut_off).astype(int)
                                                cs_typedf_not_same = cs_typedf[(cs_typedf["flag"] == 0) | (cs_typedf["is_attack"] == -1)] #deflate but keep all atks
                                                # print(cs_typedf_not_same)
                                                # print(atk_labels_df_type)
                                                atk_labels_df_type = atk_labels_df_type.loc[cs_typedf_not_same.index]

                                                if False:
                                                    fig, ax = plt.subplots(figsize=(16,9))
                                                    ax.plot(cs_typedf.index, cs_typedf["charge_speed_lag_0"], 'o', label="charge_speed_lag_0")
                                                    ax.plot(cs_typedf_not_same.index, cs_typedf_not_same["charge_speed_lag_0"], 'o', label="charge_speed_lag_0 not same")
                                                    plt.legend()
                                                    plt.show()
                                                    plt.close()                                            
                                                    exit()

                                            clf_name = row["clf"].split(".")[0]
                                            clf_full = row["clf"].replace("."+row["clf_r"],"")
                                            clf_r_name = row["clf_r"].split(".")[0]
                                            clf_n_params=".".join(clf_full.split(".")[1:])
                                            clf_r_params=".".join(row["clf_r"].split(".")[1:])

                                            y = atk_labels_df_type["is_attack"].values
                                            y_pred = [1 if x >= row["clf_o"] else -1 for x in atk_labels_df_type["decision_function"].values]
                                            #y_pred = [1 if x >= 0 else -1 for x in atk_labels_df_type["decision_function"].values]
                                            tn, fp, fn, tp = confusion_matrix(y, y_pred, labels=[1, -1]).ravel()
                                            #accuracy = accuracy_score(y, y_pred)
                                            if (tp+fn) != 0: tpr = tp/(tp+fn)
                                            else:  tpr=0
                                            if (fp+tn) != 0: fpr = fp/(fp+tn)
                                            else:  fpr=0

                                            totals_dict[actor_prefix][clf_name+"_"+clf_r_name][atk_type][row["clf_r"]][clf_full][group][atk_size_l]["fpr"].append(fpr)
                                            totals_dict[actor_prefix][clf_name+"_"+clf_r_name][atk_type][row["clf_r"]][clf_full][group][atk_size_l]["tpr"].append(tpr)

                        #plt.show()
                        #exit()
                        pbar.update(len(best_d_ag.keys()))

        plt.rc('text', usetex=USE_TEX)
        plt.rc('font', family='serif')
        font_size=22
        plt.rc('xtick',labelsize=font_size)
        plt.rc('ytick',labelsize=font_size)
        plt.rc('legend',fontsize=font_size)


        for fig_out_dir,totals_dict_ac in out_dict.items():
            fig_out_dir += "_comp"
            if FIX_atk_labels_based_on_CPO_data:
                fig_out_dir += "_fix"
            for actor_new,totals_dict in totals_dict_ac.items():
                if len(totals_dict)>0:
                    out_fig_d = fig_out_dir+("/clf_result_figs_"+args.case.replace("eval_clf_results","")+"/")+actor_new+"/"+"-".join(subcase)+"/"
                    Path(out_fig_d).mkdir(parents=True, exist_ok=True)
                    fig_dict=defaultdict(lambda:defaultdict(lambda:dict()))
                    for clf_name, d in tqdm(totals_dict.items(), desc=f"saving eval_clf_results2 {actor_new} {fig_out_dir}"):
                    #for clf_n, d in totals_dict.items():
                        for atk_type, d2 in d.items():
                            #for r_name, d3 in d2.items():
                            for r_name in d2.keys():
                                d3 = d2[r_name]
                                #fig_out_dir+="_"+r_name
                                for n_name in d3.keys():
                                    #rint("n_name",n_name)
                                    #fig_out_dir+="_"+n_name
                                    fig, ax = plt.subplots(figsize=(22,8))
                                    plt.rc('axes', prop_cycle=(
                                                                #cycler('linestyle', ['-', '--', ':', '-.']) * #'-', '--', ':', '-.'
                                                                #cycler('color', [u'#1f77b4', u'#ff7f0e', u'#2ca02c', u'#d62728',]) #[  u'#1f77b4', u'#ff7f0e', u'#2ca02c', u'#d62728', u'#9467bd', u'#8c564b', u'#e377c2', u'#7f7f7f', u'#bcbd22', u'#17becf']
                                                                cycler('color', [  u'#1f77b4', u'#ff7f0e', u'#2ca02c', u'#d62728', u'#9467bd', u'#8c564b', u'#e377c2', u'#7f7f7f', u'#bcbd22', u'#17becf']) * #[  u'#1f77b4', u'#ff7f0e', u'#2ca02c', u'#d62728', u'#9467bd', u'#8c564b', u'#e377c2', u'#7f7f7f', u'#bcbd22', u'#17becf']
                                                                cycler('linestyle', ['-', ])  #'-', '--', ':', '-.'
                                                                ))
                                    
                                    d4 = d3[n_name]
                                    tpr_dict=defaultdict(lambda:list())
                                    fpr_dict=defaultdict(lambda:list())
                                    for group, data in d4.items():
                                        for atk_size_l, data2 in data.items():
                                            #print(atk_size_l, data2)
                                            for fpr, tpr in zip(data2["fpr"], data2["tpr"]):
                                                fpr_dict[atk_size_l].append(fpr)
                                                tpr_dict[atk_size_l].append(tpr)

                                    # print(fpr_dict)
                                    # print(tpr_dict)
                                    plot_fpr=defaultdict(lambda:list())
                                    plot_tpr=defaultdict(lambda:list())
                                    total_fpr=list()
                                    for diff_t in ["ocpp_diff_rel", "grid_diff_rel"]:
                                        for atk_size in range(-50,51,2):
                                            atk_size_l = diff_t+"."+str(atk_size)
                                            if atk_size_l in fpr_dict:
                                                plot_fpr[diff_t].append(mean(fpr_dict[atk_size_l]))
                                                total_fpr.append(mean(fpr_dict[atk_size_l]))
                                            else:
                                                plot_fpr[diff_t].append(0)
                                            if atk_size_l in tpr_dict:
                                                plot_tpr[diff_t].append(mean(tpr_dict[atk_size_l]))
                                            else:
                                                plot_tpr[diff_t].append(0)
                                    plot_atk_size=[atk_size for atk_size in range(-50,51,2)]
                                    plot_mean_fpr=[mean(total_fpr) for atk_size in range(-50,51,2)]
                                    # print(plot_fpr)
                                    # print(plot_tpr)
                                    # print(plot_atk_size)
                                    
                                    for diff_t in ["ocpp_diff_rel", "grid_diff_rel"]:
                                        if diff_t == "ocpp_diff_rel":
                                            diff_t_label="Change in Reported Load"
                                        elif diff_t == "grid_diff_rel":
                                            diff_t_label="Change in Grid Load"
                                            
                                        ax.plot(
                                            plot_atk_size,
                                            plot_tpr[diff_t],
                                            #color="b",
                                            label=f"TPR over {diff_t_label}", #{atk_type.replace('None','').replace('FDI','').replace('MAD','')} 
                                            #label=f"TPR over {diff_t_label} (mean = {round(mean([v for v in plot_tpr[diff_t] if v != 0]), 2)})", #{atk_type.replace('None','').replace('FDI','').replace('MAD','')} 
                                            lw=2,
                                            alpha=0.8,
                                        )
                                        # ax.plot(
                                        #     plot_atk_size,
                                        #     plot_fpr[diff_t],
                                        #     #color="b",
                                        #     label=f"FPR over {diff_t_label} (mean = {round(mean(plot_fpr[diff_t]), 2)})", #{atk_type.replace('None','').replace('FDI','').replace('MAD','')} 
                                        #     lw=2,
                                        #     alpha=0.8,
                                        # )
                                    ax.plot(
                                        plot_atk_size,
                                        plot_mean_fpr,
                                        #color="b",
                                        label=f"FPR (mean = {round(mean(plot_mean_fpr), 2)})", #{atk_type.replace('None','').replace('FDI','').replace('MAD','')} 
                                        lw=2,
                                        alpha=0.8,
                                    )
                                    out_fig_f=out_fig_d+f"{r_name}_{n_name}_{atk_type}_44.pdf"
                                    # print(r_name,n_name)
                                    # print(out_fig_f)

                                    box = ax.get_position()
                                    ax.set_position([box.x0, box.y0, box.width * (8/22)*0.8, box.height*0.8])

                                    ax.legend(loc='right', fontsize=font_size, bbox_to_anchor=(2.95,0.505))
                                    #fig.show()
                                    #ax.legend(loc='best', fontsize=font_size)
                                    plt.xticks([-50,-25,0,25,50])

                                    plt.gca().get_yaxis().get_major_formatter().set_useOffset(False)
                                    plt.gca().get_yaxis().get_major_formatter().set_scientific(False)
                                    plt.ylim([0,1])
                                    plt.xlim([-50,50])
                                    plt.xlabel(r'\textbf{Min. Attack Magnitude}', fontsize=font_size+2)
                                    plt.ylabel(r'\textbf{TPR/FPR}', fontsize=font_size+2)
                                    if os.path.isfile(out_fig_f):
                                        logger.warning(f"overwriting {out_fig_f=}")
                                    fig.savefig(out_fig_f)
                                    #plt.show()
                                    plt.close()
                                    # exit()